/*****************************************************
**  lib-rep-read.c                       (library)  **
**  ----------------------------------------------  **
**                                                  **
************** BORING LEGAL NOTICE *******************
**                                                  **
**  Assassins' Game (Semi-)Automatic Umpire System  **
**  ----------------------------------------------  **
**  Prerelease: Copyright 2000-2002 Edward M Nokes  **
**  This code may not be distributed in source or   **
**  executable form without the express consent of  **
**  the author, Edward M Nokes (emn23@cam.ac.uk).   **
**  The author accepts no responsibility or         **
**  liability for anything occurring as a result    **
**  of using this software, etc. etc.               **
**                                                  **
*****************************************************/

#ifdef included__lib_rep_read
 #error "** lib-rep-read multiply #included **"
#else
 #define included__lib_rep_read fromfile

/***************************************************************************************************************************
** Definitions and declarations                                                                                           **
***************************************************************************************************************************/


/***************************************************************************************************************************
** Read information on report(s) from the "reports.txt" file. If (*rp_readp) is null on entry, a 'report' struct is       **
** allocated and filled in for each report in-file, and the head of the resultant linked list returned as (*rp_readp)     **
** If non-null on entry, (*rp_read) should point to a list of 'interesting' reports, linked via the 'link_all' member:    **
** each entry in the list must already contain a report id; the rest of each structure will be filled in from the file.   **
** If (freps) is non-null, the file is NOT closed on completion; the open file pointer will be returned as (*freps).      **
** (mode) b0 set => load report text for all reports, if (*rp_readp) null, else only for reports with RF_LoadText flag    **
**                  set on entry; b0 clear => don't load any report text.                                                 **
** (mode) b1 set => set 'fpos' field in report structs to start of report headers; clear => set to start of report text.  **
** The RF_Already flag is cleared for every report.                                                                       **
***************************************************************************************************************************/
int repfile(report_p *rp_readp, player_p plist, char** memst, char* mend, FILE** freps, int mode)
{
	int i, n = 0, rc = 0;
	report_p rptr, *lrpp, rp_check, rp_head;
	player_p pptr;
	FILE *fp = 0;
	char *mem, *rid = 0;
	long rlen, dec, rstfpos;
	report_flags rflg;


	/* Paranoia */
	if (!rp_readp) erf(ERF_BadParms, 1);

	/* Ensure at least bytes free in buffer */
	if ((mend - *memst) < chk[CHK_reps].len) erf(ERF_BufferExhaust, 60);

	/* List of 'interesting' reports to check against */
	for (rptr = rp_check = rp_head = *rp_readp; (rptr); rptr = rptr->link_all)
	{
		/* Optimisation: form sub list to be checked */
		rptr->link_sub = rptr->link_all;

		/* Paranoia - check report ids are non-null */
		if (!rptr->rid) erf(ERF_BadParms, 30);
	}

	/* Open reports file */
	if (!(fp = fopen(FILE_Reports, "rb"))) erf(ERF_fopenReps, 2);

	/* Check file header (ignoring initial comments) */
	if (i = chkheader(fp, CHK_reps, *memst, 0, 0, 0, 0)) erf(i, 61);

	/* Scan through for valid lines */
	while ((i = fgetc(fp)) != EOF)
	{
		/* Ignore lines not beginning with "*" character */
		switch (i)
		{
			case '*'  : break;
			default   : toeol(fp);
			case '\n' : continue;
		}

		/* Get position of start of report */
		if ((mode & 2) && (-1L == (rstfpos = ftell(fp)))) erf(ERF_ftellFailed, 3);

		/* Skip initial tabs, read 'repid' */
		if (i = tread(fp, rid = mem = *memst, mend, &mem)) erf(ERF_BadTread+i, 4); 

		/*fprintf(stderr, "rep[%s]", rid);*/

		/* Acquire report flags */
		if (i = tread(fp, mem, mend, 0)) erf(ERF_BadTread+i, 5);

		/* Convert it */
		rflg = strtoul(mem, 0, 16) & ~(RF_Already|RF_LoadText);

		/* Acquire report length */
		if (i = tread(fp, mem, mend, 0)) erf(ERF_BadTread+i, 6);

		/* Convert it */
		rlen = atol(mem);

		/* Interested in this one? */
		for (rptr = *(lrpp = &rp_head); (rptr); rptr = *(lrpp = &rptr->link_sub)) if (streq(rptr->rid, rid)) break;

		/* Skip this one (uninterested) */
		if ((!rptr) && (rp_check)) toeol(fp);

		/* Read full report information */
		else
		{
			/* 'New' report - add to list */
			if (!rptr)
			{
				/* Allocate new struct */
				if (!(rptr = malloc(sizeof(struct report)))) erf(ERF_mallocFailed, 7);

				/* Link into chain */
				rptr->link_all = rp_head;
				rptr->rid = rid;
				rptr->flags = (mode & 1) ? RF_LoadText : 0;
				rp_head = rptr;
			}

			/* Optimisation: delink from future checklist */
			else *lrpp = rptr->link_sub;

			/* Fill in struct */
			rptr->flags = rflg | ((mode & 1) ? (rptr->flags & RF_LoadText) : 0);
			rptr->length = rlen;
			rptr->link_sub = 0;
			rptr->text = 0;
			rptr->fpos = -1L;

			/* Acquire author's id */
			if (i = tread(fp, rptr->pid = mem, mend, &mem)) erf(ERF_BadTread+i, 8);

			/* Attempt to match id in player list */
			for (pptr = plist; (pptr); pptr = pptr->link) if (streq(pptr->id, rptr->pid)) break;

			/* Store player pointer */
			if (!(rptr->pptr = pptr) && (plist))
			{
				fprintf(stderr, "%li:   repfile(8): missing player \"%s\" %p.\n", uniqid, rptr->pid, plist);
				erf(ERF_MissingPid, 9);
			}

			/* Pseudonym */
			if ((i = tread(fp, rptr->pseud = mem, mend, &mem)) > 1) erf(ERF_BadTread+i, 10);

			/* "-" means to use default */
			if (streq(rptr->pseud, "-")) rptr->pseud = 0;

			/* Optional fields */
			if (i == 0)
			{
				/* Read time of last report edit */
				if ((i = tread(fp, mem, mend, 0)) > 1) erf(ERF_BadTread+i, 20);

				/* Convert it */
				if (!str2tim(mem, &rptr->uptim)) erf(ERF_BadDateText, 21);

				/* Ignore event ref field */
			}

			/* Default time of 00:00:00 if none given */
			else ZERO_TIMER(rptr->uptim);

			/* Ignore the rest of the line */
			toeol(fp);
	
			/* Store offset of report header in file */
			if (mode & 2) rptr->fpos = rstfpos - 1;

			/* Or store offset of report body in file */
			else if (-1L == (rptr->fpos = ftell(fp))) erf(ERF_ftellFailed, 11);

			/* Don't load report text? remember is skip over it, if so. */
			if (!(rptr->flags & RF_LoadText)) rptr = 0;

			/* Load text into memory */
			else
			{
				/* Check free space in buffer */
				if ((mend - mem) <= (rlen + 1)) erf(ERF_BufferExhaust, 12);

				/* Load piecewise */
				for (rptr->text = mem; (rlen > 0) && (!ferror(fp)) && (!feof(fp)); mem += dec, rlen -= dec)
				{
					dec = fread(mem, 1, (size_t)(rlen), fp);
				}
				*mem++ = '\0';

				/* Trap errors */
				if (rlen != 0)
				{
					fprintf(stderr, "%li:   repfile fread on \"%s\" failed at %li of %li: dec=%li ferror=%i feof=%i.\n",
							uniqid, rid, rlen, rptr->length, dec, ferror(fp), feof(fp));
					erf(ERF_freadFailed, 13);
				}
			}

			/* Don't everwrite memory */
			++n; *memst = mem;
		}

		/* Seek past body */
		if ((!rptr) && (fseek(fp, rlen, SEEK_CUR))) erf(ERF_fseekFailed, 14);

		/* Optimisation: early return if no more 'interesting' reports left */
		if ((rp_check) && (!rp_head)) break;
	}

	/* Report */
	fprintf(stderr, "%li:   repfile(%p): %i reports loaded, mode %i.\n", uniqid, rp_check, n, mode);

	/* Some reports missing from file */
	if ((rp_check) && (rp_head))
	{
		for (rptr = rp_head; (rptr); rptr = rptr->link_sub)
		{
			fprintf(stderr, "%li:   --XXX--> report \"%s\" not found.\n", uniqid, rptr->rid);
		}
		erf(ERF_UnknownRepIdx, 31);
	}

	/* Don't close on success? */
	if (freps) *freps = fp;

	/* Tidy up if not to be left open */
	else fclose(fp);

	/* Return head of report list? */
	if (!rp_check) *rp_readp = rp_head;

	/* Success */
	return 0;


_erf: /***** Error Handling *****/

	/* Tidy up */
	if (fp) fclose(fp);

	/* Debug */
	if (rc) fprintf(stderr, "%li:   repfile(%i):%i %i-th \"%s\" mode %i\n",
					uniqid, ERF_CONTEXT(rc), ERF_TYPE(rc), n, STR(rid), mode);

	/* Only retain low bits of normal rc's */
	return (ERF_LOWBITS(rc));
}

#endif
/*** The End ***/
