/*****************************************************
**  lib-uinfo-edit.c                     (library)  **
**  ----------------------------------------------  **
**  Library: code for editing the contents of the   **
**  "uinfo.txt" file.                               **
**                                                  **
************** BORING LEGAL NOTICE *******************
**                                                  **
**  Assassins' Game (Semi-)Automatic Umpire System  **
**  ----------------------------------------------  **
**  Prerelease: Copyright 2000-2002 Edward M Nokes  **
**  This code may not be distributed in source or   **
**  executable form without the express consent of  **
**  the author, Edward M Nokes (emn23@cam.ac.uk).   **
**  The author accepts no responsibility or         **
**  liability for anything occurring as a result    **
**  of using this software, etc. etc.               **
**                                                  **
*****************************************************/

#ifdef included__lib_uinfo_edit
 #error "** lib-uinfo-edit multiply #included **"
#else
 #define included__lib_uinfo_edit fromfile

/*#define DEB(x) x*/
#define DEB(x) 0

/***************************************************************************************************************************
** Declarations and definitions                                                                                           **
***************************************************************************************************************************/

/* Type to hold options */
typedef unsigned int eduf_flags;

/* Option flags */
#define EDUF_UpdateTargets		((eduf_flags)(1U<<0))	/* Automatically update "last seen" targets from new ones */
#define EDUF_ReturnUnwritten	((eduf_flags)(1U<<1))	/* Pass back the head of the list of unwritten changes */
#define EDUF_PartialFix			((eduf_flags)(1U<<2))	/* Respect the PF_PartialFix flag */
#define EDUF_FullFix			((eduf_flags)(1U<<2))	/* Respect the PF_FullFix flag */
#define EDUF_Repair				((eduf_flags)(1U<<8))	/* Attempt to repair old file formats */

/* Handy type for editing users */
typedef struct editpl* editpl_p;

/* A structure for editing users */
struct editpl
{
	target_p nasses;
	target_p ntargs;
	target_p noldts;
	unsigned long nfl_mask, nfl_bits;
	char *opid;
	char *nname;
	char *npass;
	char *naddr;
	char *ncoll;
	char *nmafia;
	char *nemail;
	char *nwant;
	char *npseud;
	long *nscore, *ndanger, *nrank;
	int *nkills, *ndeaths;
	timmy *ntimp;
	editpl_p link, next;
};


/***************************************************************************************************************************
** Helper functions
***************************************************************************************************************************/
int edituinfo_wrids(target_p tarp, char* p, char c_out, char c_in, FILE* fout)
{
	if (tarp)
	{
		while (tarp)
		{
			if (tarp->id)
			{
				if (3 > fprintf(fout, "\t%c%s", c_out, tarp->id)) return ERF_printfFailed;
			}
			tarp = tarp->next;
		}
	}

	else
	{
		while (*p)
		{
			if (p[0] == c_in)
			{
				if (3 > fprintf(fout, "\t%c%s", c_out, p+1)) return ERF_printfFailed;
			}
			p += 1 + strlen(p);
		}
	}

	return 0;
}

/***************************************************************************************************************************
** Core routine: scan throught the "uinfo.txt" file and update it according to the linked list of 'editpl' structs passed **
** with head (list) and links via 'next' fields. The 'next' fields are overwritten by this function. Guaranteed either    **
** to sucessfully complete all writes, or return an error and commit no changes. Do not pass lists with duplicates.       **
***************************************************************************************************************************/
int edituinfo(eduf_flags flags, editpl_p *list, editpl_p otheredp, char* bufst, char* bufend, FILE* fred, char* comnt)
{
	int c, i, j, rc = 0;
	editpl_p head, pedp, *oedpp;
	FILE *fin = 0, *fout = 0;
	char *bp, *p_tmp[6], *p_str[6], *p_ids, *p_wanted, *p_pseud, *pid = 0, *p;
	long lmin = 0L, lmax = 0L;
	unsigned long pfl, pflm;
	target_p tarp;
	timmy *nt;


	/* Check parameters for valid list of changes */
	if (!list) erf(ERF_InternalError, 1);

	/* Ensure at least bytes free in buffer */
	if ((bufend - bufst) < chk[CHK_uinfo].len) erf(ERF_BufferExhaust, 2);

	/* Initialise list of players "left to check" */
	head = *list;

	/* Debug */
	DEB(for (pedp = head; (pedp); pedp = pedp->next) fprintf(stderr, "(%s)", pedp->opid));

	/* Open current player info file */
	if (!(fin = fopen(FILE_UserInfo, "rb"))) erf(ERF_fopenUinfo, 3);

	/* No redirection? */
	if (!(fout = fred))
	{
		/* Open new player info file */
		if (!(fout = fopen(FILE_NewUserInfo, "wb"))) erf(ERF_fopenUinfo, 4);

		/* Add comment prior to header? */
		if (comnt)
		{
			for (i = 0; (comnt[i] >= 32); i++);
			comnt[i] = '\0'; i += 48;
			if (i != fprintf(fout, "###\n### (edituinfo) %s\n"
								   "##########################\n", comnt)) erf(ERF_printfFailed, 5);
		}
	}

	/* Check file header (ignoring initial comments) */
	if (i = chkheader(fin, CHK_uinfo, bufst, (flags & EDUF_Repair) ? &j : 0, &lmin, 0, 0)) erf(i, 6);

	/* Error on incorrect/missing header */
	if ((flags & EDUF_Repair) && (j != CHK_VER))
	{
		/* No header, go back to start (otherwise, lmin is current pos, just after header) */
		if ((j == CHK_NoHeader) && (fseek(fin, lmin, SEEK_SET))) erf(ERF_fseekFailed, 100);

		fprintf(stderr, "Moo! Bad header %i.\n", j);
	}

	/* Write current file header */
	if (i = prtheader(fout, CHK_uinfo, 0)) erf(i, 7);

	/* Scan through for valid lines */
	while ((c = fgetc(fin)) != EOF)
	{
		/* Ignore lines not beginning with "*" character */
		switch (c)
		{
			case '*'  : break;
			default   : toeol(fin);
			case '\n' : continue;
		}
		
		/* Reuse buffer */
		bp = bufst;

		/* Skip initial tabs, then read 'id' up to first tab or newline */
		if (i = tread(fin, pid = bp, bufend, &bp)) erf(ERF_BadTread+i, 8);

		/* Debug */
		DEB(fprintf(stderr, "%li:   edituinfo found player \"%s\"\n", uniqid, pid));

		/* Is this player on the list of changes? */
		for (pedp = *(oedpp = &head); (pedp); pedp = *(oedpp = &pedp->next)) if (streq(pedp->opid, pid)) break;

		/* No, but perhaps an edit struct was passed for those not matched on list */
		if ((!pedp) && (pedp = otheredp)) oedpp = 0;

		/* Matched */
		if (pedp)
		{
			/* Skip past any extra initial tabs to preserve column alignment */
			if (i = tskip(fin)) erf(ERF_BadTread+i, 9);

			/* Make note of position (after id field) */
			if ((lmax = ftell(fin)) == -1L) erf(ERF_ftellFailed, 10);

			/* Debug */
			DEB(fprintf(stderr, "%li:   edituinfo match %s\"%s\" ftofcpy[%li,%li)\n",
						uniqid, (oedpp) ? "" : "other ", pid, lmin, lmax));

			/* Copy bytes [lmin,lmax) from fin to fout, in a paranoid manner */
			if (i = ftofcpy(fin, fout, lmax - lmin, fpbuf, MAX_fpbuf, lmin, lmax)) erf(i, 11);

			/* Get Password, Name, College, Mafia, Address, Email fields (12-18) */
			for (j = 0; (j < 6); j++) if (i = tread(fin, p_str[j] = bp, bufend, &bp)) erf(ERF_BadTread+i, 12+j);

			/* Read flags field */
			if (i = tread(fin, bp, bufend, 0)) erf(ERF_BadTread+i, 22);

			/* Convert flags and check PF_FullFix flag */
			if (((pfl = strtoul(bp, 0, 16)) & PF_FullFix) && (flags & EDUF_FullFix))
			{
				DEB(fprintf(stderr, "%li:   edituinfo fullfix: ignoring \"%s\"\n", uniqid, pid));
				lmin = lmax;
				goto tonext;
			}

			/* Read restime/kills/deaths/score/rank/dangerousness fields as strings - don't convert data (23-28) */
			for (j = 0; (j < 6); j++) if (i = tread(fin, p_tmp[j] = bp, bufend, &bp)) erf(ERF_BadTread+i, 23+j);

			/* String which will contain all assassin/target/oldtarg ids */
			p_ids = bp;

			/* Read assassins/targets/previous-targets (30-32) */
			for (j = 0; (1); j++)
			{
				if (i = tread(fin, p = bp, bufend, &bp)) erf(ERF_BadTread+i, 30);
				if (p[0] == '-') break;
				/*fprintf(stderr, "%li:   footarg\"%s\"\n", uniqid, p);*/
				if ((p[0] != 'A') && (p[0] != 'T') && (p[0] != 'O')) erf(ERF_InternalError, 31);
				if (p[0] <= 32) erf(ERF_InternalError, 32);
			}

			/* Terminate list */
			*(bp++) = '\0';

			/* Read "wantedness" field */
			if (i = tread(fin, p_wanted = bp, bufend, &bp)) erf(ERF_BadTread+i, 40);

			/* Read pseudonym field */
			if ((i = tread(fin, p_pseud = bp, bufend, &bp)) > 1) erf(ERF_BadTread+i, 41);

			/* Make new note of position at end of entry */
			if ((lmin = ftell(fin)) == -1L) erf(ERF_ftellFailed, 42);

			/* Recover final tab if more text remaining at end of line */
			if (i == 0) --lmin;

			/* Debug: finished reading data */
			DEB(fprintf(stderr, "%li:   edituinfo read \"%s\" flags=&%lX new_lmin=%li\n", uniqid, pid, pfl, lmin));

			/* New password? */
			if ((pedp->npass) && (*(p_str[0] = pedp->npass) < 32)) erf(ERF_BadUserData, 44);

			/* New name */
			if ((pedp->nname) && (*(p_str[1] = pedp->nname) < 32)) erf(ERF_BadUserData, 45);

			/* New college? */
			if ((pedp->ncoll) && (*(p_str[2] = pedp->ncoll) < 32)) erf(ERF_BadUserData, 46);

			/* New mafia? */
			if ((pedp->nmafia) && (*(p_str[3] = pedp->nmafia) < 32)) erf(ERF_BadUserData, 47);

			/* New address? */
			if ((pedp->naddr) && (*(p_str[4] = pedp->naddr) < 32)) erf(ERF_BadUserData, 48);

			/* New email? */
			if ((pedp->nemail) && (*(p_str[5] = pedp->nemail) < 32)) erf(ERF_BadUserData, 49);

			/* Prevent certain flags from being altered */
			pflm = (((pfl & PF_PartialFix) && (flags & EDUF_PartialFix)) ? PF_MASK_Fixed : 0) | PF_MASK_NoEdit;

			/* Adjust flags */
			pfl = (pfl & ((~pedp->nfl_mask) | pflm)) ^ (pedp->nfl_bits & (~pflm));

			/* Debug */
			DEB(fprintf(stderr, "%li:   edituinfo \"%s\" pflm=%lx new pfl=%lx\n", uniqid, pid, pflm, pfl));

			/* Check new assassins/targets/oldtargs? */
			for (tarp = pedp->nasses; (tarp); tarp = tarp->next)
			{
				if ((tarp->id) && (tarp->id[0] < 32)) erf(ERF_BadUserData, 50);
			}

			/* Check new targets */
			for (tarp = pedp->ntargs; (tarp); tarp = tarp->next)
			{
				if ((tarp->id) && (tarp->id[0] < 32)) erf(ERF_BadUserData, 51);
			}

			/* Check new oldtargs? */
			for (tarp = pedp->noldts; (tarp); tarp = tarp->next)
			{
				if ((tarp->id) && (tarp->id[0] < 32)) erf(ERF_BadUserData, 52);
			}

			/* Debug */
			DEB(fprintf(stderr, "%li:   edituinfo \"%s\" wibble32\n", uniqid, pid));			

			/* New "wantedness"? */
			if ((pedp->nwant) && (*(p_wanted = pedp->nwant) < 32)) erf(ERF_BadUserData, 60);

			/* New pseudonym? */
			if ((pedp->npseud) && (*(p_pseud = pedp->npseud) < 32)) erf(ERF_BadUserData, 61);

			/* Debug */
			DEB(fprintf(stderr, "%li:   edituinfo \"%s\" wibble64\n", uniqid, pid));			

			/* Write (edited) Password, Name, College, Mafia, Address, Email, Flags fields */
			if (14 > fprintf(fout, "%s\t%s\t%s\t%s\t%s\t%s\t%lX", p_str[0],
							 p_str[1], p_str[2], p_str[3], p_str[4], p_str[5], pfl)) erf(ERF_printfFailed, 62);

			/* If timer flags have been partially fixed for this player, protect old timer value */
			nt = (pflm & PF_MASK_Timers) ? 0 : pedp->ntimp;

			/* New timer - check validity */
			if ((nt) && (9 != fprintf(fout, "\t%02d:%02d:%02d", nt->day, nt->hour, nt->min))) erf(ERF_printfFailed, 63);

			/* Old timer - assume valid if not too short */
			if ((!nt) && (9 > fprintf(fout, "\t%s", p_tmp[0]))) erf(ERF_printfFailed, 64);

			/* New kills */
			if ((pedp->nkills) && (2 > fprintf(fout, "\t%i", *pedp->nkills))) erf(ERF_printfFailed, 65);

			/* Old kills - assume valid if not too short */
			if ((!pedp->nkills) && (2 > fprintf(fout, "\t%s", p_tmp[1]))) erf(ERF_printfFailed, 66);

			/* New deaths */
			if ((pedp->ndeaths) && (2 > fprintf(fout, "\t%i", *pedp->ndeaths))) erf(ERF_printfFailed, 67);

			/* Old deaths - assume valid if not too short */
			if ((!pedp->ndeaths) && (2 > fprintf(fout, "\t%s", p_tmp[2]))) erf(ERF_printfFailed, 68);

			/* New score */
			if ((pedp->nscore) && (2 > fprintf(fout, "\t%li", *pedp->nscore))) erf(ERF_printfFailed, 69);

			/* Old score - assume valid if not too short */
			if ((!pedp->nscore) && (2 > fprintf(fout, "\t%s", p_tmp[3]))) erf(ERF_printfFailed, 70);

			/* New rank */
			if ((pedp->nrank) && (2 > fprintf(fout, "\t%li", *pedp->nrank))) erf(ERF_printfFailed, 78);

			/* Old rank - assume valid if not too short */
			if ((!pedp->nrank) && (2 > fprintf(fout, "\t%s", p_tmp[4]))) erf(ERF_printfFailed, 79);

			/* New dangerousness */
			if ((pedp->ndanger) && (2 > fprintf(fout, "\t%li", *pedp->ndanger))) erf(ERF_printfFailed, 71);

			/* Old dangerousness - assume valid if not too short */
			if ((!pedp->ndanger) && (2 > fprintf(fout, "\t%s", p_tmp[5]))) erf(ERF_printfFailed, 72);

			/* Write existing or new assassins */
			if (i = edituinfo_wrids(pedp->nasses, p_ids, 'A', 'A', fout)) erf(i, 73);

			/* Write existing or new targets */
			if (i = edituinfo_wrids(pedp->ntargs, p_ids, 'T', 'T', fout)) erf(i, 74);

			/* Automatically update "last mailed" targets from current set */
			j = (flags & EDUF_UpdateTargets) ? 'T' : 'O';

			/* Write new or existing or recent old targets */
			if (i = edituinfo_wrids(pedp->noldts, p_ids, 'O', j, fout)) erf(i, 75);

			/* Terminate id list, and write Wantedness, Pseudonym fields */
			if (6 > fprintf(fout, "\t-\t%s\t%s", p_wanted, p_pseud)) erf(ERF_printfFailed, 77);

			/* Debug. NB: Remainder and termination of line are dealt with by subsequent block copies */
			DEB(fprintf(stderr, "%li:   edituinfo wrote %s new_flags=&%lX\n", uniqid, pid, pfl));
		}

tonext:	/* Slight optimisation: remove from list of structs to check */
		if ((pedp) && (oedpp)) { *oedpp = pedp->next; pedp->next = 0; }

		/* Optimisation: stop searching if no more changes to make */
		if ((!head) && (!otheredp)) break;

		/* Next please */
		toeol(fin);

		/* Forget */
		pid = 0;
	}

	/* Copy remaining chunk of fin to fout, in a paranoid manner */
	if (i = ftofcpy(fin, fout, -1L, fpbuf, MAX_fpbuf, lmin, -1L)) erf(i, 81);

	/* Close input file */
	fclose(fin); fin = 0;

	/* Return head of list of unwritten change structs? */
	if (flags & EDUF_ReturnUnwritten) *list = head;

	/* Any unwritten changes? */
	if (head)
	{
		/* Debug: note the offenders */
		for (pedp = head, i = 1; (pedp); pedp = pedp->next, i++)
		{
			fprintf(stderr, "%li:   XXX unwritten edit %i on \"%s\"\n", uniqid, i, pedp->opid);
		}

		/* Complain */
		erf(ERF_IncompleteEdit, 82);
	}

	/* No redirection */
	if (!fred)
	{
		/* Close output file if not redirected somewhere */
		if (fout) { fclose(fout); fout = 0; }

		/* Remove very old uinfo file */
		remove(FILE_OldUserInfo);

		/* Use output file as new uinfo file, rename old file away */
		if (rename(FILE_UserInfo, FILE_OldUserInfo) || rename(FILE_NewUserInfo, FILE_UserInfo)) erf(ERF_FSwapFailed, 83);
    }
   

_erf: /***** Error handling *****/

	/* Tidy up */
	if (fin) fclose(fin);
	if ((fout) && (!fred)) fclose(fout);

	/* No error, or reserved error */
	if (rc <= 0) return (rc);

	/* Debug */
	fprintf(stderr, "%li !! edituinfo(%i):%i pid=\"%s\"\n", uniqid, ERF_CONTEXT(rc), ERF_TYPE(rc), STR(pid));

	/* Only retain low bits of normal rc's */
	return (ERF_LOWBITS(rc));
}

#endif
/**** The End ****/
