/*****************************************************
**  lib-retarg.c                         (library)  **
**  ----------------------------------------------  **
**  Retargetting algorithm.                         **
**                                                  **
************** BORING LEGAL NOTICE *******************
**                                                  **
**  Assassins' Game (Semi-)Automatic Umpire System  **
**  ----------------------------------------------  **
**  Prerelease: Copyright 2000-2002 Edward M Nokes  **
**  This code may not be distributed in source or   **
**  executable form without the express consent of  **
**  the author, Edward M Nokes (emn23@cam.ac.uk).   **
**  The author accepts no responsibility or         **
**  liability for anything occurring as a result    **
**  of using this software, etc. etc.               **
**  ----------------------------------------------  **
*****************************************************/

#ifdef included__lib_retarg
 #error "** lib-retarg multiply #included **"
#else
 #define included__lib_retarg fromfile

/***************************************************************************************************************************
** Declarations and definitions                                                                                           **
***************************************************************************************************************************/

/* Targetting constants */
#define TARGS_RECURSE_MAX	100		/* Limit on depth of recursions */
#define TARGS_ITER_MAX		8000	/* Maximum number of iterations */
#define TARGS_MIN			3		/* Number of targets to attempt to allocate */
#define TARGS_MAX_PL		300
#define TARGS_MAX_PER_PL	5

/* Targetting flags */
#define DOTARG_Strict	(1U<<0)		/* Don't allocate same player as target and assassin */
#define DOTARG_Random	(1U<<1)		/* Randomly reorder list prior to reallocation */

/***************************************************************************************************************************
** Updated targetting algorithm                                                                                           **
***************************************************************************************************************************/
#define IFDEB(x)  /*do (x); while (0) */

int dotargetting(player_p *pl_in, int pn_in, unsigned int state, retarg_p *outrtp, int **cnts)
{
	static player_p *pl;
	static retarg_p retarps, bestretarps;
	static int pn, iter, goodness, best, *ntar, *nass, freetars, recurse;
	static int fairness[TARGS_MIN], bestfair[TARGS_MIN];
	player_p pptr, victpptr, nextpptr;
	target_p xp, newxp;
	retarg_p retp;
	int i, j, this, vict;

	/* Initialisation */
	if (pn_in >= 0)
	{
		/* Player list */
		pl = pl_in;
		pn = pn_in;

		/* Clear results */
		if (outrtp) *outrtp = 0;

		/* No new target assignments yet */
		bestretarps = retarps = 0;

		/* Strength of assignments */
		goodness = best = 0;

		/* Randomise results? */
		if (state & DOTARG_Random)
		{
			/* Move each player to a random position */
			for (i = 0; (i < pn); i++)
			{
				/* Random position 0 <= j < pn */
				j = myrand(pn);

				/* Swap */
				pptr = pl[i];
				pl[i] = pl[j];
				pl[j] = pptr;
			}
		}

		/* Allocate block of counts */
		if (!(ntar = malloc(pn * 2 * sizeof(int)))) return -10;

		/* Block of assassin counts */
		nass = &ntar[pn];

		/* Debug */
		IFDEB(printf("<h4>Debug:</h4><p>Init on pn=%i players.<br>\n", pn));

		/* No iterations yet */
		iter = 0; recurse = 0;

		/* Fairness of target allocation: initialise */
		for (i = 0; (i < TARGS_MIN); i++) fairness[i] = 0;

		/* Initialise each player */
		for (i = freetars = 0; (i < pn); i++)
		{
			/* Count existing targets */
			for (ntar[i] = 0, xp = pl[i]->targs; (xp); ntar[i]++, xp = xp->next) /*xp->ext = 0*/;

			/* Count existing assassins */
			for (nass[i] = 0, xp = pl[i]->asses; (xp); nass[i]++, xp = xp->next) /*xp->ext = 0*/;

			/* Note first player with too few targets */
			if (ntar[i] < TARGS_MIN)
			{
				freetars += TARGS_MIN - ntar[i];
				++fairness[ntar[i]];
			}
		}

		/* Fairness of best target allocation: initialise */
		for (i = 0; (i < TARGS_MIN); i++) bestfair[i] = fairness[i];

		/* No targets needed */
		if (freetars == 0) return -1;
	}

	/* Too complicated... give up */
	if (++iter > TARGS_ITER_MAX) return -2;

	/* Loop through players with too few targets */
	for (pptr = 0, i = this = 0; (this < pn); this++)
	{
		IFDEB(printf("(needtarfor:%i:\"%s\":%i,%i)", this, pl[this]->id, ntar[this], nass[this]));

		/* Already fully allocated */
		if (ntar[this] >= TARGS_MIN) continue;

		/* Player requiring a target */
		pptr = pl[this];

		/* Debug */
		IFDEB(printf("<br>Find targ for \"%s\" (has %i targs, %i asses):<ul>\n", pptr->id, ntar[this], nass[this]));
	
		/* Search for a suitable victim */
		for (victpptr = 0, i = vict = 0; (vict < pn); vict++)
		{
			IFDEB(printf("(tryvict:%i:\"%s\":%i,%i)", vict, pl[vict]->id, ntar[vict], nass[vict]));

			/* No self-allocation */
			if ((victpptr = pl[vict]) == pptr) continue;
	
			/* Limit number of assassins per victim */
			if (nass[vict] >= TARGS_MIN) continue;
	
			/* Check if victim is already target of player  */
			for (j = 0, xp = victpptr->asses; (xp); xp = xp->next, j++)
			{
				if (j >= nass[vict]) printf("<b>AssListErr:%i_of_%i!</b>", j, nass[vict]);
				if (streq(xp->id, pptr->id)) break;
			}
	
			IFDEB(printf("(foo)"));

			/* Already linked as target - go to next possible victim */
			if (xp)
			{
				IFDEB(printf("<li>\"%s\" free (%d asses), but is already target.</li>\n", victpptr->id, nass[vict]));
				continue;
			}
	
			/* Check if victim is already assassin of player  */
			for (j = 0, xp = victpptr->targs; (xp); xp = xp->next, j++)
			{
				if (j >= ntar[vict]) printf("<b>TarListErr:%i_of_%i!</b>", j, ntar[vict]);
				if (streq(xp->id, pptr->id)) break;
			}
	
			/* Already linked as assassin? */
			if (xp) IFDEB(printf("<li>\"%s\" free (%d asses); is already assassin.</li>\n", victpptr->id, nass[vict]));
	
			/* If so, and in strict mode, go to next possible victim */
			if ((xp) && (state & DOTARG_Strict)) continue;
	
			/* Debug */
			IFDEB(printf("<li>\"%s\" valid (%d asses) as target for \"%s\".</li>\n", victpptr->id, nass[vict], pptr->id));
	
			/* Make assignment: claim structure */
			if (!(retp = malloc(sizeof(struct retarg)))) return -14;
	
			/* Assassin and target */
			retp->assid = pptr->id;
			retp->tarid = victpptr->id;
	
			/* Link into list */
			retp->next = retarps;
			retarps = retp;
	
			/* Update lists: claim structures */
			if (!(newxp = malloc(2 * sizeof(struct target)))) return -15;
	
			/* New target for player */
			newxp->id = victpptr->id;
			newxp->next = pptr->targs;
			pptr->targs = newxp;
	
			/* New assassin for victim */
			(newxp+1)->id = pptr->id;
			(newxp+1)->next = victpptr->asses;
			victpptr->asses = (newxp+1);
	
			/* Remeve old entry from fairness array */
			if (ntar[this] < TARGS_MIN) --fairness[ntar[this]];

			/* One more assignment */
			++ntar[this];
			++nass[vict];

			/* Update fairness array with new entry */
			if (ntar[this] < TARGS_MIN) ++fairness[ntar[this]];

			IFDEB(printf("[+|retarps|%p]", retarps));
	
			/* Same number of allocations: test for "fairer" allocation */
			if (++goodness == best)
			{
				/* Iterate whilst the same */
				for (i = 0; (i < TARGS_MIN); i++) if (fairness[i] != bestfair[i])
				{
					/* Either better, or worse */
					i = (fairness[i] > bestfair[i]); break;
				}

				IFDEB(fprintf(stderr, "%li:   [fairer? goodness=%i new %i,%i,%i best %i,%i,%i result=%i]\n",
							  uniqid, goodness, fairness[0], fairness[1], fairness[2],
							  bestfair[0], bestfair[1], bestfair[2], i));
			}

			/* If (i) is zero, then "fairer" */
			else i = 1;

			/* Is this the best arrangement so far? Record if so. */
			if ((goodness > best) || (i == 0))
			{
				retarg_p fromrtp, tortp;
	
				/* Debug */
				IFDEB(printf("<li><b>New best arrangement:</b> %i, old %i.</li>\n", goodness, best));

				/* Free old structure block */
				if (bestretarps) free(bestretarps);
	
				/* Allocate structure block */
				if (!(bestretarps = malloc(goodness * sizeof(struct retarg)))) return -16;
	
				/* Initialise structure */
				for (tortp = bestretarps, fromrtp = retarps, i = 0; (fromrtp); i++, fromrtp = fromrtp->next)
				{
					/* Paranoia */
					if (i >= goodness) return -17;
	
					/* Note assignments */
					tortp->assid = fromrtp->assid;
					tortp->tarid = fromrtp->tarid;
	
					/* Link up list */
					tortp = tortp->next = tortp + 1;
				}
		
				/* Paranoia */
				if (i != goodness) return -19;

				/* Terminate list */
				(tortp - 1)->next = 0;

				/* Strenght of best arrangement so far */
				best = goodness;

				/* Fairness of best arrangement so far */
				for (i = 0; (i < TARGS_MIN); i++) bestfair[i] = fairness[i];
			}
	
			/* Limit recursion depth */
			if ((++recurse) > TARGS_RECURSE_MAX) return -3;

			/* Recurse */
			if ((i = dotargetting(0, -1, state, 0, 0)) < 0) return (i);

			/* Recursion reduced */
			--recurse;

			IFDEB(printf("[-|retarps|%p]", retarps));

			/* Remeve new entry from fairness array */
			if (ntar[this] < TARGS_MIN) --fairness[ntar[this]];

			/* One fewer assignment */
			--goodness;
			--ntar[this];
			--nass[vict];

			/* Reinstate old entry in fairness array */
			if (ntar[this] < TARGS_MIN) ++fairness[ntar[this]];
	
			/* Paranoia */
			if ((pptr->targs != newxp) || (victpptr->asses != newxp + 1)) return -20;

			/* Remove target, assassin */
			pptr->targs = pptr->targs->next;
			victpptr->asses = victpptr->asses->next;
	
			/* Free target structures */
			free(newxp);
	
			/* Remove assignment */
			retarps = retp->next;
	
			/* Free assignment structure */
			free(retp);

			/* End of recursion? (early termination) */
			if (i) break;
		}

		/* Debug */
		IFDEB(puts("</ul>\n"));

		/* End of recursion? (early termination) */
		if (i) break;
	}

	/* Currently recursed - return true if no more targets need be allocated */
	if (pn_in < 0) return ((i) || (pptr == 0));

	/* Return results */
	if (outrtp) *outrtp = bestretarps;

	/* Preserve and return workspace */
	if (cnts) *cnts = ntar;

	/* Deallocate workspace */
	else free(ntar);

	/* Finish off debugging */
	IFDEB(printf("<p>End of debug: best=%i retarps=%p</p>", best, retarps));

	fprintf(stderr, "%li:   Targetting done.\n", uniqid);

	/* Paranoia: everything should have been undone */
	if (retarps) return -18;

	/* Return number of new allocations in best arrangement */
	return (best);
}

#undef IFDEB

/***************************************************************************************************************************
** Print retargetting summary and data as HTML                                                                            **
***************************************************************************************************************************/
int prtretarg(player_p *pl, int pn, FILE* fp, int m, retarg_p untarps, unsigned int dorand, int* retn)
{
	int n, i, j, k, l, r, *cnts = 0, rc = 0;
	char *p, *q, *p1id;
	retarg_p rtp, *prtp, retarps = 0;
	player_p p1ptr, p2ptr;


	/* Retain names */
	if ((dorand & 2) && (pl) && (pn > 0))
	{
		for (i = j = 0; (i < pn); i++)
		{
			if (pl[i]) fprintf(fp, "<input type=\"hidden\" name=\"ol_%did\" value=\"%s\">", 1+i, pl[i]->id);
		}
	}

	/* Run the algorithm */
	n = dotargetting(pl, pn, DOTARG_Strict | ((dorand & 1) ? DOTARG_Random : 0), &retarps, &cnts);

	/* Tidyness */
	p = (m & 1) ? "<tr><td colspan=2>" : "";

	/* Return the number of possible things */
	if (retn) *retn = n;

	/* Error */
	if (n < -3) erf(ERF_Targetting, -n);

	/* Too complicated */
	else if (n < -1) fprintf(fp, "%s<h4>Targetting failed: too complicated.</h4>", p);

	/* No targetting necessary */
	else if (n == -1) fprintf(fp, LF "%s<h4>No additional targetting required.</h4>"
								  LF "<p>None of the relevent players are short of targets</p>", p);

	/* No targetting possible */
	else if (n == 0) fprintf(fp, LF "%s<h4>No additional targetting is possible.</h4>"
								 LF "<p>No good allocations between the relevent players could be found.</p>", p);

	/* Targetting successful */
	else
	{
		/* Appropriate message */
		if (m & 1) fprintf(fp, LF "<tr><td><input type=\"radio\"%sname=\"ev_retarg\" value=\"retarg\">&nbsp;</td>"
							   LF "<td>Deallocate victim's previous targets and assassins (%i player%s)<br>"
							   LF "Make %d assassin &mdash;&gt; target allocation%s from them thusly:</td></tr>",
							   (m & 2) ? " checked " : "", pn, (pn > 1) ? "s" : "", n, (n > 1) ? "s" : "");

		/* Title */
		else fprintf(fp, LF "<h4>Target allocation amongst %i players:</h4>"
						 LF "<p>Can make %d assassin &mdash;&gt; target allocation%s from them as follows:</p>"
						 LF "<p><table cellpadding=0 cellspacing=2 border=0 cols=2>",
						 pn, n, (n > 1) ? "s" : "");

		/* Count number of assassins per player */
		for (rtp = retarps; (rtp); rtp = rtp->next)
		{
			/* Match assassin with player */
			for (i = 0; (i < pn); i++) if (streq(pl[i]->id, rtp->tarid)) break;

			/* Increment player count of assassins */
			if (i < pn) ++cnts[pn + i];
		}

		/* Group by player */
		for (i = r = l = 0; (i < pn) || (retarps) || (untarps); i++)
		{
			/* Common assassin */
			if (i < pn) p1id = (p1ptr = pl[i])->id;

			/* Leftovers */
			else { p1ptr = 0; p1id = (retarps) ? retarps->assid : untarps->assid; }

			/* Write out the retargettings */
			for (k = 0, rtp = *(prtp = &retarps); (rtp); rtp = rtp->next)
			{
				/* Only ones which concern the current player */
				if (!streq(p1id, rtp->assid))
				{
					/* Retain others in list */
					prtp = &rtp->next; continue;
				}

				/* Lookup target involved */
				for (j = 0; (j < pn); j++) if (streq(pl[j]->id, rtp->tarid)) break;
	
				/* Found? */
				p2ptr = (j < pn) ? pl[j] : 0;
	
				/* HTML output */
				fprintf(fp, LF "<input type=\"hidden\" name=\"pl_%dt%d\" value=\"%s\">"
							LF "<tr><td>(%i)</td><td><b><a href=\"plinfo.cgi?id=%s\">%s</a></b> could target"
							LF "<b><a href=\"plinfo.cgi?id=%s\">%s</a>.</b></td></tr>",
							l+1, ++k, rtp->tarid, ++r,
							p1id, (p1ptr) ? p1ptr->name : "[unknown]",
							rtp->tarid, (p2ptr) ? p2ptr->name : "[unknown]");
	
				/* Delink from chain */
				*prtp = rtp->next;
			}

			/* Write out the untargettings */
			for (j = 0, rtp = *(prtp = &untarps); (rtp); rtp = rtp->next)
			{
				/* Only ones which concern the current player */
				if (!streq(p1id, rtp->assid))
				{
					/* Retain others in list */
					prtp = &rtp->next; continue;
				}				

				/* HTML output */
				fprintf(fp, LF "<input type=\"hidden\" name=\"pl_%du%d\" value=\"%s\">", l+1, ++j, rtp->tarid);
	
				/* Delink from chain */
				*prtp = rtp->next;
			}

			/* Count players with new targets */
			if ((k) || (j)) fprintf(fp, LF "<input type=\"hidden\" name=\"pl_%did\" value=\"%s\">", ++l, p1id);

			/* Right number of targets for player? */
			if (p1ptr)
			{
				p = (k + cnts[i] == TARGS_MIN) ? 0 : (k + cnts[i] < TARGS_MIN) ? "few" : "many";
				q = (cnts[pn + i] == TARGS_MIN) ? 0 : (cnts[pn + i] < TARGS_MIN) ? "few" : "many";

				if ((p) || (q)) fprintf(fp, LF "<tr><td></td><td><b><a href=\"plinfo.cgi?id=%s\">%s</a></b>"
											LF "would still have ", p1id, p1ptr->name);

				if (p) fprintf(fp, LF "<b>too %s</b> (%i) targets%s", p, k + cnts[i], (q) ? " and " : "");
				if (q) fprintf(fp, LF "<b>too %s</b> (%i) assassins", q, cnts[pn + i]);

				fputs(".</td></tr>", fp);
			}
		}

		puts((m & 1) ? LF "<tr><td colspan=2>" : LF "</table></p><p>");
		fprintf(fp, "These give additional targets to %i player%s.%s", l, (l > 1) ? "s" : "", (m & 1) ? "" : "</p>");
	}

	if (dorand & 1) puts("<p><b>Targetting was randomised</b></p>");

	/* Free up memory */
	if (cnts) free(cnts);

	/* Tidy up HTML */
	if (m & 1) puts("</td></tr></table>");
	

_erf: /***** Error handling *****/

	/* Debug */
	if (rc) fprintf(stderr, "%li:   prtretarg(%i):%i on %i.\n", uniqid, ERF_CONTEXT(rc), ERF_TYPE(rc), pn);

	/* Finished */
	return (ERF_LOWBITS(rc));
}

/***************************************************************************************************************************
** Carry out retargettings (on player kill event form, or player status change form)                                      **
***************************************************************************************************************************/
int killretarg(int m, pstate_p killpstp, player_p killpptr, int sel, player_p plist)
{
	int i, j, n, rc = 0;
	target_p tarp, remtarp, *ptrp;
	player_p pptr, testpptr, pptrlist = 0, *pptrarr;
	retarg_p untargs = 0, rtp;


	/* List targets that would be deallocated */
	puts("<tr><td></td><td>One fewer assassin for:");

	/* Run through targets */
	for (n = i = j = 0, tarp = killpptr->targs; (tarp) || (++j == 1); tarp = tarp->next, i++)
	{
		/* Finished list of targets */
		if (!tarp)
		{
			/* List assassins that would be deallocated (ie. victim's targets) */
			printf("%s</td></tr>" LF "<tr><td></td><td>One fewer target for:", (i == 0) ? "No players." : "");

			/* No victim's assassins yet */
			i = 0;

			/* Run through victim's assassins */
			if (!(tarp = killpptr->asses)) break;
		}

		/* Get target player */
		for (pptr = plist; (pptr); pptr = pptr->link) if (streq(pptr->id, tarp->id)) break;

		/* Not found */
		if (!pptr) erf(ERF_InternalError, 10);

		/* Print referenced player name */
		printf(LF "<b><a href=\""CGI_URL"plinfo.cgi?id=%s\">%s</a></b>%s",
			   pptr->id, pptr->name, (tarp->next) ? "," : ".");

		/* No self references */
		if (pptr == killpptr) continue;

		/* Prepare for targetting */
		if (m)
		{
			/* Untarg struct */
			if (!(rtp = malloc(sizeof(struct retarg)))) erf(ERF_mallocFailed, 11);
	
			/* Fill in struct */
			rtp->assid = (j) ? pptr->id : killpptr->id;
			rtp->tarid = (j) ? killpptr->id : pptr->id;
	
			/* Link up struct in untarg list */
			rtp->next = untargs;
			untargs = rtp;

			/* Search for backlink */
			for (ptrp = (j) ? &pptr->targs : &pptr->asses; (remtarp = *ptrp); ptrp = &remtarp->next)
			{
				/*printf("[%s]", remtarp->id);*/

				/* Match? */
				if (streq(remtarp->id, killpptr->id)) break;
			}

			/* Remove victim from current target/assassin lists */
			if (remtarp) *ptrp = remtarp->next;

			/*else printf("(Debug:%p:%s<->%s:)", remtarp, killpptr->id, pptr->id);*/
		}

		/* Already on seen list */
		for (testpptr = pptrlist; (testpptr); testpptr = testpptr->ext) if (testpptr == pptr) break;

		/* Already seen? Don't add again. */
		if (testpptr) continue;

		/* Not so. Add to list of relevent players */
		pptr->ext = pptrlist;
		pptrlist = pptr;

		/* Count them */
		++n;
	}

	/* None found? */
	printf("%s</td></tr>", (i == 0) ? "No players." : "");

	/* Don't do actual reallocation? */
	if (!m) return 0;

	/* Message */
	printf("<tr><td></td><td>%s</td></tr>",
			(n == 0) ? "<b>Victim has no targets and assassins around which to base retargetting.</b>" : "");

	/* None found */
	if (n == 0) return 0;

	/* Array of player pointers */
	if (!(pptrarr = malloc(n * sizeof(player_p)))) erf(ERF_mallocFailed, 20);

	/* Initialise array (paranoia: array bounds) */
	for (i = 0, pptr = pptrlist; (pptr); pptr = (pptrarr[i] = pptr)->ext, i++) if (i >= n) erf(ERF_Paranoia, 21);

	/* Carry out retagetting */
	rc = prtretarg(pptrarr, n, stdout, (sel) ? 3 : 1, untargs, 0, 0);

	/* Deallocate memory */
	free(pptrarr);


_erf: /***** Error handling *****/

	/* Debug */
	if (rc) fprintf(stderr, "%li:   killretarg(%i):%i on %i.\n", uniqid, ERF_CONTEXT(rc), ERF_TYPE(rc), n);

	/* Finished */
	return (ERF_LOWBITS(rc));
}

int readretargs__aux(int n, char c, player_p pptr, retarg_p *pretargs, char* strb)
{
	retarg_p rtp;
	int j;
	char *p;


	/* Run through list */
	for (j = 1; (j <= TARGS_MAX_PER_PL); j++)
	{
		/* Form "pl_7t4" string */
		if (sprintf(strb, "pl_%d%c%d", n, c, j) < 6) return ERF_InternalError;

		/* End of list of targets */
		if (!(p = findval(strb, 0))) break;

		/* Allocate retargetting structure */
		if (!(rtp = malloc(sizeof(struct retarg)))) return ERF_mallocFailed;

		/* Assassin and target */
		rtp->assid = pptr->id;
		rtp->tarid = p;

		/* Add into list for player */
		rtp->next = *pretargs;
		*pretargs = rtp;

		fprintf(stderr, "%li:   targ: \"%s\" on \"%s\"\n.", uniqid, pptr->id, p);
	}

	/* Done */
	return 0;
}

int readretargs(retarg_p *puntargs, retarg_p *pretargs, player_p plist)
{
	int i, n = 0, rc = 0;
	char* p, strb[MAX_id+32];
	player_p pptr;


	/* Scan player ids */
	while (1)
	{
		/* Form "pl_1id" string */
		if (sprintf(strb, "pl_%did", ++n) < 6) erf(ERF_InternalError, 5);

		/* End of search */
		if (!(p = findval(strb, 0))) break;

		/* Check array bounds, and sanity of query */
		if (n >= TARGS_MAX_PL) erf(ERF_TooManyPlayers, 7);

		/* Find player */
		for (pptr = plist; (pptr); pptr = pptr->link) if (streq(p, pptr->id)) break;

		/* No match on list? */
		if (!pptr) erf(ERF_MissingPid, 9);

		/* Get targets */
		if ((pretargs) && (i = readretargs__aux(n, 't', pptr, pretargs, strb))) erf(i, 11);

		/* Get detargettings */
		if ((puntargs) && (i = readretargs__aux(n, 'u', pptr, puntargs, strb))) erf(i, 12);
	}


_erf: /***** Error handling *****/

	/* Debug */
	if (rc) fprintf(stderr, "%li:   killretarg(%i):%i on %i.\n", uniqid, ERF_CONTEXT(rc), ERF_TYPE(rc), n);

	/* Finished */
	return (ERF_LOWBITS(rc));
}

/***************************************************************************************************************************
** Old, old, old stuff
***************************************************************************************************************************/
#if 0
struct retargp { char* name; int ass[6], bass[6]; } *fidata;
int fiddletargs_uf(char* id, char* name, char* coll, char* mafia, char* addr, unsigned long p_flags, timmy* p_timer,
				   int kills, int deaths, char* email, char* ass[9], char* wanted, char* pseud, long p_score, int state)
{
	static int idscnt, pcnt, idsmax, nleft, verb;
	static char** ids;

	char **q;
	int i, j, k = 0;


	/* Initial - id = dead player id (or null), ass = array of/for player ids, kills/deaths = init/max size of array */
	if (state == -1)
	{
		/* Start with this many player ids in list */
		pcnt = idscnt = (nleft = kills) + 2;

		/* Room for all primary player ids? */
		if (idscnt > (idsmax = deaths)) return -(DOC_TooManyIds);

		/* Add null player id to list */
		(ids = ass)[0] = NON_ID;

		/* Add dead player id to list */
		if (!id) ids[1] = NON_ID;

		/* Dead player id is the null player? */
		else if (!strcmp(ids[0], (ids[1] = id))) return -(DOC_BadId);

		/* Initialise primary player entries */
		for (i = 2; (i < pcnt); i++) fidata[i].name = 0;

		/* The null player */
   		fidata[0].name = "[[Null]]";

		/* The dead player */
   		fidata[1].name = "[[Dead]]";

		verb = (p_score) ? 1 : 0;

		/* Everything ok */
		return 0;
	}

	/* Final - tidy up null or root ids */
	else if (state == -2)
	{
		/* Any primary players left on whom information is required? */
		if (nleft) return -(DOC_MissingIds);

		if (verb) printf("<br>uFunction complete.<br>");

		/* All successful */
		return 0;
	}

	/* Try to match with primary players */
	for (i = 2; (i < pcnt); i++) if (!strcmp(ids[i], id))
	{
		/* Overmatched? */
		if (fidata[i].name) return -(DOC_OverMatch);

		/* Store player name */
		fidata[i].name = name;

		if (verb) printf("<br>Matched player %i (%s) = %s and ", i, ids[i], fidata[i].name);
		
		/* Search amongst assassins/targets */
		for (j = 0; (j < 6); j++)
		{
			/* Try to match */
			for (k = 0; ((k < idscnt) && (strcmp(ids[k], ass[j]))); k++);

			/* No match - add entry */
			if (k == idscnt)
			{
				/* Out of space? */
				if (++idscnt >= idsmax) return -(DOC_TooManyIds);

				/* Store original string pointer */
				ids[k] = ass[j];
			}

			/* Link to unique match number - destroy references to the dead player */
			fidata[i].ass[j] = (k > 1) ? k : 0;

			if (verb) printf("[%i->%i:%s]", j, k, ids[k]);
		}

		/* Any left? */
		return 8+1 + ((--nleft > 0) ? (0) : (2));
	}

	/* No match - ask for next */
	return 1;
}
int fiddletargs_aux(int m, int n, char** ids)
{
	static int s, b, c = 0, mode = 0, pcnt;
	int i, j, k, l, e, f = 0;

	/* Initialisation - 'm' is the negation of the number of primaries */
	if (m < 0)
	{
		c = b = 0; pcnt = 2 - m;

		if (mode = n) printf("<br>Init with m=-%i, pcnt=%i <br>", -m, pcnt);

		/* Find the valid primarys' first unallocated target */
		for (m = 2, s = 1; (s) && (m < pcnt); m++) for (n = 3; (s) && (n < 6); n++) if (!fidata[m].ass[n]) s = 0;

		/* None such */
		if (s) return -2;

		/* Correct overshoot*/
		n--; m--;
	}

	/* Too complicated... give up */
	if (c++ > 100) return -3;

	/* Debug */
	if (mode) printf("<br>Find target %d (%s) for (%d) %s:<ul>", n, ids[fidata[m].ass[n]], m, ids[m]);

	/* Try each valid primary's unallocated assassins (to find target for (m,n)) */
	for (e = 0, i = 2; (!e) && (i < pcnt); i++) if (i != m) for (j = 0; (j < 3); j++) if (!fidata[i].ass[j])
	{
		/* Check if this primary is already linked to the given assassin  */
		for (k = 0; (k < 6); k++) if (m == fidata[i].ass[k]) break;

		/* Debug */
		if (mode) printf("<li>ass #%d free for (%d) %s%s</li>", j, i, ids[i], ((k != 6) ? " ...Unsuitable" : ""));

		/* Already linked - go to next unallocated assassin */
		if (k != 6) continue;

		/* Assign target and assassin */
		fidata[i].ass[j] = m;
		fidata[m].ass[n] = i;

		/* Is this the best arrangement so far? Record if so. */
		if (++s > b) for (b = s, k = 2; (k < pcnt); k++) for (l = 0; (l < 6); l++) fidata[k].bass[l] = fidata[k].ass[l];

		/* Find the valid primaries' next unallocated target and try to fill the gap */
		for (k = 2, f = -1; (!e) && (k < pcnt); k++) for (l = 3; (l < 6); l++) if (!fidata[k].ass[l])
		{
			/* Return early if possible */
			f = 0; if (e = fiddletargs_aux(k, l, ids)) break;
		}

		/* Undo target and assassin assignment */
		s--; fidata[m].ass[n] = fidata[i].ass[j] = 0;

		/* Early return if possible */
		if ((e) || (e = f)) break;
	}

	if (mode) puts("</ul>\n");

	/* Return number of new links in best arrangement */
	return (s) ? (e) : (b);
}
int prepretarget(player_p pl[], int plnum, int lim /* char* ids[], int max, char* deadpid */)
{
	int i, j, k, ch, n, ct, t, c = 0;
	struct retargp data[max + 2];

	/* Need some data */
	if ((max <= 0) || (lim <= max)) { puts("<p>No players for retargetting.</p>"); return 0; }

	/* Allow data access by helper functions */
	fidata = data;

	/* Initialisation - 'max' ids, with room for 'lim-max' others */
	if (i = fiddletargs_uf(deadpid, 0,0,0,0,0,0, max, lim, 0, ids, 0,0,0, -1)) return (-i);

	/* Set read primary target/assassin information and thus secondary targets/assassins */
	if (i = uread(0, fiddletargs_uf, 0, plbuf_pos, plbuf_end, &plbuf_pos)) return (i);

	/* Report unknown ids or operation on the null player */
	if (i = fiddletargs_uf(0,0,0,0,0,0,0,0,0,0,0,0,0,0,-2)) return (-i);

	/* Retargetting algorithm... unsucessful? */
	if ((n = fiddletargs_aux(-max, 0, ids)) <= 0)
	{
		if (n == -3) puts("<p>Retargetting too complicated. Increase limit...</p>");
		else if (!n) puts("<p>Retargetting not possible - all possible allocations between the relevent "
						  "players would duplicate an existing link.</p>");
		else         puts("<p>Retargetting unnecessary - none of the relevent players are short of targets.</p>");
		return 0; 
	}

	/* Otherwise, print synopsis */
	if (deadpid) printf(LF "<p><input type=\"checkbox\" name=\"ev_retarg\" value=\"set\" checked> &nbsp;"
						LF "Make %d target-assassin link%s, as follows:</p>", n, (n != 1) ? "s" : "");

	/* Start list */
	puts("<ol>");

	/* Write out each assignment */
	for (t = 0, i = 2; (i < (max + 2)); i++)
	{
		/* Any changes to targets or assassins? */
		for (j = ct = ch = 0; (j < 6); j++) if (fidata[i].ass[j] != fidata[i].bass[j])
		{
			/* Unallocated target or assassin */
			if ((k = fidata[i].bass[j]) < 1)
			{
				printf("<li><b><a href=\"plinfo.cgi?id=%s\">%s</a></b> could not be allocated a new %s (number %d).",
					   ids[i], fidata[i].name, ((j > 2) ? "target" : "assassin"), ((j > 2) ? (j - 3) : (j)));
				ct++;
			}

			/* Valid new target */
			else if (j > 2)
			{
				printf("<li><b><a href=\"plinfo.cgi?id=%s\">%s</a></b> could now target "
					   "<b><a href=\"plinfo.cgi?id=%s\">%s</a>.</b></li>",
					   ids[i], fidata[i].name, ids[k], fidata[k].name);
				ct++;
			}
			ch++;
		}

		/* No change, so don't write anything */
		if (!ch) continue;

		/* Output internal id */
		printf("<input type=\"hidden\" name=\"tar_%did\" value=\"%s\">", c, ids[i]);

		/* And internal new target details */
		for (j = 0; (j < 6); j++) printf("<input type=\"hidden\" name=\"tar_%da%d\" value=\"%s\">",
										 c, j, ids[fidata[i].bass[j]]);

		/* Update change counts */
		c++; if (ct) t++;
	}

	printf("</ol><p>These changes affect %i players, and alter %i player%s targets.</p>",
		   c, t, (t == 1) ? "'s" : "s'");

	return (-n);
}
int fiddletargs(player_p plptr)
{
	char* ids[2+6*7];
	int i, j;

	for (i = j = 0; (i < 6); i++) if (strcmp(plptr->con[i], NON_ID)) ids[2 + j++] = plptr->con[i];

	return prepretarget(ids, j, 2+6*7, plptr->id);
}
int genretarget(int mode, int max)
{
	int i, j = 0;
	char vid[32];


	/* Prepare the retargetting for up to 'max' players */
	if (mode == 0)
	{
		int lim = 2 + ((6 * max) > MAX_players) ? (6 * max) : (MAX_players);
		char* ids[lim];
	
		/* Count and acquire the player ids passed */
		while ((j < max) && (sprintf(vid, "tar_%did", j) > 6) && (*(ids[2 + j] = findval(vid, 1)))) j++;

		/* Too many player ids? */
		if (j >= max) return (DOC_ManyTargets);

		/* Output HTTP headers */
		printf("Content-type: text/html" LF
			   LF "<HTML><HEAD><TITLE>" GAME " Assassins' Semi-Automatic Umpire</TITLE></HEAD>"
			   LF "<BODY bgcolor=#000000 text=#FFFFDD link=#BBBBFF vlink=#FFCCFF alink=#FFFFFF>"
			   LF "<h1>Semi-Automatic Umpire: General Retargetting</h1>"
			   LF "<form method=\"POST\" action=\"" CGI_URL THIS_CGI "\">"
			   LF "<input type=\"hidden\" name=\"security\" value=\"" SECURITY "\">"
			   LF "<p>Trying to maximise targetting amongst selected %i players... results:</p>"
			   LF "<input type=\"hidden\" name=\"sort\" value=\"%s\">", j, findval("sort", 1));

		/* Handle the preparation */
		i = prepretarget(ids, j, lim, 0);

		/* Allow submission if sucessful */
		if (i < 0) printf(LF "<p><input value=\"Make %i link%s as shown\" type=\"submit\" name=\"sub_dotarg\"></p>",
						  -i, (i != -1) ? "s" : "");

		/* Report errors */
		else if (i > 0) printf("<h4>Targetting error: code %d</h4>", i);

		/* Footer */
		printf(LF "</form><p><form method=\"POST\" action=\"" CGI_URL "stats.cgi\">"
			   LF "<input type=\"hidden\" name=\"security\" value=\"" SECURITY "\">"
			   LF "<input value=\"Return to stats page\" type=\"submit\" name=\"resort\"></form></p>"
			   LF "<p align=right>Produced at %s.</p></BODY></HTML>" LF, ctime(&thetime));

		return 0;
	}

	/* Carry out the retargetting */
	else
	{
		unsigned long tf;
		struct neventp nevp;
		FILE* fp = 0;

		/* General targetting submissions */
		if (mode == 1)
		{
			/* Prepare temporary file for chained stdin */
			if (!(fp = prepforstdin())) return (DOC_TmpFileFailed);

			/* Encode special display type */
			urlencode(fp, "sub_show", "AutoGen", 1);
		}

		/* Insert up to 'max' target events */
		for (i = j = 0; (j < 1) && (i < max) && (sprintf(vid, "tar_%did", i) > 6) && (*(nevp.pid = findval(vid, 1))); i++)
		{
			/* Read in new targets/assassins */
			for (j = 0; (j < 6) && (sprintf(vid, "tar_%da%d", i, j) > 6) && (*(nevp.nids[j] = findval(vid, 1))); j++);
	
			/* Missing targets/assassins */
			if (j < 6) { j = (DOC_MissingIds); break; }
	
			/* No reference yet */
			nevp.ref = 0;
			nevp.bonup = 0;
	
			/* Event time is now */
			nevp.whenp = nevp.ntimp = &dtim;
	
			/* Event flags */
			*(nevp.flsp = &tf) = (EF_SetTargets);
	
			/* Add retargetting event */
			if ((j = newevent(&nevp, 0, 0, nevp.pid, 0, &dtim, &plbuf_pos, plbuf_end)) > 0) break;

			if (mode == 1)
			{
				fputs("&id_", fp);
				urlencode(fp, nevp.pid, "show", 1);
			}
		}

		/* Run scoring update executable */
		if ((i) && (mode == 1)) system(SYS_Scores" 2>> "FILE_Log" 1>&2 &");

		/* Return errors */
		if (j > 0) return (j);

		/* Normal return, otherwise chain stats.cgi */
		else if (mode != 1) return 0;

		/* Encode sorting method */
		urlencode(fp, "sort", findval("sort", 1), 0);

		/* Encode password */
		urlencode(fp, "security", SECURITY, 0);

		/* Execute the new binary, taking input from temporary file */
		return launchwithstdin(CGI_ROOT "stats.cgi", fp);
	}
}
#endif

#endif
/***** End of file *****/
