/*****************************************************
**  event.c                           >> event.cgi  **
**  ----------------------------------------------  **
**  Build a CGI executable to deal with addition/   **
**  editing/viewing of all events and reports, in   **
**  particular assassinations/attempts. Note that   **
**  for other types of event, the initial event     **
**  submission form is generated by a different     **
**  executable, but is actually acted upon by this  **
**  one (possibly after some pre-processing).       **
**  ----------------------------------------------  **
**  Updating to use new time system.                **
**                                                  **
************** BORING LEGAL NOTICE *******************
**                                                  **
**  Assassins' Game (Semi-)Automatic Umpire System  **
**  ----------------------------------------------  **
**  Prerelease: Copyright 2000-2002 Edward M Nokes  **
**  This code may not be distributed in source or   **
**  executable form without the express consent of  **
**  the author, Edward M Nokes (emn23@cam.ac.uk).   **
**  The author accepts no responsibility or         **
**  liability for anything occurring as a result    **
**  of using this software, etc. etc.               **
**  ----------------------------------------------  **
*****************************************************/

#ifndef THIS_CGI
 #define THIS_CGI			"event.cgi"
#endif

#define MAX_statbuf			(1024*1024)
#define MAX_fpbuf			65536
#define MAX_cgibuf 			65536		/* XXX --- Size of buffer for CGI variables --- use statbuf intstead */
#define MAX_npairs 			512			/* Limit on number of name=val pairs accepted by CGI routine */
#define EVENT_EXT_TYPE		int			/* type of  event_p->ext is irrelevent */
#define PLAYER_EXT_TYPE		player_p	/* type of player_p->ext is player_p   */
#define PSTATE_EXT_TYPE		pexist_p	/* type of pstate_p->ext is pexist_p   */
#define PSTATE_PLP_TYPE		player_p	/* type of pstate_p->plp is player_p   */
#define PEXIST_EXT_TYPE		int			/* type of pexist_p->ext is irrelevent */

#include <math.h>
#include "!lib-general.c"
#include "!custom.c"
#include "!lib-uinfo-scan.c"
#include "!lib-uinfo-misc.c"
#include "!lib-event-read.c"
#include "!lib-event-edit.c"
#include "!lib-evolve.c"
#include "!lib-rep-read.c"
#include "!lib-rep-edit.c"
#include "!lib-retarg.c"

/***************************************************************************************************************************
** Declarations and definitions.                                                                                          **
***************************************************************************************************************************/

/* Priviledges granted under this authentication */
#define AD_ASS			1					/* Authenticated as assassin */
#define AD_MIN			2					/* Authenticated as admin */
#define	AD_TAR			4					/* Authenticated as victim */
#define AD_INV			(AD_ASS|AD_TAR)		/* Involved in event (assassin or victim) */
#define AD_ASS_PSEUD	8					/* Can fiddle assassins's pseud */
#define	AD_TAR_PSEUD	16					/* Can fiddle target's pseud */
#define AD_OBS			32

/* A type to hold a record of which event flags are editable via options */
typedef unsigned long tst_flags;

/* Editable flags */
#define TST_Invisible			(((tst_flags)1)<<0)
#define TST_Open				(((tst_flags)1)<<1)
#define TST_Ignore				(((tst_flags)1)<<2)
#define TST_InvertAutoRes		(((tst_flags)1)<<3)
#define TST_SetP2Pseudonym		(((tst_flags)1)<<4)
#define TST_NoDangerBonus		(((tst_flags)1)<<5)
#define TST_SetP1Pseudonym		(((tst_flags)1)<<6)
#define TST_Annulled			(((tst_flags)1)<<7)
#define TST_Comment				(((tst_flags)1)<<8)
#define TST_DetargP1			(((tst_flags)1)<<9)
#define TST_DetargP2			(((tst_flags)1)<<10)

/* A type to hold a record of problems when adding player status events */
typedef unsigned int bad_actev_type;

/* Things that can go wrong */
#define BAD_actev_PseudMatch	(((bad_actev_type)1)<<0)
#define BAD_actev_Pseud			(((bad_actev_type)1)<<1)
#define BAD_actev_Boost			(((bad_actev_type)1)<<2)
#define BAD_actev_WantAct		(((bad_actev_type)1)<<3)
#define BAD_actev_Min			(((bad_actev_type)1)<<4)
#define BAD_actev_Hour			(((bad_actev_type)1)<<5)
#define BAD_actev_Day			(((bad_actev_type)1)<<6)
#define BAD_actev_PID			(((bad_actev_type)1)<<7)
#define BAD_actev_Missing		(((bad_actev_type)1)<<8)
#define BAD_actev__MAX_IDX		(9)

/* Descriptions of each problem */
char* doact_badstuff[BAD_actev__MAX_IDX] =
{
	"This pseudonym has already been taken by another player.<br/>",
	"You have not supplied a new pseudonym.<br/>",
	"Invalid boost to incompetence deadline.<br/>",
	"You have not specified a valid action.<br/>",
	"(xxx)<br/>",
	"(xxx)<br/>",
	"(xxx)<br/>",
	""
};

/* A type to hold a record of problems when adding assassination events or editing events/reports */
typedef unsigned long bad_event_type;

/* Things that can go wrong */
#define BAD_event_PID				(((bad_event_type)1)<<0)
#define BAD_event_SpecVic			(((bad_event_type)1)<<1)
#define BAD_event_DeadVic			(((bad_event_type)1)<<2)
#define BAD_event_SpecAss			(((bad_event_type)1)<<3)
#define BAD_event_DeadAss			(((bad_event_type)1)<<4)
#define BAD_event_Min				(((bad_event_type)1)<<5)
#define BAD_event_Hour				(((bad_event_type)1)<<6)
#define BAD_event_Day				(((bad_event_type)1)<<7)
#define BAD_event_Type				(((bad_event_type)1)<<8)
#define BAD_event_PseudAss			(((bad_event_type)1)<<9)
#define BAD_event_PseudAssMatch		(((bad_event_type)1)<<10)
#define BAD_event_PseudVic			(((bad_event_type)1)<<11)
#define BAD_event_PseudVicMatch		(((bad_event_type)1)<<12)
#define BAD_event_PseudRep			(((bad_event_type)1)<<13)
#define BAD_event_PseudRepMatch		(((bad_event_type)1)<<14)
#define BAD_event_Headline			(((bad_event_type)1)<<15)
#define BAD_event_RetAss			(((bad_event_type)1)<<16)
#define BAD_event_RetVic			(((bad_event_type)1)<<17)
#define BAD_event_RepOwner			(((bad_event_type)1)<<18)
#define BAD_event_RepText			(((bad_event_type)1)<<19)
#define BAD_event_RepNotFromEvent	(((bad_event_type)1)<<20)
#define BAD_event_MainRep			(((bad_event_type)1)<<21)
#define BAD_event__MAX_IDX			(22)

/* Descriptions of each problem (first parameter is a string: the report id) */
char* genfef_badstuff[BAD_event__MAX_IDX] =
{
	"Unknown or missing player(s).<br/>",
	"Invalid choice of victim.<br/>",
	"The chosen victim was already dead at the time!<br/>",
	"Invalid choice of assassin.<br/>",
	"The chosen assassin was dead at the time!<br/>",
	"Invalid time for event (min).<br/>",
	"Invalid time for event (hour).<br/>",
	"You can't submit events on that date.<br/>",
	"Invalid or missing event outcome.<br/>",
	"Missing or invalid pseudonym data for assassin.<br/>",
	"The pseudonym for the assassin has already been taken by another player.<br/>",
	"Missing or invalid pseudonym data for victim.<br/>",
	"The pseudonym for the victim has already been taken by another player.<br/>",
	"Missing or invalid pseudonym data.<br/>",
	"That pseudonym has already been taken by another player.<br/>",
	"Invalid or missing headline HTML fragment.<br/>",
	"The chosen assassin has retired!<br/>",
	"The chosen victim has retired!<br/>",
	"Can't add reports under this authentication.<br/>",
	"Care to provide some text to go with your report?<br/>",
	"Report \"%s\" does not belong to this event.<br/>",
	"Invalid or missing main news HTML fragment.<br/>"
};

/* A type to hold a record of sucessfully added/edited events or reports */
typedef unsigned int good_event_type;

/* The things that can go well */
#define GOOD_event_RepAdded			(((good_event_type)1)<<0)
#define GOOD_event_EventAdded		(((good_event_type)1)<<1)
#define GOOD_event_RepEdited		(((good_event_type)1)<<2)
#define GOOD_event_EventEdited		(((good_event_type)1)<<3)
#define GOOD_event__MAX_IDX			(4)

/* Descriptions of sucesses (first parameter is a string: the report id) */
char* genfef_goodstuff[GOOD_event__MAX_IDX] =
{
	"Report \"%s\" added successfully.<br/>",
	"Event added successfully.<br/>",
	"Report \"%s\" edited successfully.<br/>",
	"Event edited successfully.<br/>"
};

/* A type to hold a record of warnings when adding/editing events or reports */
typedef unsigned int warn_event_type;

/* The different warning flags */
#define WARN_event_RepLength		(((warn_event_type)1)<<0)
#define WARN_event__MAX_IDX			(1)

/* Descriptions of sucesses (first parameter is a string: the report id) */
char* genfef_warnstuff[WARN_event__MAX_IDX] =
{
	"Report text was too long... excess has been truncated."
};

/* Null-terminated list of CGI variable names to exclude from regurgitation if authentication fails */
char* excl_auth[4] =
{
	"ev_ref", "authname", "authpass", 0
};

/* More global variables */
static FILE* flaunch = 0;
static player_p plist = 0;
static int lockfd = 0;
static char* suggestid = 0;

/* Various things may need updating after execution */
#define RESC_Scoring		(1U<<0)
#define RESC_MkNews			(1U<<1)
#define RESC_Flaunch		(1U<<2)
#define RESC_ERR__Shift		(8)
#define RESC_ERR_Scoring	(1U<<8)
#define RESC_ERR_MkNews		(1U<<9)
#define RESC_ERR_Flaunch	(1U<<10)

/* By default, update nothing */
unsigned int rescore = 0;

/* Function declarations */
int send_doc(int which, player_p authp);
void authform(int which, char* file, player_p authp, char* authname);

/***************************************************************************************************************************
** Helper functions for dealing with the event and player state event forms.                                              **
***************************************************************************************************************************/
int genfef_hook(event_p ep, int match, void* x5, evolve__ctx_t* ectx)
{
	pstate_p pstp;
	pexist_p exp;

	/*fprintf(stderr, "%li   hook called on ref %s\n", uniqid, (ep) ? (ep->ref) : "late");*/

	/* Continue to current time */
	ectx->endtime = thetime;

	/* Don't call this hook again, and be inclusive from now on */
	ectx->flags &= ~(EVOLVE_FuncEndCond|EVOLVE_FuncExhaust|EVOLVE_Exclusive);

	/* Don't stop at any specific event */
	ectx->endref = 0;

	/* Note all current data */
	for (pstp = ectx->plist; (pstp); pstp = pstp->next) if (pstp->state & EVOLVE_Init)
	{
		/*fprintf(stderr, "%li  h %s \"%s\"", uniqid, pstp->pid, (pstp->exist->dps) ? (pstp->exist->dps->text) : "???");*/

		/* Allocate new existence structure */
		if (!(exp = malloc(sizeof(struct pexist)))) return -(ERF_mallocFailed);

		/*fprintf(stderr, "%li  fefhook %s &%lx\n", uniqid, pstp->pid, pstp->exist->fl);*/

		/* Copy part of current existence structure into it */
		exp->fl       = pstp->exist->fl;
		exp->score    = pstp->exist->score;
		exp->timer    = pstp->exist->timer;
		exp->twant    = pstp->exist->twant;
		exp->born     = pstp->exist->born;
		exp->died     = pstp->exist->died;
		exp->killedby = 0;
		exp->pseuds   = exp->dps = pstp->exist->dps;
		exp->scoretab = 0;
		exp->ext      = 0;
		exp->link_all = exp->link_pl = 0;

		/* Wantedness comment */
		if (exp->comment = pstp->exist->comment)
		{
			/* Allocate space for current wantedness */
			if (!(exp->comment = malloc(1 + strlen(pstp->exist->comment)))) return -(ERF_mallocFailed);

			/* Copy current wantedness */
			else strcpy(exp->comment, pstp->exist->comment);
		}

		/* Save pointer */
		pstp->ext = exp;
	}

	/* Continue, and recheck term conditions */
	return 0;
}

/* Print out the basic top-level event submission form */
int printblankform(int mode, player_p p1ptr, player_p p2ptr, bad_event_type bad, timeval when,
				   player_p authp, unsigned long f, int head, char* prep, char* phlin)
{
	int i;

	if (head & 1) puts(LF "<HTML><HEAD><TITLE>"GAME" Assassins' Automatic Umpire</TITLE></HEAD>"
					   LF "<BODY bgcolor=#000000 text=#FFFFDD link=#BBBBFF vlink=#FFCCFF alink=#FFFFFF>"
					   LF "<h1>Automatic Umpire: Add new event</h1>");

	if (head & 4) printf(LF "<p>Authenticated as <b>%s</b>%s.</p><p><b>Everything should be working now.</b></p>",
						 authp->name, ((authp->flags & PF_Admin) ? " (admin)" : ""));

	if (bad)
	{
		puts(LF "<h3><img src=\""CROSS_URL"\" align=left "CROSS_SIZ" hspace=8>"
			 LF "<font color=#FFCCCC>That was really rather dubious.</font></h3><p>");

		for (i = 0; (i < BAD_event__MAX_IDX); i++) if (bad & (1UL << i)) puts(genfef_badstuff[i]);

		puts("</p>");
	}

	if (mode < 2)
	{
		printf(LF "<form method=\"POST\" action=\""CGI_URL THIS_CGI"\">"
			   LF "<input type=\"hidden\" name=\"authname\" value=\"%s\">"
			   LF "<input type=\"hidden\" name=\"authpass\" value=\"%s\">"
			   LF "<h2><hr/><font color="WHITE">Add incident</font><hr/></h2>"
			   LF "<p><table cols=2 cellspacing=2 cellpadding=2>"
			   LF "<tr><td align=right>By:</td><td><select name=\"ev_by\" size="OPT_SelBySize">",
			   authp->id, authp->pass);

		if (i = makeidmenu(((p1ptr) ? (p1ptr->id) : 0), PF_Playing, PF_Playing, (authp->flags & PF_Observer) ? 0 : 1))
		{
			fprintf(stderr, "%li:   rc %d from makeidmenu(1) in printblankform(%d,%d)\n", uniqid, i, mode, head);
		}

		printf(LF "</select></td></tr><tr><td align=right>Outcome:</td>"
			   LF "<td><select name=\"ev_type\" size="OPT_SelTypeSize">"
			   LF "<option %svalue=\"fail\">Failed attempt</option>"
			   LF "<option %svalue=\"kill\">Successful kill</option></select>",
			   ((f & EF_Kill) ? "" : "selected "), ((f & EF_Kill) ? "selected " : ""));

		if ((authp->flags & PF_Admin))
		{
			puts("&nbsp; <input type=\"checkbox\" name=\"ev_retarg\" "OPT_AutoRetarg" value=\"init\">"
				 "&nbsp; Attempt retargetting");
		}

		puts("</td></tr><tr><td align=right>On:</td><td><select name=\"ev_on\" size="OPT_SelOnSize">");

		if (i = makeidmenu(((p2ptr) ? (p2ptr->id) : 0), PF_Victim, PF_Victim, (authp->flags & PF_Observer) ? 0 : 1))
		{
			fprintf(stderr, "%li:   rc %d from makeidmenu(2) in printblankform(%d,%d)\n", uniqid, i, mode, head);
		}

		puts(LF "</select></td></tr><tr><td align=right>Time and date:</td><td>");

		printdatesel("ev_d", "ev_h", "ev_m", when, (authp->flags & PF_Admin) ? 1 : 0);

		puts(LF "</td></tr><tr><td></td><td><input value=\"Add incident\" type=\"submit\" name=\"sub_prime\">"
			 LF "</td></tr></table></p></form>");
	}

	if (mode == 2)
	{
		printf(LF "<form method=\"POST\" action=\""CGI_URL THIS_CGI"\">"
			   LF "<input type=\"hidden\" name=\"authname\" value=\"%s\">"
			   LF "<input type=\"hidden\" name=\"authpass\" value=\"%s\">"
			   LF "<input type=\"hidden\" name=\"ev_by\" value=\"%s\">"
			   LF "<input type=\"hidden\" name=\"ev_on\" value=\"%s\">"
			   LF "<h2><hr/><font color="WHITE">Add comment to news</font><hr/></h2>"
			   LF "<p><table cols=2 cellspacing=2 cellpadding=2>"
			   LF "<tr><td align=right>Time and date:</td><td>",
			   authp->id, authp->pass, ((p1ptr) ? (p1ptr->id) : ""), ((p2ptr) ? (p2ptr->id) : ""));

		printdatesel("ev_d", "ev_h", "ev_m", when, ((authp->flags & PF_Admin) ? 1 : 0));

		printf("</td></tr>" LF "<tr><td align=right>Event flags:</td><td>"
			   LF "<input type=\"checkbox\" name=\"ev_sopen\" value=\"set\"%s> &nbsp; Open &nbsp;"
			   LF "<input type=\"checkbox\" name=\"ev_sunsc\" value=\"set\"%s> &nbsp; Ignore &nbsp;"
			   LF "<input type=\"checkbox\" name=\"ev_sinv\" value=\"set\"%s> &nbsp; Invisible &nbsp;</td></tr>"
			   LF "<tr><td align=right valign=top><input type=\"checkbox\" name=\"ev_shead\" value=\"set\"%s>"
			   LF "&nbsp; Show Headline:<br/></td><td><textarea name=\"ev_head\" rows=3 cols=60 wrap=virtual>",
			   ((f & EF_Open) ? " checked" : ""),
			   ((f & EF_Ignore) ? " checked" : ""),
			   ((f & EF_Invisible) ? " checked" : ""),
			   ((f & EF_ShowHeadline) ? " checked" : ""));

		if (phlin) encapsulate(phlin, stdout);

		fputs("</textarea></td></tr><tr><td align=right valign=top>Base HTML fragment:</td>"
			  LF "<td><textarea name=\"ev_html\" rows=32 cols=60 wrap=virtual>", stdout);

		if (prep) encapsulate(prep, stdout);

		puts("</textarea></td></tr>"
			 LF "<tr><td></td><td><input value=\"Add comment\" type=\"submit\" name=\"sub_com\"></td></tr>"
			 LF "</table></p></form>");
	}

	if (head & 2) puts("<hr/><p>"FOOTER"<hr/></p></body></html>");

	return 0;
}

/* Strip pseudonym, and check it has not already been used. */
ulong checkpseud(char *p1, char *p2, char* p3, unsigned long *flptr, char** ps, char* plid, pstate_p pstl)
{
	pstate_p pstp = 0;
	char *p, *q;

	/* Use given text */
	if (!p1) p = p2;

	/* Search for standard pseudonym variables */
	else if (!(p = findval(p1, 0))) p = findval(p2, 1);

	/* Allow changing of default pseudonym */
	if (p3) *flptr = ((findval(p3, 0)) ? EF_SetPseudonym : EF_AddPseudonym);

	/* Clean up text - Strip leading white space */
	while ((*p == ' ') || (*p == '')) p++;

	/* Strip trailing white space */
	for (q = p + strlen(p); (--q > p) && ((*q == ' ') || (*q == '')); *q = 0);

	/* Missing pseudonym text? */
	if (!(*ps = (*p) ? p : 0)) return (BAD_event_PseudAss);

	/* Scan each evolved player */
	for (pstp = pstl; (pstp); pstp = pstp->next)
	{
		ppseud_p pseudp;

		/* Scan all evolved players' pseudonyms */
		for (pseudp = pstp->allpseuds; (pseudp); pseudp = pseudp->link_pl)
		{
			/* Matching? */
			if (!strcasecmp(p, pseudp->text)) break;
		}

		/* Early return for match if does not already belong to this player */
		if ((pseudp) && (strcmp(pstp->pid, plid))) break;
	}

	/* Someone else has already claimed this pseudonym? */
	if (!pstp) return 0;

	fprintf(stderr, "%li:   Pseudo-match on \"%s\"\n", uniqid, pstp->pid);

	return (BAD_event_PseudAssMatch);
}

/* Print out a pseudonym selector fragment of HTML */
void pseudselector(char* def, int set, char* seln1, char* seln2, char* seln3, ppseud_p pslist)
{
	ppseud_p pseudp;
	int i = 0;

	/*puts("<table border=0 cellspacing=2 cellpadding=0><tr><td>");*/

	if ((pslist) || (set & 2))
	{
		printf("<select name=\"%s\" size=1>", seln1);

		if (set & 2) printf("<option value=\"-\"%s>Standard pseudonym</option>"
							"<option value=\"-\">-----(or)-----</option>", (def) ? "" : " selected");

		puts("<option value=\"\">New pseudonym:");

		if (pslist) puts("</option><option value=\"\">-----(or)-----");
	
		for (pseudp = pslist; (pseudp); pseudp = pseudp->link_pl)
		{
			i = (i) ? 2 : ((def) && (!strcasecmp(pseudp->text, def))) ? 1 : 0;
		
			printf("</option>" LF "<option value=\"");
			encapsulate(pseudp->text, stdout);
			printf("\"%s>", (i == 1) ? " selected" : "");
			encapslim(pseudp->text, stdout, MAX_PseudDisplay);
		}
	
		fputs("</option></select>&nbsp;&nbsp;\n" /*"</td></tr><tr><td>"*/, stdout);
	}

	printf("<input type=\"text\" name=\"%s\" value=\"", seln2);
	
	if ((i == 0) && (def)) encapsulate(def, stdout);
	
	if (!seln3) printf("\">");

	else printf("\">&nbsp;&nbsp;&nbsp;<input type=\"checkbox\" name=\"%s\" value=\"set\"%s>"
				"&nbsp;&nbsp;Set&nbsp;default", seln3, ((set & 1) ? " checked" : ""));

	/*puts("</td></tr></table>");*/
}

/* Print out a table of information about an assassination event */
void printevoverview(event_p eptr, unsigned long f, player_p p1ptr, player_p p2ptr, char* p1ps, char* p2ps, int admin)
{
	puts("</td></tr><tr><td colspan=2 bgcolor="GREEN">");

	/* Print table header for assassinations */
	if (f & EF_Assassination)
	{
		printf(LF "<table width=100%% border=0 cols=5 bgcolor=#000000 cellpadding=4 cellspacing=2>"
	 		   LF "<thead><td colspan=2 bgcolor="GREEN"><b>%s</b></td>"
	 		   LF "<td colspan=2 bgcolor="GREEN">Status</td>"
	 		   LF "<td bgcolor="GREEN">Standard pseudonyms:</td></thead>"
			   LF "<tr><td bgcolor="GREEN">Assassin:</td>",
	 		   (f & EF_Kill) ? "Successful kill" : "Failed attempt");
	}

	/* Print table header for other events */
	else printf(LF "<table width=100%% border=0 cols=5 bgcolor=#000000 cellpadding=4 cellspacing=2>"
				LF "<tr><td bgcolor="GREEN">Player:</td>");

	/* Special players aren't linked to */
	if (p1ptr->flags & PF_Special) printf("<td bgcolor="RED" colspan=3><b>%s</b></td><td bgcolor="RED">", p1ptr->name);

	/* Show name and plinfo link for normal players */
	else
	{
		printf(LF "<td bgcolor="RED"><a href=\"plinfo.cgi?id=%s\">", p1ptr->id);

		if (admin & AD_OBS) puts(p1ptr->name);
		else if (p1ptr->pseud) encapsulate(p1ptr->pseud, stdout);
		else printf("[**NoPseud:%s**]", p1ptr->id);

		printf("</a></td>" LF "<td bgcolor="RED"><a href=\"plinfo.cgi?id=%s&fref=%s&prior=1\">Pre</a></td>"
			   LF "<td bgcolor="RED"><a href=\"plinfo.cgi?id=%s&fref=%s\">Post</a></td>"
			   LF "<td bgcolor="RED">", p1ptr->id, eptr->ref, p1ptr->id, eptr->ref);
	}

	/* Print pseudonym */
	if (p1ps) encapsulate(p1ps, stdout);

	/* None given */
	else puts("<b>Default</b> (no pseudonym specified)");

	/* Another player involved? */
	if (p2ptr)
	{
		/* Begin table segment */
		puts(LF "</td></tr><tr><td bgcolor="GREEN">Victim:</td>");
	
		/* Special players aren't linked to */
		if (p2ptr->flags & PF_Special) printf("<td bgcolor="BLUE" colspan=3><b>%s</b></td><td bgcolor="BLUE">",
											  p2ptr->name);
	
		/* Show name and plinfo link for normal players */
		else
		{
			printf(LF "<td bgcolor="BLUE"><a href=\"plinfo.cgi?id=%s\">", p2ptr->id);

			if (admin & AD_OBS) puts(p2ptr->name);
			else if (p2ptr->pseud) encapsulate(p2ptr->pseud, stdout);
			else printf("[**NoPseud:%s**]", p2ptr->id);

			printf("</a></td>" LF "<td bgcolor="BLUE"><a href=\"plinfo.cgi?id=%s&fref=%s&prior=1\">Pre</a></td>"
				   LF "<td bgcolor="BLUE"><a href=\"plinfo.cgi?id=%s&fref=%s\">Post</a></td>"
				   LF "<td bgcolor="BLUE">", p2ptr->id, eptr->ref, p2ptr->id, eptr->ref);
		}
	
		/* Print victim's pseudonym */
		if (p2ps) encapsulate(p2ps, stdout);

		/* None given */
		else puts("<b>Default</b> (no pseudonym specified)");
	}

	puts("</td></tr></table>");
}

/***************************************************************************************************************************
** Process submissions from event form, in its various stages: 0/1 sub event, 2 sub comment; 3 generate event overview    **
** form; 4 handle event edit submits; 5/6 generate add/edit report form, 7 (and internally, 8) deal with its submission.  **
** Parameter: (strparm) for (mode) 3 can be null, or an event ref; for (mode) 6, must be report id; ignored for others.   **
***************************************************************************************************************************/
int genfulleventform(int mode, player_p authp, player_p p1ptr, player_p p2ptr, player_p p3ptr,
					 char** statbuf_pos, char* statbuf_end, char* strparm)
{
	char *p, ridbuf[MAX_rid+1], *rid = 0, *ref = "XXX", *title, *comment = 0;
	unsigned long r, f = 0, p1f = 0, p2f = 0, p3f = 0;
	pstate_p psarr[4] = { 0, 0, 0, 0 }, *psidx = psarr, pslist = 0, p1sp = 0, p2sp = 0, p3sp = 0;
	pexist_p ex1p = 0, ex2p = 0, ex3p = 0;
	char *ipids[4] = { 0, 0, 0, 0 }, *p1ps = 0, *p2ps = 0, *mainrep = 0, *headline = 0;
	int rc = 0, i, j, admin, gotreps = 0, retargetting = 0;
	bad_event_type bad = 0, oldbad = 0;
	good_event_type good = 0;
	warn_event_type warn = 0;
	struct report newrep;
	report_p newrptr = 0, rptr, rlist = 0;
	struct period *perp;
	event_p eptr, oldeptr = 0;
	timeval tmptime, when = thetime;
	tst_flags testf;
	long bonus = 0, l;
	u32b boost = 0;
	struct editev nev;
	retarg_p untargs = 0, retargs = 0;


	/* Priv of submitter */
	admin = (authp->flags & PF_Admin) ? (AD_MIN|AD_ASS|AD_ASS_PSEUD|AD_TAR|AD_TAR_PSEUD) : 0;

	/* Output HTTP headers */
    printf("Content-type: text/html" LF LF);

	/* Paranoia: check mode parameter */
	if ((mode >= 8) || (mode < 0)) erf(ERF_BadParms, 1);

	/* Check string parameter - should be report id in some modes */
	if ((mode == 6) && ((!strparm) || (!strparm[0]))) erf(ERF_BadParms, 2);

	/* New event */
	if (mode <= 2)
	{
		u32b d, m, h;

		/* Require first player id */
		if (!p1ptr) bad |= BAD_event_PID;

		/* "Add comment" event */
		if (mode == 2)
		{
			/* Must be authenticated as an admin */
			if (!admin) erf(ERF_BadAuth, 3);

			/* Don't need another player */
			p2ptr = 0;

			/* Don't need to evolve flags */
			p1f = p1ptr->flags;
		}

		/* Not a comment event... */
		else
		{
			/* Require another player id */
			if (!p2ptr) bad |= BAD_event_PID;

			/* Require either a "kill" or a "fail" */
			if ((i = strcasecmp(p = findval("ev_type", 1), "fail")) && (strcasecmp(p, "kill"))) bad |= BAD_event_Type;
	
			/* Set to flags from ev_type */
			f = (i == 0) ? (0) : (EF_Kill);
	   	}

		/* Require valid event time: min */
		if (!(p = findval("ev_m", 0)) || (str2u32b(&m, p, 1, 2)) || (m > 59))
		{
			m = T_m(thetime);
			bad |= BAD_event_Min;
		}
	
		/* Require valid event time: hour */
		if (!(p = findval("ev_h", 0)) || (str2u32b(&h, p, 1, 2)) || (h > 23))
		{
			h = T_h(thetime);
			bad |= BAD_event_Hour;
		}

		/* Require valid event time: hour */
		if (!(p = findval("ev_d", 0)) || (str2u32b(&d, p, 1, 2)))
		{
			h = T_h(thetime);
			bad |= BAD_event_Day;
		}

		/* Convert to timeval */
      
		if (!(when = TVAL(d, h, m))) erf(ERF_BadTimeVal, 3);

		/* Check range */
		if ((when < (tmptime = (admin) ? TIME_eventFullMin : TIME_eventMin)) ||
			(when > (tmptime = (admin) ? TIME_eventFullMax : TIME_eventMax)))
		{
			when = tmptime;
			bad |= BAD_event_Day;
		}
	}

	/*fprintf(stderr, "[here]\n");*/

	/* New event, not comment (so both p1ptr and p2ptr valid) */
	if (mode <= 1)
	{
		/* Collect mistakes found so far */
		if (bad) erf(printblankform(mode, p1ptr, p2ptr, bad, when, authp, f, 1+2+4,0,0), 4);

		/* Players involved in events can do more */
		if (authp == p1ptr) admin |= AD_ASS;
		if (authp == p2ptr) admin |= AD_TAR;

		/* Players involved in events can do more */
		#if OPT_ChangePseud
		if (authp == p1ptr) admin |= AD_ASS_PSEUD;
		if (authp == p2ptr) admin |= AD_TAR_PSEUD;
		#endif

		/* No extra privs -> skip secondary stage */
		if ((!admin) && (mode == 0)) mode = 1;

		/* Interesting players */
		ipids[0] = p1ptr->id;
		ipids[1] = p2ptr->id;
		ipids[2] = 0;

		/* Evolve some relevent players - determine participants' states at time of event */
		if (rc = evolve(EVOLVE_FuncEndCond|EVOLVE_FuncExhaust|EVOLVE_TimersToLimit|EVOLVE_MatchAll|EVOLVE_Targetting|
						EVOLVE_Comments|EVOLVE_AllPseuds|EVLadd_AllPseuds|EVOLVE_InheritPseud|EVLadd_Comments,
						ipids, &psidx, &pslist, 0, 0,0,0,0,0, when, 0,
						*statbuf_pos, statbuf_end, statbuf_pos, 0,0,0, genfef_hook, 0, 0)) {erf(rc, 5);}

		fprintf(stderr, "%li   evolved p1sp=%p p2sp=%p\n", uniqid, psidx[0], psidx[1]);

		/* Existences at time of event */
		if ((!(p1sp = psidx[0])) || !(ex1p = p1sp->ext)) erf(ERF_InternalError, 6);
		if ((!(p2sp = psidx[1])) || !(ex2p = p2sp->ext)) erf(ERF_InternalError, 7);

		/* Player flags at time of event */
		p1f = (p1ptr->flags & ~(PF_MASK_Scoring)) | (ex1p->fl & PF_MASK_Scoring);
		p2f = (p2ptr->flags & ~(PF_MASK_Scoring)) | (ex2p->fl & PF_MASK_Scoring);

		/* Dead assasins can't do anything */
		if (p1f & PF_Dead) bad |= BAD_event_DeadAss;
	
		/* Special players cannot do anything */
		if ((p1f & PF_Special) || !(p1f & PF_Playing)) bad |= BAD_event_SpecAss;
	
		/* Retired players cannot do anything */
		if ((p1f & PF_Retired)) bad |= BAD_event_RetAss;

		/* Dead victims cannot be killed */
		if ((f) && (p2f & PF_Dead)) bad |= BAD_event_DeadVic;
	
		/* Some players cannot be attacked */
		if (!(p2f & PF_Victim)) bad |= BAD_event_SpecVic;

		/* Retired players cannot do anything */
		if ((p2f & PF_Retired)) bad |= BAD_event_RetVic;

		/*fprintf(stderr, "[xxx]\n");*/

		/* Default pseudonyms */
 		p1ps = (ex1p->dps) ? (ex1p->dps->text) : 0;
		p2ps = (ex2p->dps) ? (ex2p->dps->text) : 0;

		/* Present mistakes */
		if ((bad) && !(admin & AD_MIN)) erf(printblankform(mode, p1ptr, p2ptr, bad, when, authp, f, 1+2+4,0,0), 8);

		/* Targetting */
		if ((admin & AD_MIN) && (f & EF_Kill))
		{
			if (!(p = findval("ev_retarg", 0))) retargetting = 0;
			else if (!strcasecmp(p, "detarg_vict")) retargetting = 2;
			else if (!strcasecmp(p, "retarg")) retargetting = 3;
			else if (!strcasecmp(p, "init")) retargetting = 4;
			else retargetting = 1;
		}
	}

	/* First call */
	if ((mode == 0) && (admin & AD_MIN))
	{
		/* If doing targetting, DON'T select deallocate option by default */
		/*if ((f & EF_Kill) && (findval("ev_retarg", 0))) f |= EF_DetargP2;*/

		/* Default comment is current one */
		comment = ex1p->comment;

		/* Default inactivity boost */
		boost = DEF_BOOST;
	}

	/* Secondary call */
	if (mode == 1)
	{
		/* Authorized to select assassin's pseudonym? */
		if (admin & AD_ASS_PSEUD)
		{
			/* Check for duplication */
			if (l = checkpseud("ev_p1psu", "ev_p1psn", "ev_p1psd", &r, &p1ps, p1ptr->id, pslist)) bad |= l;

			/* Not duplicated */
			else f |= r;
		}

		/* Automatic selection of assassin's pseudonym? */
		else p1ps = (ex1p->dps) ? (ex1p->dps->text) : 0;

		/* Authorized to select victim's pseudonym? */
		if (admin & AD_TAR_PSEUD)
		{
			/* Check for duplication */
			if (l = checkpseud("ev_p2psu", "ev_p2psn", "ev_p2psd", &r, &p2ps, p2ptr->id, pslist)) bad |= (l << 2);

			/* Not duplicated */
			else f |= ((r & EF_SetPseudonym) ? (EF_SetP2Pseudonym) : 0);
		}

		/* Automatic selection of victim's pseudonym  */
		else p2ps = (ex2p->dps) ? (ex2p->dps->text) : 0;

		/* Other administrative extras */
		if (admin & AD_MIN)
		{
			/* Scoring bonus */
			if ((p = findval("ev_boost", 0)) && str2u32b(&boost, p, 1, 2)) boost = 0;

			/* Scoring bonus */
			if ((p = findval("ev_bonus", 0)) && (bonus = rint((1 << SCORE_SHIFT) * strtod(p, 0)))) f |= EF_ScoreBonus;

			/* Detarget P2 */
			if (p = findval("ev_sdetarg_vict", 0)) f |= EF_DetargP2;

			/* Act on changes in wantedness */
			if (findval("ev_swant", 0)) f |= (p1f & PF_Wanted) ? EF_MakeUnWanted|EF_MakeCompetent : EF_MakeWanted;

			/* New comment string */
			comment = findnameval("ev_comment", 1+32, 0);

			/* (Un|Re)targetting */
			if ((retargetting == 3) || (retargetting == 2))
			{
				/* Read in results */
				if (i = readretargs(&untargs, (retargetting == 3) ? &retargs : 0, plist)) {
               erf(i, 170);
            }
			}

			/**** frag_3 was here ****/
		}
	}

	/*fprintf(stderr, "%li   there: p1ps=\"%s\" p2ps=\"%s\"\n", uniqid, STR(p1ps), STR(p2ps));*/

	/* Deal with news fragments (admin-only) */
	if (((mode == 1) || (mode == 2)) && (admin & AD_MIN))
	{
		/* Option: Ignore event */
		if (findval("ev_sunsc", 0)) f |= (EF_Ignore);

		/* Option: Invisible */
		if (findval("ev_sinv", 0)) f |= (EF_Invisible);

		/* Option: Open */
		if (findval("ev_sopen", 0)) f |= (EF_Open);

		/* Option: Show headline */
		if (findval("ev_shead", 0)) f |= (EF_ShowHeadline);

		/* Option: Invert normal resurrection discipline */
		if (findval("ev_sinvres", 0) && !(p2f & PF_Special) && (f & EF_Kill)) f |= (EF_InvertAutoRes);

		/* Acquire man report */
		if (!(mainrep = findval("ev_html", 0+4))) bad |= BAD_event_MainRep;

		/* Acquire headline */
		if (!(headline = findval("ev_head", 0))) bad |= BAD_event_Headline;
	}

	/* Default news fragments */
	else if ((mode <= 1))
	{
		/* Non-umpirical events are naturally open */
		f |= (EF_Open|EF_ShowHeadline);

		/* XXX --- Used to switch here on player death status at time of event */
		/* switch ((f & EF_Kill) | ((!ex2p) ? 0 : (ex2p->fl & PF_Dead) ? EF_Temp : 0)) */

		if (f & EF_Kill)
		{
			headline = "[P1] [t] [N2] ([L2])";
			mainrep = "<h4><hr />[[[d0]] [P1] [t] [N2] ([L2]) <hr /></h4>[Z]";
		}

		else
		{
			/* Use auto-title feature */
			headline = "[P1] [t] [P2]";
			mainrep = "<h4><hr />[[[d0]] [P1] [t] [P2]<hr /></h4>[Z]";
		}
	}

	/* Adding new event - Degrade to mode = 0 on input error */
	if ((bad) && (mode == 1) && (!(admin & AD_MIN) || !findval("ev_forcebad", 0))) mode = 0;

	/* Adding comments event - Present input errors */
	if ((bad) && (mode == 2)) erf(printblankform(2, p1ptr, 0, bad, when, authp, f, 1+2+4, mainrep, headline), 9);

	/* Primary call - collect more data */
	if (mode == 0)
	{
		ppseud_p pseudp;

		/* Sucessful kill? */
		p = (f & EF_Kill) ? "Sucessful kill" : "Unsucessful kill attempt";
		
		printf(LF "<HTML><HEAD><TITLE>"GAME" Assassins' Semi-Automatic Umpire</TITLE></HEAD>"
			   LF "<BODY bgcolor=#000000 text=#FFFFDD link=#BBBBFF vlink=#FFCCFF alink=#FFFFFF>"
			   LF "<h1>Semi-Automatic Umpire: Add event</h1>"
			   LF "<p>Authenticated as <b>%s</b>%s.</p><p><b>Meaningful messages can go here.</b></p>"
			   LF "<h2><hr/><font color="WHITE">Add %s event</font><hr/></h2>"
			   LF "<form method=\"POST\" action=\""CGI_URL THIS_CGI"\">",
			   authp->name, ((authp->flags & PF_Admin) ? " (admin)" : ""), p);
	
		if (bad)
		{
			puts("<h3><img src=\""CROSS_URL"\" align=left "CROSS_SIZ" hspace=8>"
				 "<font color=#FFCCCC>A Few Minor Imperfections Were Found...</font></h3><p><font color=#99DDFF>");

			for (i = 0; (i < BAD_event__MAX_IDX); i++) if (bad & (1UL << i)) puts(genfef_badstuff[i]);

			if (admin & AD_MIN) puts("</font></p><p><font color=#FFCCCC>"
									 LF "<input type=\"checkbox\" name=\"ev_forcebad\" value=\"set\"> &nbsp;"
									 LF "Ignore these trivial imperfections... (be careful)");

			puts("</font><br clear=left></p><hr/>");
		}

		printf(LF "<input type=\"hidden\" name=\"authname\" value=\"%s\">"
			   LF "<input type=\"hidden\" name=\"authpass\" value=\"%s\">"
			   LF "<input type=\"hidden\" name=\"ev_d\" value=\"%lu\">"
			   LF "<input type=\"hidden\" name=\"ev_h\" value=\"%lu\">"
			   LF "<input type=\"hidden\" name=\"ev_m\" value=\"%lu\">"
			   LF "<input type=\"hidden\" name=\"ev_type\" value=\"%s\">"
			   LF "<input type=\"hidden\" name=\"ev_by\" value=\"%s\">"
			   LF "<p><table cols=2 cellspacing=4 cellpadding=4>"
			   LF "<tr><td bgcolor=#AA0000 align=right><font color="WHITE"><b>Event:</b></font></td>"
			   LF "<td bgcolor=#AA0000><font color="WHITE"><b>%s at "TDISP" on %s</b></font></td></tr>",
			   authp->id, authp->pass, Tdhm(when),
			   ((f & EF_Kill) ? "kill" : "fail"),
			   p1ptr->id, p, Thm(when), fdayname(T_d(when), 0));

		if (admin & AD_MIN)
		{
			printf(LF "<tr><td align=right bgcolor="GREEN" valign=top>Base HTML fragment:"
				   LF "<br/></td><td bgcolor="GREEN"><textarea name=\"ev_html\" rows=8 cols=60 wrap=virtual>");

			if (mainrep) encapsulate(mainrep, stdout);

			printf("</textarea></td></tr>" LF "<tr><td bgcolor="GREEN" align=right>"
				   LF "<input type=\"checkbox\" name=\"ev_shead\" value=\"set\"%s> &nbsp; Show Headline HTML:"
				   LF "</td><td bgcolor="GREEN"><textarea name=\"ev_head\" rows=3 cols=60 wrap=virtual>",
				   ((f & EF_ShowHeadline) ? " checked" : 0));

			if (headline) encapsulate(headline, stdout);

			printf("</textarea></td></tr>"
				   LF "<tr><td bgcolor="GREEN" align=right>Event flags:</td><td bgcolor="GREEN">"
				   LF "<input type=\"checkbox\" name=\"ev_sopen\" value=\"set\"%s> &nbsp; Open &nbsp;"
				   LF "<input type=\"checkbox\" name=\"ev_sinv\" value=\"set\"%s> &nbsp; Invisible &nbsp;"
				   LF "<input type=\"checkbox\" name=\"ev_sunsc\" value=\"set\"%s> &nbsp; Ignore event</td></tr>",
				   ((f & EF_Open) ? " checked" : ""),
				   ((f & EF_Invisible) ? " checked" : ""),
				   ((f & EF_Ignore) ? " checked" : ""));
		}	

		printf(LF "<tr bgcolor="BLUE"><td align=right>By player:</td><td><b><a href=\"plinfo.cgi?id=%s\">%s</a></b>"
			   LF "(Status at time of event: <b>%s%s</b>)",
			   p1ptr->id, p1ptr->name, ((p1f & PF_Special) ? "" : (p1f & PF_Dead) ? "Dead, " : "Live, "),
			   pltypes[PF_type(p1f)]);

		if (admin & AD_ASS_PSEUD)
		{
			puts(LF "</td></tr><tr bgcolor="BLUE"><td align=right>Under pseudonym:</td><td>");

			if (!(p1f & PF_Special))
			{
				puts("<select name=\"ev_p1psu\" size=1>"
					 "<option value=\"\">New pseudonym:</option>"
					 "<option value=\"\">-----(or)-----");

				for (i = 0, pseudp = p1sp->allpseuds; (pseudp); pseudp = pseudp->link_pl)
				{
					i = (i) ? 2 : ((p1ps) && (!strcasecmp(pseudp->text, p1ps))) ? 1 : 0;
					printf("</option>" LF "<option value=\"");
					encapsulate(pseudp->text, stdout);
					printf("\"%s>", (i == 1) ? " selected" : "");
					encapslim(pseudp->text, stdout, MAX_PseudDisplay);
				}

				puts("</option></select>" LF "<br/><input type=\"text\" name=\"ev_p1psn\" value=\"");
				if ((i == 0) && (p1ps)) encapsulate(p1ps, stdout);
				printf("\"> &nbsp;<input type=\"checkbox\" name=\"ev_p1psd\" value=\"set\"%s> &nbsp; "
					   "Make default pseudonym", ((f & EF_SetPseudonym) ? " checked" : ""));
			}
			else
			{
				puts("<input type=\"text\" name=\"ev_p1psu\" value=\"");
				if (p1ps) encapsulate(p1ps, stdout);
				puts("\">");
			}
		}

		if (admin & AD_MIN)
		{
			u32b d = T_d(when);

			printf("</td></tr><tr bgcolor="BLUE"><td align=right>Scoring bonus:</td>"
				   LF "<td><input type=\"text\" name=\"ev_bonus\" size=8 maxlength=8 value=\"%+.5f\"></td></tr>",
				   (bonus / ((double) (1 << SCORE_SHIFT))));

			puts("<tr bgcolor="BLUE"><td align=right>Inactivity deadline:</td>" LF
				 "<td><select name=\"ev_boost\" size=1><option value=\"0\">No change</option>");
	
			for (i = 1; i < 20; i++) printf("<option value=\"%d\" %s>(%d days) %s</option>",
											i, ((i == boost) ? ("selected") : ("")), i, fdayname(d + i, 0));
	
			p = (!(p1f & PF_Wanted)) ? "unwanted" : (p1f & PF_Incompetent) ? "incompetent" : "wanted";

			printf("</select></td></tr>" LF "<tr bgcolor="BLUE"><td align=right>Criminality:</td>"
				   LF "<td> Currently %s. &nbsp; <input type=\"checkbox\"%sname=\"ev_swant\" value=\"set\">"
				   LF "&nbsp; Make %s. &nbsp;" LF "<input type=\"text\" name=\"ev_comment\" value=\"",
				   p, (findval("ev_swant", 0)) ? " checked " : " ",
				   (p1f & PF_Wanted) ? "unwanted" : "wanted");

			if (comment) encapsulate(comment, stdout);
			fputs("\">" LF "<input type=\"hidden\" name=\"ev_owant\" value=\"", stdout);
			if (ex1p->comment) encapsulate(ex1p->comment, stdout);
			fputs("\">", stdout);
		}

		printf("</td></tr>"LF"<input type=\"hidden\" name=\"ev_on\" value=\"%s\">"
			   LF "<tr bgcolor="RED"><td align=right>On player:</td>"
			   LF "<td><b><a href=\"plinfo.cgi?id=%s\">%s</a></b>"
			   LF "(Status at time of event: <b>%s%s</b>)",
			   p2ptr->id, p2ptr->id, p2ptr->name, ((p2f & PF_Special) ? "" : (p2f & PF_Dead) ? "Dead, " : "Live, "),
			   pltypes[PF_type(p2f)]);

		if (admin & AD_TAR_PSEUD)
		{
			puts(LF "</td></tr><tr bgcolor="RED"><td align=right>Under pseudonym:</td><td>");

			if (!(p2f & PF_Special))
			{
				puts("<select name=\"ev_p2psu\" size=1>"
					 "<option value=\"\">New pseudonym:</option>"
					 "<option value=\"\">-----(or)-----");

				for (i = 0, pseudp = p2sp->allpseuds; (pseudp); pseudp = pseudp->link_pl)
				{
					i = (i) ? 2 : ((p2ps) && (!strcasecmp(pseudp->text, p2ps))) ? 1 : 0;
					printf("</option>" LF "<option value=\"");
					encapsulate(pseudp->text, stdout);
					printf("\"%s>", (i == 1) ? " selected" : "");
					encapslim(pseudp->text, stdout, MAX_PseudDisplay);
				}

				puts("</option></select>" LF "<br/><input type=\"text\" name=\"ev_p2psn\" value=\"");
				if ((i == 0) && (p2ps)) encapsulate(p2ps, stdout);
				printf("\"> &nbsp;<input type=\"checkbox\" name=\"ev_p2psd\" value=\"set\"%s> &nbsp; "
					   "Make default pseudonym", ((f & EF_SetP2Pseudonym) ? " checked" : ""));
			}
			else
			{
				puts("<input type=\"text\" name=\"ev_p2psu\" value=\"");
				if (p2ps) encapsulate(p2ps, stdout);
				puts("\">");
			}
		}

		/* Show auto-resurrection stuff */
		if ((admin & AD_MIN) && !(p2f & PF_Special) && (f & EF_Kill))
		{
			if (!(tmptime = addhours(when, reshours(when)))) erf(ERF_BadTimeVal, 26);

			printf(LF "</td></tr><tr bgcolor="RED"><td align=right>Effects:</td>"
				   LF "<td><table border=0 cellspacing=0 cellpadding=0><tr><td>"
				   LF "<input type=\"checkbox\" name=\"ev_sinvres\" value=\"set\"%s> &nbsp;</td>"
				   LF "<td>%sesurrect <b>victim</b> at "TDISP" %s (reverse normal resurrection discipline).",
				   (f & EF_InvertAutoRes) ? " checked" : "", (OPT_AutoRes) ? "Don't r" : "R",
				   Thm(tmptime), fdayname(T_d(tmptime), 0));

			if (tmptime >= TIME_AutoResLimit) puts("<br/><b>Has no effect: event time is past normal"
												   " resurrection deadline.</b>");
			puts("</td></tr></table>");
		}
		
		/* Targetting */
		if ((admin & AD_MIN) && (retargetting))
		{
			printf("</td></tr>" LF "<tr bgcolor="RED"><td align=right>Retargeting:</td>"
				   LF "<td><table cellpadding=0 cellspacing=2 border=0 cols=2>"			   
				   LF "<tr><td><input type=\"radio\"%sname=\"ev_retarg\" value=\"none\">&nbsp;</td>"
				   LF "<td>Don't do any target/assassin allocation or deallocation</td></tr>"
				   LF "<tr><td><input type=\"radio\"%sname=\"ev_retarg\" value=\"detarg_vict\">&nbsp;</td>"
				   LF "<td>Only deallocate all of the victim's targets and assassins:</td></tr>",
				   (retargetting == 1) ? " checked " : " ", (retargetting == 2) ? " checked " : " ");

			if (i = killretarg(retargetting, p2sp, p2ptr, (retargetting >= 3), plist)) erf(i, 27);
		}

		puts(LF "</td></tr><tr><td></td><td><input value=\"Add event\" type=\"submit\" name=\"sub_ev\"></td></tr>"
			 LF "</form></table></p><hr/><p>"FOOTER"<hr/></p>"
			 LF "</body></html>");

		return 0;
	}

	/* Secondary call - prepare final event information */
	else if (mode == 1)
	{
		/* Retain old errors as warnings */
		oldbad = bad; bad = 0;

		/* Basic flags */
		f |= (EF_Assassination|EF_News);

		/* Player types */
		if (p2ptr->flags & PF_Special)   f |= (EF_OnSpecial);

		#if 0
		if (p2ptr->flags & PF_IllegalNP) f |= (EF_OnIllegalNP);
		if (p2ptr->flags & PF_LegalNP)   f |= (EF_OnLegalNP);
		#endif

		/* Change comment? */
		if ((comment) && (admin & AD_MIN) && (!streq(comment, NSTR(ex1p->comment)))) f |= EF_Wantedness;
	}

	/* Add comment event */
	else if (mode == 2)
	{
		/* Basic flags */
		f |= (EF_Comment|EF_News);
	}

	/* Add new [mode = (1) game, (2) comment] event */
	if (mode <= 2)
	{
		struct editev tev;
		unsigned long tfl;

		/* Deadline boost */
		if (boost) f |= EF_SetInactTimer;

		/* Initialise timer */
		if (!(tmptime = addhours(when, 24 * boost))) erf(ERF_BadTimeVal, 28);

		/* Debug */
		fprintf(stderr, "%li:   add(%d) f=&%lx p1id=\"%s\" p2id=\"%s\" time="TDEBfmt" p1f=&%lx p2f=&%lx\n",
				uniqid, mode, f, p1ptr->id, (p2ptr) ? (p2ptr->id) : "xxx", TDEBparm(when), p1f, p2f);

		/* No report id yet */
		*(newrep.rid = ridbuf) = '\0';

		/* Main report */
		newrep.pid   = authp->id;
		newrep.pptr  = 0;
		newrep.text  = mainrep;
		newrep.pseud = 0;
		newrep.link_sub = 0;
		newrep.link_all = 0;
		newrep.flags = (RF_Main|RF_Priv|RF_Normal) | ((mode == 2) ? RF_Comment : (admin & AD_MIN) ? 0 : RF_AutoGen);
		newrep.fpos = -1L;
		newrep.uptim = thetime;

		/* Build initial report fragment  */
		if ((i = addreport(&newrep, 0, 0, 0, 0, *statbuf_pos, statbuf_end)) <= 0)
		{
			/* Retain pointer to filled in report struct */
			rid = (rlist = newrptr = &newrep)->rid;
		}

		/* Addition failed */
		else erf(i, 20);

		/* Debug */
		fprintf(stderr, "%li:   add((%d)) f=&%lx p1id=\"%s\" p2id=\"%s\" time="TDEBfmt" p1f=&%lx p2f=&%lx\n",
				uniqid, mode, f, p1ptr->id, (p2ptr) ? (p2ptr->id) : "xxx", TDEBparm(when), p1f, p2f);

		/* Initialise (add the report id, automatically allocate new ref) */
		init_editev_struct(&nev, 0, 0UL, 0, 0, when, rid);

		/* New event structure */
		nev.npid        = p1ptr->id;
		nev.npseud      = p1ps;
		nev.nvict_pid   = (p2ptr) ? (p2ptr->id) : 0;
		nev.nvict_pseud = p2ps;
		nev.nnews_date  = thetime;
		nev.nnews_head  = headline;
		nev.nwanted     = comment;
		nev.nbonusp		= &bonus;
		nev.nflsp       = &f;
		nev.ntimp       = tmptime;

		/* Separate targetting event */
		if ((retargs) || (untargs))
		{
			/* Initialise */
			init_editev_struct(&tev, 0, 0UL, 0, 0, thetime, 0);

			/* Event flags */
			*(tev.nflsp = &tfl) = EF_Targetting;

			/* By player victem */
			tev.npid = p2ptr->id;

			/* Time of event is now */
			tev.ntimp = thetime;

			/* Reallocation of targets */
			tev.nretarg = retargs;
			tev.nuntarg = untargs;

            /* Link up */
			nev.next = &tev;
		}

		/* No retargetting */
		nev.nretarg = nev.nuntarg = 0;

		/* Form comment */
		if ((statbuf_end - (p = *statbuf_pos)) > (96+MAX_id))
		{
			/* Comment string */
			j = sprintf(p, " Added at realtime "TDEBfmt" by auth \"%.*s\":", TDEBparm(thetime), (int)MAX_id, authp->id);

			/* Check string length */
			if ((j < 20) || (j > (96+MAX_id))) p = 0;

			/* Advance buffer position */
			else *statbuf_pos += 1+j;
		}

		/* No comment */
		else p = 0;

		/* Add new event */
		if (i = editevents(0, &nev, statbuf_pos, statbuf_end, p, 0)) erf(i, 21);

		/* Get event struct */
		eptr = nev.ep;

		/* old add new event code */
		/*if ((i = newevent(&nevp, &eptr, rid, 0, &when, statbuf_pos, statbuf_end)) > 0) erf(i, 21);*/

		/* Run the scoring program later and regenerate the news */
		rescore |= RESC_Scoring|RESC_ERR_Scoring|RESC_MkNews|RESC_ERR_MkNews;

		/* Note success */
		good |= GOOD_event_EventAdded;

		/* Deal with retargetting */
		/*if ((findval("ev_retarg", 0)) && (j = genretarget(2, 6))) return (j);*/
	}

	/* Acquire event information for pre-existing events [mode >= 3] */
	else if (mode >= 3)
	{
		player_p ppt;

		/*fprintf(stderr, "%li:   flob %p\n", uniqid, plist);*/
		/*for (ppt = plist; (ppt); ppt = ppt->link) fprintf(stderr, "%li:   goo %p \"%s\"\n", uniqid, ppt, ppt->id);*/

		/* Allowed to do anything other than look? */
		if ((mode == 4) && (authp->flags & PF_Public)) erf(ERF_CantPubEditEv, 26);

		/* Event reference passed as parameter? */
		ref = ((mode == 3) && (strparm) && (*strparm)) ? strparm : 0;

		/* Require event reference */
		if ((!ref) && !(ref = findval("ev_ref", 0))) erf(ERF_UnknownRef, 24);

		/* Lookup event */
		if (rc = getevinfo(ref, &eptr, &rlist, statbuf_pos, statbuf_end, 0)) erf(rc, 25);

		/* Not found */
		if (!eptr) erf(ERF_UnknownRef, 26);

		/* Event flags */
		f = eptr->flags;

		/* Time of event */
		when = eptr->when;

		/* Debug */
		/*fprintf(stderr, "%li:   genfef(%d) f=&%lx ref=\"%s\" epid=\"%s\" %p %p\n",
				uniqid, mode, f, eptr->ref, eptr->id, *statbuf_pos, eptr->ref);*/

		/* Check authentication permissions */
		if (!(admin & AD_MIN) && ((f & (EF_Ignore|EF_Invisible)) || !(f & EF_Open))) erf(ERF_CantEditEv, 27);

		/* Event participants pids */
		ipids[0] = eptr->id;
		ipids[1] = (f & EF_Assassination) ? (eptr->vict_pid) : 0;
		ipids[2] = 0;

		/* Pseudonyms */
		p1ps = (f & EF_MASK_Pseudonym) ? (eptr->pseud) : 0;
		p2ps = (f & EF_Assassination) ? (eptr->vict_pseud) : 0;

		fprintf(stderr, "%li:   genfef(%d)a 1id=\"%s\" 2id=\"%s\" p1ptr=%p p2ptr=%p by=\"%s\" on \"%s\"\n",
				uniqid, mode, STR(ipids[0]), STR(ipids[1]), p1ptr, p2ptr, eptr->id, STR(eptr->vict_pid));

		/* Possible extra permissions if authentication matches assassin */
		if (streq(authp->id, ipids[0]))
		{
			p1ptr = authp;
			if (f & EF_Assassination)
			{
				admin |= AD_ASS;
				#if OPT_ChangePseud
				admin |= AD_ASS_PSEUD;
				#endif
			}			
		}

		/* Lookup player 1 */
		else for (p1ptr = plist; (p1ptr); p1ptr = p1ptr->link)
		{
			fprintf(stderr, "%li:    >> test \"%s\"\n", uniqid, p1ptr->id, ipids[0]);
			if (!strcmp(p1ptr->id, ipids[0])) break;
		}

		/* Unknown player 1 */
		if (!p1ptr) erf(ERF_MissingPid, 29);

		/* *Current* Player flags */
		p1f = p1ptr->flags;

		/* No player 2 */
		if (!ipids[1]) p2ptr = 0;

		else
		{
			/* Extra permissions if authentication matches target */
			if (!strcmp(ipids[1], authp->id))
			{
				p2ptr = authp;
				if (f & EF_Assassination)
				{
					admin |= AD_TAR;
					#if OPT_ChangePseud
					admin |= AD_TAR_PSEUD;
					#endif
				}
			}
	
			/* Quick check: p2 is p1 */
			else if (!strcmp(ipids[0], ipids[1])) p2ptr = p1ptr;
		
			/* Lookup player 2 */
			else for (p2ptr = plist; (p2ptr); p2ptr = p2ptr->link) if (!strcmp(p2ptr->id, ipids[1])) break;
	
			/* Unknown player 2 */
			if (!p2ptr) erf(ERF_MissingPid, 30);

			/* *Current* Player flags */
			p2f = p2ptr->flags;
		}

		/*fprintf(stderr, "%li:   genfef(%d)b 1id=\"%s\" 2id=\"%s\" p1ptr=%p p2ptr=%p\n",
				uniqid, mode, STR(ipids[0]), STR(ipids[1]), p1ptr, p2ptr);*/

		/*fprintf(stderr, "%li:   genfef(%d) f=&%lx pids=\"%s\",\"%s\" ref=\"%s\" epid=\"%s\" %p,%p pptrs=%p,%p\n",
				uniqid, mode, f, STR(p1ptr->id), (p2ptr) ? STR(p2ptr->id) : "xxx", STR(eptr->ref), STR(eptr->id),
				eptr->ref, eptr->id, p1ptr, p2ptr);*/

		/* Scoring bonus (if any) */
		bonus = (f & EF_ScoreBonus) ? (eptr->bonus) : 0;

		/**** frag_0 was here ****/
	}

	/* Paranoia: Must have valid event pointer */
	if (!eptr) erf(ERF_InternalError, 40);

	/* Paranoia: Must have correct p1ptr player pointer */
	if ((!p1ptr) || (strcmp(eptr->id, p1ptr->id))) erf(ERF_InternalError, 41);

	/* Paranoia: p2ptr non-null if and only if event is an assassination */
	if ((f & EF_Assassination) && ((!p2ptr) || (strcmp(eptr->vict_pid, p2ptr->id)))) erf(ERF_InternalError, 42);
	if (!(f & EF_Assassination) && (p2ptr)) erf(ERF_InternalError, 43);

	/* Period in which event occurred */
	perp = per(eptr->when);

	/* Distinguish between acting on addition and editing of reports by presence of report id string */
	if ((mode == 7) && (rid = findval("rep_rid", 0))) mode = 8;

	/* But when preparing reports for addition/editing, report id passed as parameter to routine */
	else rid = (mode == 6) ? strparm : 0;

	/* Acquire existing report information */
	if ((mode == 6) || (mode == 8))
	{
		if (mode == 6) fprintf(stderr, "%li:   request to edit rep rid=\"%s\" for eref=\"%s\"\n",
							   uniqid, rid, eptr->ref);

		/* Report owner not yet known */
		p3ptr = 0;

		/* Does report belong to this event? */
		for (rptr = rlist; (rptr); rptr = rptr->link_all) if (!strcmp(rptr->rid, rid)) break;

		/* No, so disallow edit */
		if (!rptr) bad |= BAD_event_RepNotFromEvent;

		/* Acquire report information */
		else
		{
			/* Make sure chosen report text will be loaded */
			rptr->flags |= RF_LoadText;

			/* Acquire report information */
			if (rc = repfile(&rlist, plist, statbuf_pos, statbuf_end, 0, 1+2)) erf(rc, 44);

			/* All good so far */
			p3ptr = (newrptr = rptr)->pptr;

			/* Note that report information has been found */
			gotreps = 1;
		}
	}

	/* No player specified for report? */
	if ((mode >= 5) && (mode <= 8) && (!p3ptr)) erf(ERF_MissingPid, 45);

	/* Optimisation: duplicate players */	
	if ((p1ptr) && (p3ptr == p1ptr)) { p3sp = p1sp; p3f = p1f; ex3p = ex1p; }

	/* Optimisation: duplicate players */
	else if ((p2ptr) && (p3ptr == p2ptr)) { p3sp = p2sp; p3f = p2f; ex3p = ex2p; }

	/* Need player information at time of event? */
	if (((admin & AD_INV) && ((!p1sp) || (!p2sp))) || ((p3ptr) && (!p3sp)))
	{
		/* Interesting players */
		ipids[0] = (p1sp) ? "" : (p1ptr->id);
		ipids[1] = (p2ptr && !p2sp) ? (p2ptr->id) : "";
		ipids[2] = (p3ptr && !p3sp) ? (p3ptr->id) : "";
		ipids[3] = 0;

		/* Clear output array */
		psidx[0] = psidx[1] = psidx[2] = 0;

		fprintf(stderr, "%li   about to evolve \"%s\" \"%s\" \"%s\"\n", uniqid, ipids[0], ipids[1], ipids[2]);

		/* Evolve some relevent players - determine participants' states at time of event */
		if (rc = evolve(EVOLVE_FuncEndCond|EVOLVE_FuncExhaust|EVOLVE_TimersToLimit|EVOLVE_MatchAll|EVOLVE_Prior|
						EVOLVE_Comments|EVOLVE_AllPseuds|EVLadd_AllPseuds|EVOLVE_InheritPseud|EVLadd_Comments,
						ipids, &psidx, &pslist, 0,0,0,0, eptr->ref, 0,0,0,
						*statbuf_pos, statbuf_end, statbuf_pos, 0,0,0, genfef_hook, 0, 0)) erf(rc, 46);

		fprintf(stderr, "%li   did evolution => %p %p %p\n", uniqid, psidx[0], psidx[1], psidx[2]);

		/* Existences at time of event */
		if ((*ipids[0]) && ((!(p1sp = psidx[0])) || !(ex1p = p1sp->ext))) erf(ERF_InternalError, 47);
		if ((*ipids[1]) && ((!(p2sp = psidx[1])) || !(ex2p = p2sp->ext))) erf(ERF_InternalError, 48);
		if ((*ipids[2]) && ((!(p3sp = psidx[2])) || !(ex3p = p3sp->ext))) erf(ERF_InternalError, 49);

		/* Player flags at time of event */
		if (*ipids[0]) p1f = (p1ptr->flags & ~(PF_MASK_Scoring)) | (ex1p->fl & PF_MASK_Scoring);
		if (*ipids[1]) p2f = (p2ptr->flags & ~(PF_MASK_Scoring)) | (ex2p->fl & PF_MASK_Scoring);
		if (*ipids[2]) p3f = (p3ptr->flags & ~(PF_MASK_Scoring)) | (ex3p->fl & PF_MASK_Scoring);
	}

	/* Edit existing event: be careful to distinguish between old and new parameters */
	if (mode == 4)
	{
		/* Check permissions */
		if (!(admin & AD_INV)) erf(ERF_CantEditEv, 80);

		/* Safety: which options are present in the input data? */
		testf = (p = findval("ed_testf", 0)) ? strtoul(p, 0, 16) : 0;

		/* Admin: Change general options: f = eptr->flags at this point */
		if (admin & AD_MIN)
		{
			/* Option: Deallocate P1 targets */
			if (testf & TST_DetargP1) f = (findval("ed_sdetarg_ass", 0)) ? (f | EF_Detarg) : (f & ~EF_Detarg);

			/* Option: Ignore event */
			if (testf & TST_Ignore) f = (findval("ed_sunsc", 0)) ? (f | EF_Ignore) : (f & ~EF_Ignore);
	
			/* Option: Invisible */
			if (testf & TST_Invisible) f = (findval("ed_sinv", 0)) ? (f | EF_Invisible) : (f & ~EF_Invisible);
	
			/* Option: Open */
			if (testf & TST_Open) f = (findval("ed_sopen", 0)) ? (f | EF_Open) : (f & ~EF_Open);

			/* Option: Comment */
			if (testf & TST_Comment) f = (findval("ed_scomm", 0)) ? (f | EF_Comment) : (f & ~EF_Comment);

			/* Option: Annulled */
			if (testf & TST_Annulled) f = (findval("ed_sanul", 0)) ? (f | EF_Annulled) : (f & ~EF_Annulled);

			/* Scoring bonus */
			if (p = findval("ed_bonus", 32+1))
			{
				/* Set or clear flag according nullity of bonus value */
				f = (bonus = rint((1 << SCORE_SHIFT) * strtod(p, 0))) ? (f | EF_ScoreBonus) : (f & ~EF_ScoreBonus);
			}
		}

		/* Admin: Change assassination options */
		if ((eptr->flags & EF_Assassination) && (admin & AD_MIN))
		{
			/* Option: Detarget P2 */
			if (testf & TST_DetargP2)
			{
				f = (findval("ed_sdetarg_vict", 0)) ? (f | EF_DetargP2) : (f & ~EF_DetargP2);
			}

			/* Option: Invert normal resurrection discipline */
			if (testf & TST_InvertAutoRes)
			{
				f = (findval("ed_sinvres", 0)) ? (f | EF_InvertAutoRes) : (f & ~EF_InvertAutoRes);
			}

			/* Option: Invert normal resurrection discipline */
			if (testf & TST_NoDangerBonus)
			{
				f = (findval("ed_snodbonus", 0)) ? (f | EF_NoDangerBonus) : (f & ~EF_NoDangerBonus);
			}
		}		

		/* Change parameters that player 1 can edit in assassination events */
		if (admin & AD_ASS_PSEUD)
		{
			/* Check for presence of new pseudonym, and check for duplication */
			if ((l = checkpseud("ed_p1psu", "ed_p1psn", "ed_p1psd", &r, &p, p1ptr->id, pslist)) != BAD_event_PseudAss)
			{
				/* Invalid: note reason */
				if (l) bad |= l;

				/* Valid, so use */
				else p1ps = p;
			}

			/* Option: Set as default pseudonym */
			if (testf & TST_SetP1Pseudonym) f = (f & ~(EF_SetPseudonym|EF_AddPseudonym)) | r;
		}

		/* Change parameters that victim or assassin can edit in assassination events */
		if ((eptr->flags & EF_Assassination) && (admin & AD_INV) && (p = findval("ed_outcome", 0)))
		{
			/* New outcome valid? */
			if (!(i = strcasecmp(p, "kill")) || (!strcasecmp(p, "fail")))
			{
				/* XXX XXX --- Should check that that the new value is legal */

				/* Determine new flag value */
				f = (f & (~EF_Kill)) | ((i) ? 0 : EF_Kill);
			}

			/* Invalid value for outcome */
			else bad |= BAD_event_Type;
		}

		/* Change other parameters that victim can edit in assassination events */
		if ((p2ptr) && (admin & AD_TAR_PSEUD))
		{
			/* Check for presence of new pseudonym, and check for duplication */
			if ((l = checkpseud("ed_p2psu", "ed_p2psn", "ed_p2psd", &r, &p, p2ptr->id, pslist)) != BAD_event_PseudAss)
			{
				/* Invalid: note reason */
				if (l) bad |= (l << 2);

				/* Valid, so use */
				else p2ps = p;
			}

			/* Option: Set as default pseudonym */
			if (testf & TST_SetP2Pseudonym) 
			{
				f = (r & EF_SetPseudonym) ? (f | EF_SetP2Pseudonym) : (f & ~EF_SetP2Pseudonym);
			}
		}

		/* Debug */
		fprintf(stderr, "%li:   edit event \"%s\" tstf=%lX oldf=&%lX newf=&%lX bad=%lu p2ps=\"%s\"\n",
				uniqid, eptr->ref, testf, eptr->flags, f, bad, p2ps);

		/* Event details okay */
        if (!bad)
        {
			/* Edit event by ref, keep prev ref, not adding any reports */
			init_editev_struct(&nev, eptr->ref, EDEV_Edit, 0, 0, 0, 0);

			/* Preserve previous event structure */
			oldeptr = eptr;

			/* Changes to event structure */
			nev.npseud      = p1ps;
			nev.nvict_pseud = p2ps;
			nev.nnews_date  = thetime;
			nev.nbonusp		= &bonus;
			nev.nflsp       = &f;

			/* Form comment */
			if ((statbuf_end - (p = *statbuf_pos)) > (96+MAX_id))
			{
				/* Comment string */
				j = sprintf(p, " Edited at realtime "TDEBfmt" by auth \"%.*s\":",
							TDEBparm(thetime), (int)MAX_id, authp->id);

				/* Check string length */
				if ((j < 21) || (j > (96+MAX_id))) p = 0;

				/* Advance buffer position */
				else *statbuf_pos += 1+j;
			}

			/* No comment */
			else p = 0;
	
			/* Add new event */
			if (i = editevents(0, &nev, statbuf_pos, statbuf_end, p, 0)) erf(i, 81);
	
			/* Old event edit call: Edit the event; 'eptr' will refer to edited event details */
			/*if ((i = newevent(&nevp, &eptr, 0, ref, 0, statbuf_pos, statbuf_end)) > 0) erf(i, 81);*/

			/* Paranoia: require valid edited event struct */
			if (!(eptr = nev.ep)) erf(ERF_Paranoia, 82);

			/* If sucessful, downgrade back to event overview */
			mode = 3;
		
			/* Run the scoring program later and regenerate the news */
			rescore |= RESC_Scoring|RESC_ERR_Scoring|RESC_MkNews|RESC_ERR_MkNews;

			/* Note success */
			good |= GOOD_event_EventEdited;
		}
	}

	/********************************************************************************************************
	*** At this point, we are dealing with pre-existing [mode >= 3] or recently added events [mode <= 2]. ***
	*** The event structure (eptr) is up to date; no change will be made later in this function except as ***
	*** required to add new reports or update the news datestamp (not yet implemented).                   ***
	********************************************************************************************************/

	/* Deal with editing/adding reports */
	if ((mode >= 5) && (mode <= 8))
	{
		/* Paranoia: Must have done the evolution */
		if (!ex3p) erf(ERF_InternalError, 60);

		/* New report: initialise */
		if ((mode == 5) || (mode == 7))
		{
			/* Setup default report struct */
			newrptr = &newrep;
	
			/* Main report */
			newrep.rid = rid = 0;
			newrep.pid = p3ptr->id;
			newrep.text = mainrep = OPT_FragNewRep;
			newrep.pseud = 0;
			newrep.uptim = thetime; 
			newrep.pptr   = p3ptr;
			newrep.flags   = RF_Listed | ((eptr->flags & EF_Comment) ? (RF_Comment) : 0);
			newrep.link_all = 0;
			newrep.link_sub = 0;
		}

		/* Admin can mess up anything */
		if (!(bad) && !(admin & AD_MIN))
		{
			/* Players can only report as themselves or edit their own reports */
			if (p3ptr != authp)
			{
				suggestid = p3ptr->id;
				erf((authp->flags & PF_Public) ? ERF_CantPubDoReps : ERF_NotRepOwner, 61);
			}
		
			/* Must be the correct sort of player */
			if ((p3ptr->flags & PF_Special) || !(p3ptr->flags & PF_Playing)) bad |= BAD_event_RepOwner;
		}

		/* Warnings: downgrade back to event overview */
		if ((bad) && ((mode == 5) || (mode == 6))) mode = 4;

		/* Prepare and check reports before addition/edit */
		if (((mode == 7) || (mode == 8)) && (!bad))
		{
			if (mode == 7) fprintf(stderr, "%li:   request to add rep bad=%lu p3id=\"%s\" eref=\"%s\"\n",
								   uniqid, bad, p3ptr->id, eptr->ref);

			if (mode == 8) fprintf(stderr, "%li:   request to edit rep \"%s\" bad=%lu p3id=\"%s\" eref=\"%s\"\n",
								   uniqid, rid, bad, p3ptr->id, eptr->ref);
	
			/* Additional admin flags */
			if (admin & AD_MIN)
			{
				/* Override default settings */
				newrptr->flags &= ~(RF_Main|RF_Comment|RF_Listed|RF_Hidden|RF_AutoGen|RF_NoIntro|RF_Priv|RF_Repro);

				/* Main report flag */
				newrptr->flags |= (findval("rep_omain", 0)) ? (RF_Main) : (0);

				/* Report may take higher priviledges */
				newrptr->flags |= (findval("rep_opriv", 0)) ? (RF_Priv) : (0);

				/* Report is a comment */
				newrptr->flags |= (findval("rep_ocomment", 0)) ? (RF_Comment) : (0);

				/* Report is not to be publically listed? */
				newrptr->flags |= (findval("rep_ounlisted", 0)) ? (0) : (RF_Listed);

				/* Report is not to be published? */
				newrptr->flags |= (findval("rep_ohidden", 0)) ? (RF_Hidden) : (0);

				/* Report was autogenerated */
				newrptr->flags |= (findval("rep_oautogen", 0)) ? (RF_AutoGen) : (0);

				/* Report appearance */
				r = strtoul(findval("rep_repro", 1), 0, 10);

				/* Set flags appropriately */
				if (r & 1) newrptr->flags |= RF_0_Repro;
				if (r & 2) newrptr->flags |= RF_1_Repro;

				/* Force report introduction to be ommitted if automatically included */
				newrptr->flags |= (findval("rep_onoint", 0)) ? (RF_NoIntro) : (0);
			}

			/* Admin can edit reports without timestamping, otherwise */
			if (!(admin & AD_MIN) || (findval("rep_ostamp", 0))) newrptr->uptim = thetime;

			/* Report text */
			if (!(newrptr->text = mainrep = findval("rep_html", 0+4))) bad |= BAD_event_RepText;

			/* Report pseudonym */
			if ((OPT_ChangePseud) || (admin & AD_MIN))
			{
				if (p = findval("rep_psu", 32+1))
				{
					/* Standard pseudonym */
					if (!strcmp(p, "-")) newrptr->pseud = 0;
	
					/* Named pseudonym */
				 	else
				 	{
						/* Chosen from list, or typed in? */
				 		p = (*p) ? p : findval("rep_psn", 1);
	
						/* Check for duplication */
						if (l = checkpseud(0, p, 0, 0, &newrptr->pseud, p3ptr->id, pslist)) bad |= (l << 4);
				 	}
				}
			}
	    }

		/* Warnings: downgrade back to "add/edit report" stage */
		if ((bad) && ((mode == 7) || (mode == 8))) mode -= 2;

		/* Add the report */
		else if (mode == 7)
		{
			/* No report id yet */
			*(newrep.rid = ridbuf) = '\0';
	
			/* Build initial report fragment */
			if ((i = addreport(&newrep, 0, 0, 0, eptr->ref, *statbuf_pos, statbuf_end) > 0)) erf(i, 62);

			/* Only alter report ids */
			init_editev_struct(&nev, eptr->ref, EDEV_OnlyRids, 0, 0, 0, rid = newrptr->rid);

			/* Form comment */
			if ((statbuf_end - (p = *statbuf_pos)) > (96+MAX_id+MAX_rid))
			{
				/* Comment string */
				j = sprintf(p, " Added report \"%.*s\" at realtime "TDEBfmt" by auth \"%.*s\":",
							(int)MAX_rid, rid, TDEBparm(thetime), (int)MAX_id, authp->id);

				/* Check string length is valid */
				if ((j < 51) || (j > (96+MAX_id+MAX_rid))) p = 0;

				/* Advance buffer pointer */
				else *statbuf_pos += 1+j;
			}

			/* No comment */
			else p = 0;

			/* Add new report to event */
			if (i = editevents(0, &nev, statbuf_pos, statbuf_end, p, 0)) erf(i, 63);

			/* Locate end of report list */
			for (rptr = rlist; (rptr); rptr = rptr->link_all) if (!rptr->link_all) break;

			/* Add new report to list */
			if (rptr) rptr->link_all = newrptr; else rlist = newrptr;

			/* Regenerate the news */
			rescore |= RESC_MkNews|RESC_ERR_MkNews;

			/* Note success */
			good |= GOOD_event_RepAdded;
		}

		/* Edit the report */
		else if (mode == 8)
		{
			/* Save changes */
			if ((i = addreport(newrptr, rid, 0, 0, eptr->ref, *statbuf_pos, statbuf_end)) > 0) erf(i, 64);

			/* Regenerate the news */
			rescore |= RESC_MkNews|RESC_ERR_MkNews;

			/* Note success */
			good |= GOOD_event_RepEdited;

			/* Drop back to common "report added/edited" mode */
			mode = 7;
		}
	}

	/****************************************************************************************
	*** At this point, all additions/changes to events and reports have been carried out. ***
	*** All that remains is to output the appropriate HTML form.                          ***
	****************************************************************************************/

	/* Show names? */
	if (authp->flags & PF_Observer) admin |= AD_OBS;

	/* Note which flag options are included on form */
	testf = 0;

	/* Event title */
	if (!(f & EF_News)) title = "Private event";
	else if (f & EF_Comment) title = "Comment event";
	else if (f & EF_Assassination) title = (f & EF_Kill) ? "Sucessful kill" : "Unsucessful kill attempt";
	else title = "Miscellaneous event";

	/* Catch mistakes */
	p = "@@@@@@@@@@@@@@@@@@@@";

	printf(LF "<HTML><HEAD><TITLE>"GAME" Assassins' Automatic Umpire</TITLE></HEAD>"
		   LF "<BODY bgcolor=#000000 text=#FFFFDD link=#BBBBFF vlink=#FFCCFF alink=#FFFFFF>"
		   LF "<h1><hr/><font color="WHITE">Assassins' Guild, "GAME": Automatic Umpire</font><hr/></h1>"
		   LF "<p>Authenticated as <b>%s</b>%s.</p>"
		   LF "<form method=\"POST\" action=\""CGI_URL THIS_CGI"\">"
		   LF "<input type=\"hidden\" name=\"authname\" value=\"%s\">"
		   LF "<input type=\"hidden\" name=\"authpass\" value=\"%s\">"
		   LF "<input type=\"hidden\" name=\"ev_ref\" value=\"%s\">",
		   authp->name, ((authp->flags & PF_Admin) ? " (admin)" : ""), authp->id, authp->pass, eptr->ref);

	/* Notifications of success */
	if (good)
	{
			puts("<h3><img src=\""TICK_URL"\" align=left "TICK_SIZ" hspace=8>"
				 "<font color=#FFCCCC>");
			for (i = 0; (i < GOOD_event__MAX_IDX); i++) if (good & (1 << i)) printf(genfef_goodstuff[i], STR(rid));
			puts("</font><br clear=left></h3><hr/>");
	}

	/* Warnings */
	if (warn)
	{
		puts("<h3><img src=\""WARN_URL"\" align=left "WARN_SIZ" hspace=8>"
			 "<font color=#FFCCCC>Warnings:</font></h3><p><font color=#99DDFF>");

		for (i = 0; (i < WARN_event__MAX_IDX); i++) if (warn & (1UL << i)) puts(genfef_warnstuff[i]);

		puts("</font></p><hr/>");
	}

	/* Warnings from initial event addition */
	if (oldbad)
	{
		puts("<h3><img src=\""SKULL_URL"\" align=left "SKULL_SIZ" hspace=8>"
			 "<font color=#FFCCCC>Ignored Some Trivial Imperfections...</font></h3><p><font color=#99DDFF>");

		for (i = 0; (i < BAD_event__MAX_IDX); i++) if (oldbad & (1UL << i)) puts(genfef_badstuff[i]);

		puts("</font></p><p>I hope you knew what you were doing </p><hr/>");
	}

	/* Notifications of failure */
	if (bad)
	{
			puts("<h3><img src=\""CROSS_URL"\" align=left "CROSS_SIZ" hspace=8>"
				 "<font color=#FFCCCC>Changes Not Committed: Some reasons...</font></h3><p><font color=#99DDFF>");

			for (i = 0; (i < BAD_event__MAX_IDX); i++) if (bad & (1UL << i)) printf(genfef_badstuff[i], STR(rid));

			puts("</font><br clear=left></p><hr/>");
	}

	/* Event overview table header */
	printf(LF "<p><table cols=2 cellspacing=4 cellpadding=4>"
		   LF "<tr><td colspan=2 bgcolor=#AA0000>"
		   LF "<font color="WHITE">Event &quot;<b>%s</b>&quot; &mdash;"
		   LF "<b>%s</b> at <b>"TDISP" on %s</b> &mdash;"
		   LF "<b>on <a href=\"" NEWS_URL "news%s.html#%s\">%s</a></b></font>",
		   eptr->ref, title, Thm(when), fdayname(T_d(when), 0), perp->index, eptr->ref, perp->title);

	/*fprintf(stderr, "%li   baaaa\n", uniqid);*/

	/* Present form for adding/editing new reports */
	if ((mode == 6) || (mode == 5))
	{
		FILE* fp;

		/* Paranoia: suitable blank/previous report struct should by now have been defined */
		if (!newrptr) erf(ERF_InternalError, 70);

		/* Paranoia: mode 6 => non-null report id */
		if ((!(rid = newrptr->rid)) && (mode == 6)) erf(ERF_InternalError, 71);

		/* Paranoia: mode 5 => no report id */
		else if ((mode == 5) && (rid))
		{
			fprintf(stderr, "%li:   XXX - Mismatch between mode=%i and rid=\"%s\".\n", uniqid, mode, rid);
			rid = 0;
		}

		/* List BASIC assassination information (frag_5 used to be around here) */
		printevoverview(eptr, f, p1ptr, p2ptr, p1ps, p2ps, admin);

		/* Title for add report segment */
		if (!rid) puts("</td></tr><tr><td colspan=2 bgcolor=#AA0000>" LF "<font color="WHITE"><b>Adding new report</b>");

		/* Title for edit report report segment */
		else printf("</td></tr>" LF "<input type=\"hidden\" name=\"rep_rid\" value=\"%s\"><tr>"
					"<td colspan=2 bgcolor=#AA0000><font color="WHITE"><b>Editing report</b> &quot;<b>%s</b>&quot;",
					rid, rid);

		/* Player information */
		printf(LF "by <input type=\"hidden\" name=\"rep_by\" value=\"%s\">"
			   LF "<b><a href=\"plinfo.cgi?id=%s\">%s</a></b>.</font>",
		 	   p3ptr->id, p3ptr->id, p3ptr->name);

		/* Report pseuds */
		if ((OPT_ChangePseud) || (admin & AD_MIN))
		{
			puts(LF "</td></tr><tr><td bgcolor="GREEN" align=right>Under pseudonym:</td>" LF "<td bgcolor="GREEN">");
		
			/* Pseudonym selector */
			pseudselector(newrptr->pseud, 2, "rep_psu", "rep_psn", 0, ((p3sp) ? p3sp->allpseuds : 0));
		}

		/* Print time of last report edit */
		if (newrptr->uptim)
		{
			printf(LF "</td></tr><tr><td align=right bgcolor="GREEN">Last updated:</td>"
				   LF "<td bgcolor="GREEN"><b>"TDISP"</b> on %s&nbsp;&nbsp;",
			 	   Thm(newrptr->uptim), fdayname(T_d(newrptr->uptim), 0));
		}

		/* Last edit time unknown */
		else puts(LF "</td></tr><tr><td bgcolor="GREEN" align=right>Last updated:"
				  LF "</td><td bgcolor="GREEN"><b>Time and date not known</b>&nbsp;&nbsp;");

		/* Admin features */
		if (admin & AD_MIN)
		{
			/* Choose whether or not to edit report timestamp */
			if (rid) puts("<input type=\"checkbox\" name=\"rep_ostamp\" value=\"set\" checked>"
						  " &nbsp;Set&nbsp;new&nbsp;timestamp");
			
			/* Report flags */
			printf(LF "</td></tr><tr><td bgcolor="GREEN" align=right>Options:</td><td bgcolor="GREEN">"
				   LF "<input type=\"checkbox\" name=\"rep_omain\" value=\"set\"%s> &nbsp;Main &nbsp;"
				   LF "<input type=\"checkbox\" name=\"rep_opriv\" value=\"set\"%s> &nbsp;Priviledged &nbsp;"
				   LF "<input type=\"checkbox\" name=\"rep_ocomment\" value=\"set\"%s> &nbsp;Comment &nbsp;"
				   LF "<input type=\"checkbox\" name=\"rep_ounlisted\" value=\"set\"%s> &nbsp;Unlisted &nbsp;"
				   LF "<input type=\"checkbox\" name=\"rep_oautogen\" value=\"set\"%s> &nbsp;Automatically generated"
				   LF "</td></tr><tr><td bgcolor="GREEN" align=right>Appearance:</td><td bgcolor="GREEN">"
				   LF "<select name=\"rep_repro\" size=1><option value=\"0\"%s>Process normally</option>"
				   LF "<option value=\"1\"%s>Reproduce verbatim (don't process)</option>"
				   LF "<option value=\"2\"%s>Print raw text (don't process)</option>"
				   LF "%s</select> &nbsp;"
				   LF "<input type=\"checkbox\" name=\"rep_ohidden\" value=\"set\"%s> &nbsp;Invisible &nbsp;"
				   LF "<input type=\"checkbox\" name=\"rep_onoint\" value=\"set\"%s> &nbsp;No auto-intro &nbsp;",
				   (newrptr->flags & RF_Main) ? " checked" : "", (newrptr->flags & RF_Priv) ? " checked" : "",
				   (newrptr->flags & RF_Comment) ? " checked" : "", (newrptr->flags & RF_Listed) ? "" : " checked",
				   (newrptr->flags & RF_AutoGen) ? " checked" : "",
				   ((newrptr->flags & RF_Repro) == RF_Normal) ? " selected" : "",
				   ((newrptr->flags & RF_Repro) == RF_Verbatim) ? " selected" : "",
				   ((newrptr->flags & RF_Repro) == RF_Encapsulate) ? " selected" : "",
				   ((newrptr->flags & RF_Repro) >= RF_XXX_Repro) ? "<option value=\"3\" selected>XXX</option>" : "",
				   (newrptr->flags & RF_Hidden) ? " checked" : "", (newrptr->flags & RF_NoIntro) ? " checked" : "");
		}

		puts(LF "</td></tr><tr><td colspan=2 bgcolor="GREEN">");

		if (j = (fp = fopen(FILE_ReportGuide, "rb")) ? 1 : 0)
		{
			while ((i = getc(fp)) != EOF) if (EOF == putchar(i)) { j = 2; break; }
			fclose(fp);
			fp = 0;
		}
		if (j != 1) fprintf(stderr, "%li:   Error %d copying ReportGuide.\n", uniqid, j);

		fputs(LF "<p><textarea name=\"rep_html\" rows=16 cols=80 wrap=virtual>", stdout);

		if (newrptr->text) encapsulate(newrptr->text, stdout);

		puts("</textarea>" LF "</p></td></tr><tr><td colspan=2 bgcolor="GREEN">"
			 LF "<input value=\"Submit report...\" type=\"submit\" name=\"sub_rep\"></td></tr>"
			 LF "</table></p></body></html>");

		return 0;
	}

	/* List BASIC assassination information, don't allow editing */
	if (!(admin & AD_INV)) printevoverview(eptr, f, p1ptr, p2ptr, p1ps, p2ps, (admin & AD_OBS));

	/* Otherwise, print event overview/edit form */
	else
	{
		char* ct[2] = { "440000", "000044" };

		/* Table header */
		puts(LF "</td></tr><tr><td colspan=2 bgcolor="GREEN">");
	
		/* Admin: general event stuff */
		if (j = (admin & AD_MIN) ? 1 : 0)
		{
			printf(LF "<table width=100%% border=0 cols=2 bgcolor=#000000 cellpadding=4 cellspacing=2>"
			 	   LF "<tr><td bgcolor="GREEN">Options:</td><td bgcolor="RED">"
				   LF "<input type=\"checkbox\" name=\"ed_sinv\" value=\"set\"%s>&nbsp;&nbsp;Invisible &nbsp;"
				   LF "<input type=\"checkbox\" name=\"ed_sanul\" value=\"set\"%s>&nbsp;&nbsp;Annulled &nbsp;"
				   LF "<input type=\"checkbox\" name=\"ed_sunsc\" value=\"set\"%s>&nbsp;&nbsp;Ignore&nbsp;event &nbsp;"
				   LF "<input type=\"checkbox\" name=\"ed_scomm\" value=\"set\"%s>&nbsp;&nbsp;Comment &nbsp;"
				   LF "<input type=\"checkbox\" name=\"ed_sopen\" value=\"set\"%s>&nbsp;&nbsp;Open</td></tr></table><br/>",
				   ((f & EF_Invisible) ? " checked" : ""),
				   ((f & EF_Annulled) ? " checked" : ""),
				   ((f & EF_Ignore) ? " checked" : ""),
				   ((f & EF_Comment) ? " checked" : ""),
				   ((f & EF_Open) ? " checked" : ""));
	
			/* Note which options are available */
			testf |= (TST_Invisible|TST_Annulled|TST_Ignore|TST_Open|TST_Comment);
		}
	
		/* Print basic (p1) player details */
		printf(LF "<table width=100%% border=0 cols=3 bgcolor=#000000 cellpadding=4 cellspacing=2>"
			   LF "<tr><td bgcolor="GREEN"><b>%s:</b></td><td bgcolor=#%s>",
			   ((f & EF_Assassination) ? "Assassin" : "Player"), ct[j & 1]);

		/* Special players aren't linked to */
		if (p1ptr->flags & PF_Special) printf("<b>%s</b></td><td bgcolor=#%s align=right>&mdash;",
											  p1ptr->name, ct[j & 1]);
		
		/* Show name and plinfo link for normal players */
		else printf(LF "<a href=\"plinfo.cgi?id=%s\">%s</a></td><td bgcolor=#%s align=right>"
					LF "Status just <a href=\"plinfo.cgi?id=%s&fref=%s&prior=1\">before</a>"
					LF "| <a href=\"plinfo.cgi?id=%s&fref=%s\">after</a> event",
					p1ptr->id, p1ptr->name, ct[j & 1], p1ptr->id, eptr->ref, p1ptr->id, eptr->ref);

		/* Onto next row */
		printf(LF "</td></tr><tr><td bgcolor="GREEN">Pseudonym:</td><td colspan=2 bgcolor=#%s>", ct[++j & 1]);

		/* Allowed to change assassin's pseudonym? */
		if ((admin & AD_ASS_PSEUD) && (f & EF_MASK_Pseudonym))
		{
			/* Pseudonym selector */
			pseudselector(p1ps, ((f & EF_SetPseudonym) ? 1 : 0), "ed_p1psu", "ed_p1psn", "ed_p1psd",
						 ((p1sp) ? p1sp->allpseuds : 0));
	
			/* Note that p1's 'set default' flag is settable */
			testf |= TST_SetP1Pseudonym;
		}

		/* Not allowed to change assassin's pseudonym: just print it */
		else if (f & EF_MASK_Pseudonym) encapsulate(p1ps, stdout);

		/* No pseudonym associated with event */
		else puts("(<b>No pseudonym associated with event</b>)");
			
		/* Admin: show some assassin-related options */
		if (admin & AD_MIN)
		{
			/* Scoring bonus?  */
			printf(LF "</td></tr><tr><td rowspan=%d bgcolor="GREEN">Scoring:</td>"
				   LF "<td colspan=2 bgcolor=#%s>Award bonus score:&nbsp;"
				   LF "<input type=\"text\" name=\"ed_bonus\" size=8 maxlength=8 value=\"%+f\">&nbsp;&nbsp;"
				   LF "(may be negative)",
				   ((f & EF_Assassination) ? 2 : 1), ct[++j & 1], (bonus / ((double) (1 << SCORE_SHIFT))));

			/* Option: dangerousness bonus disabled? */
			if (f & EF_Assassination)
			{
				printf(LF "</td></tr><tr><td colspan=2 bgcolor=#%s>"
					   LF "<input type=\"checkbox\" name=\"ed_snodbonus\" value=\"set\"%s> &nbsp;"
					   LF "Ignore assassin and victim's dangerousness when scoring.",
					   ct[++j & 1], (f & EF_NoDangerBonus) ? " checked" : "");
	
				testf |= TST_NoDangerBonus;
			}

			printf(LF "</td></tr><tr><td bgcolor="GREEN">Wantedness:</td>"
				   LF "<td colspan=2 bgcolor=#%s>Nothing here yet.", ct[++j & 1]);

			printf(LF "</td></tr><tr><td bgcolor="GREEN">Inactivity deadline:</td>"
				   LF "<td colspan=2 bgcolor=#%s>Nothing here yet.", ct[++j & 1]);
		}

		/* Print assassination victim's details */
		if (p2ptr)
		{
			/* Begin table segment */
			printf(LF "</td></tr></table><br/>"
				   LF "<table width=100%% border=0 cols=3 bgcolor=#000000 cellpadding=4 cellspacing=2>"
				   LF "<tr><td bgcolor="GREEN"><b>Victim:</b></td><td bgcolor=#%s>", ct[++j & 1]);
		
			/* Special players aren't linked to */
			if (p2ptr->flags & PF_Special) printf("<b>%s</b></td><td bgcolor=#%s align=right>&mdash;",
												  p2ptr->name, ct[j & 1]);
		
			/* Show name and plinfo link for normal players */
			else printf(LF "<a href=\"plinfo.cgi?id=%s\">%s</a></td><td bgcolor=#%s align=right>"
						LF "Status just <a href=\"plinfo.cgi?id=%s&fref=%s&prior=1\">before</a>"
						LF "| <a href=\"plinfo.cgi?id=%s&fref=%s\">after</a> event",
						p2ptr->id, p2ptr->name, ct[j & 1], p2ptr->id, eptr->ref, p2ptr->id, eptr->ref);
		
			/* Onto next row */
			printf(LF "</td></tr><tr><td bgcolor="GREEN">Pseudonym:</td><td colspan=2 bgcolor=#%s>", ct[++j & 1]);

			/* Not allowed to change victim's pseudonym: just print it */
			if (!(admin & AD_TAR_PSEUD)) encapsulate(p2ps, stdout);
				
			/* Allowed to change assassin's pseudonym? */
			else
			{
				/* Pseudonym selector */
				pseudselector(p2ps, ((f & EF_SetP2Pseudonym) ? 1 : 0), "ed_p2psu", "ed_p2psn", "ed_p2psd",
							  ((p2sp) ? p2sp->allpseuds : 0));
		
				/* Note that p2's 'set default' flag is settable */
				testf |= TST_SetP2Pseudonym;
			}

			/* Admin: show auto-resurrection stuff */
			if (admin & AD_MIN)
			{
				if (!(tmptime = addhours(eptr->when, reshours(eptr->when)))) erf(ERF_BadTimeVal, 71);
	
				printf(LF "</td></tr><tr><td bgcolor="GREEN">Targetting:</td><td colspan=2 bgcolor=#%s>"
					   LF "<input type=\"checkbox\" name=\"ed_sdetarg_vict\" value=\"set\"%s>&nbsp;"
					   LF "Deallocate all of victim's target and assassin links.</td></tr>"
					   LF "<tr><td bgcolor="GREEN">Auto-Res:</td><td colspan=2 bgcolor=#%s>"
					   LF "<table border=0 cellspacing=0 cellpadding=0><tr><td>"
					   LF "<input type=\"checkbox\" name=\"ed_sinvres\" value=\"set\"%s>&nbsp;&nbsp;</td>"
					   LF "<td>%sesurrect <b>victim</b> at "TDISP" %s (reverse normal resurrection discipline).",
					   ct[(j + 1) & 1], (f & EF_DetargP2) ? " checked" : "",
					   ct[j & 1], (f & EF_InvertAutoRes) ? " checked" : "",
					   (OPT_AutoRes) ? "Don't r" : "R", Thm(tmptime), fdayname(T_d(tmptime), 0));
	
				if (i = (f & EF_Kill) ? 0 : 1) puts("<br/><b>Has no effect:</b> Event not fatal.");
				
				if (!(p2ptr->flags & PF_Special)) printf("%s Player is special." LF,
														 (i++) ? "" : "<br/><b>Has no effect:</b>");
	
				if (tmptime >= TIME_AutoResLimit) printf("%s Event past resurrection deadline." LF,
														 (i++) ? "" : "<br/><b>Has no effect:</b>");
				puts("</td></tr></table>");
	
				testf |= TST_DetargP2|TST_InvertAutoRes;
			}
		}

		/* Admin: news related stuff */
		if (admin & AD_MIN)
		{
			/* Begin table segment */
			printf(LF "</td></tr></table><br/>"
				   LF "<table width=100%% border=0 cols=2 bgcolor=#000000 cellpadding=4 cellspacing=2>"
				   LF "<tr><td bgcolor="GREEN"><b>News:</b></td>"
				   LF "<td bgcolor=#%s>Nothing here yet", ct[++j & 1]);
		}

		/* Close the most recent table */
		puts("</td></tr></table><p align=right>");

		/* Allow assassination outcomes to be edited */
		if (f & EF_Assassination) printf(LF "Outcome:&nbsp;&nbsp;<select name=\"ed_outcome\" size=1>"
										 LF "<option %svalue=\"fail\">Failed attempt</option>"
										 LF "<option %svalue=\"kill\">Successful kill</option></select>&nbsp;&nbsp;",
										 ((f & EF_Kill) ? "" : "selected "), ((f & EF_Kill) ? "selected " : ""));
			   
		/* Allow adjustment of things */
		printf(LF "<input type=\"hidden\" name=\"ed_testf\" value=\"0x%lX\">"
			   LF "<input value=\"Submit changed event details\" type=\"submit\" name=\"sub_eved\"></p>", testf);
	}

	/* Present standard report summary list */
	if (rlist)
	{
		/* Acquire full report information if have not already done so */
		if ((!gotreps) && (rc = repfile(&rlist, plist, statbuf_pos, statbuf_end, 0, 0))) erf(rc, 72);

		/* Report information has now been loaded */
		gotreps = 1;

		/* Loop through each relevent report */
		for (i = 0, rptr = rlist; rptr; rptr = rptr->link_all)
		{
			/* Paranoia */
			if (!rptr->pptr) erf(ERF_InternalError, 73);

			/* Is it report existence publically displayable? */
			if (!(admin & AD_MIN) && !(rptr->flags & RF_Listed)) continue;

			/* Alternate background colour */
			p = (i & 1) ? BLUE : RED;

			if (!i++) printf("</td></tr><tr><td colspan=2 bgcolor="GREEN">"
							 LF "<table width=100%% border=0 cols=%i bgcolor=#000000 cellpadding=4 cellspacing=2>"
							 LF "<thead><td colspan=2 bgcolor="GREEN"><b>Existing reports:</b></td>"
							 LF "%s<td bgcolor="GREEN">By:</td><td bgcolor="GREEN">Updated:"
							 LF "</td><td bgcolor="GREEN">Under pseudonym:</td></thead>",
							 ((admin & AD_MIN) ? 12 : 5),
							 ((admin & AD_MIN) ? "<td bgcolor="GREEN" colspan=7>Flags:</td>" : ""));

			printf(LF "<tr bgcolor=%s><td><input value=\"Edit\" type=\"submit\" name=\"sub_red_%s\"></td>"
				   LF "<td><b>&quot;%s&quot;</b> (%li bytes)</td><td>", p, rptr->rid, rptr->rid, rptr->length);

			/* Admin: show report flags */
			if (admin & AD_MIN)
			{
				char *q;

				/* Reproduction type */
				switch (rptr->flags & RF_Repro)
				{
					case RF_Normal:		 q = "";  break;
					case RF_Verbatim:	 q = "V"; break;
					case RF_Encapsulate: q = "E"; break;
					default:			 q = "?";
				}

				/* Print flags as letter switches */
				printf("%s</td><td>%s</td><td>%s</td><td>%s</td><td>%s</td><td>%s</td><td>%s</td><td>",
					   ((rptr->flags & RF_Main)     ? "M" : ""), ((rptr->flags & RF_Comment)  ? "C" : ""),
					   ((rptr->flags & RF_Listed)   ? "" : "U"), ((rptr->flags & RF_AutoGen)  ? "A" : ""),
					   ((rptr->flags & RF_Hidden)   ? "H" : ""), ((rptr->flags & RF_NoIntro)  ? "N" : ""), q);
			}

			j = (!strcmp(rptr->pptr->id, LEG_ID) || !strcmp(rptr->pptr->id, ILL_ID));
			if (!j) printf("<a href=\"plinfo.cgi?id=%s\">", rptr->pptr->id);
			if (admin & AD_OBS) puts(rptr->pptr->name);
			else if (rptr->pptr->pseud) encapsulate(rptr->pptr->pseud, stdout);
			else printf("[**NoPseud:%s**]", rptr->pptr->id);
			if (!j) puts("</a>");

			/* Special role in event? */
			if (rptr->pptr == p1ptr) puts(" (assassin)");
			if ((p2ptr) && (rptr->pptr == p2ptr)) puts(" (victim)");

			/* Last edit time unknown */
			if (!rptr->uptim) puts("</td><td>&mdash;</td><td>");

			/* Time of last edit */
			else printf("</td><td>%s "TDISP"</td><td>", fdayname(T_d(rptr->uptim), "%d %b"), Thm(rptr->uptim));

			/* Specific pseudonym for report */
			if (rptr->pseud) encapslim(rptr->pseud, stdout, MAX_PseudDisplay);

			/* Standard pseudonym for player */
			else puts("(<b>standard</b>)");

			puts("</td></tr>");
		}

		if ((i) && !(admin & AD_MIN)) puts("</table>");

		else if (i) puts("</tr><tr><td colspan=12 bgcolor="GREEN">Key: <tt>M</tt>ain, <tt>C</tt>omment,"
						 " <tt>U</tt>nlisted, <tt>A</tt>uto-generated, <tt>H</tt>idden, <tt>V</tt>erbatim,"
						 " <tt>N</tt>o introduction</td></tr></table>");

		else puts("</td></tr><tr><td colspan=2 bgcolor="GREEN"><b>No reports (under this authentication).</b>");
	}
	else printf("</td></tr><tr><td colspan=2 bgcolor="GREEN"><b>No reports yet.</b>");

	/* Add new report section */
	printf("<p align=right>Add new report by: &nbsp; <select name=\"rep_by\" size=1>");

	/* List of players to author new report */
	if (i = makeidmenu(authp->id, PF_Playing, PF_Playing, OPT_IDMenuMode))
	{
		fprintf(stderr, "%li:   rc %d from makeidmenu(10) in genfef.\n", uniqid, i);
	}

	/* Finish off event overview form */
	printf("</select> &nbsp; <input value=\"Add report...\" type=\"submit\" name=\"sub_addrep\"></p></td></tr>"
		   LF "</table></p></form><p>"
		   LF "<form method=\"POST\" action=\""CGI_URL THIS_CGI"\">"
		   LF "<input type=\"hidden\" name=\"authname\" value=\"%s\">"
		   LF "<input type=\"hidden\" name=\"authpass\" value=\"%s\">"
		   LF "<input value=\"Add new event\" type=\"submit\"></form><br/>"
		   LF "<form method=\"POST\" action=\""CGI_URL"front.cgi\">"
		   LF "<input type=\"hidden\" name=\"id\" value=\"%s\">"
		   LF "<input type=\"hidden\" name=\"pass\" value=\"%s\">"
		   LF "<input value=\"Return to player front-end\" type=\"submit\"></form></p>"
		   LF "<p><hr/>" FOOTER "<hr/></p></body></html>",
		   authp->id, authp->pass, authp->id, authp->pass);


_erf: /****** Error handling ******/

	/* Debug */
	if (rc) fprintf(stderr, "%li:   genfulleventform(%i):%i auth=\"%s\" mode=%i p1id=%s p2id=%s p3id=%s bad=%lu\n",
					uniqid, ERF_CONTEXT(rc), ERF_TYPE(rc), authp->id, mode,
					((p1ptr)?(p1ptr->id):"zz"), ((p2ptr)?(p2ptr->id):"zz"), ((p3ptr)?(p3ptr->id):"zz"), bad);

	/* Only retain low bits of normal rc's, and pass most of then on */
	return(ERF_LOWBITS(rc));
}

/***************************************************************************************************************************
** Handle assassination events and event/report edits                                                                     **
***************************************************************************************************************************/
int doev(int mode, char *authname, char *authpass, char **buf_posp, char *buf_end, char *parm)
{
	char *p, *ipids[5] = { authname, "", "", "", 0};
	player_p pptrs[4] = { 0, 0, 0, 0 };
	player_p *plidx = pptrs, authp = 0, ppt;
	pshort_p moo = 0;
	int i, j, rc = 0;
	FILE *fp;


	/* (Possible) Participants in event? */
	ipids[1] = findval("ev_by", 1);
	ipids[2] = findval("ev_on", 1);

	/* Author of report */
	if ((mode >= 5) && (mode <= 7)) ipids[3] = findval("rep_by", 1);

	/* Read full info for all players */
	if (rc = scanplinfo(SCANPL_AddFull|SCANPL_GetTargs, ipids, &plidx,
						&moo, &plist, *buf_posp, buf_end, buf_posp)) erf(rc, 4);

	/* Desire non-empty player list */
	if (!plist) erf(ERF_InternalError, 5);

	/* Debug */
	fprintf(stderr, "%li:   authname=\"%s\" authpass=\"%s\"\n", uniqid, authname, authpass);

	/* Invalid authentication? */
	if (!(authp = pptrs[0]) || (!authp->name) || !(authp->flags & PF_Login) ||
		(!(authp->flags & PF_Public) && strcmp(authp->pass, authpass)))
	{
		/* No valid auth player struct */
		authp = 0; rc = ERF_BadAuth;
	}

	/* Can public auth users add events? */
    #if !OPT_PubAddEvents
	else if ((mode <= 2) && (authp->flags & PF_Public)) rc = ERF_CantPubAddEv;
	#endif

	/* Can players add events? */
	#if OPT_NoAddEvents
	else if ((mode <= 2) && !(authp->flags & PF_Admin)) rc = ERF_CantAddEv;
	#endif

	/* Can public edit events? */
	else if ((mode == 4) && (authp->flags & PF_Public)) rc = ERF_CantPubEditEv;

	/* Can players edit events? */
	#if OPT_NoEditEvents
	else if ((mode >= 4) && !(authp->flags & PF_Admin)) rc = ERF_CantEditEv;
	#endif

	/* Output source from */
	else if (mode <= -1) rc = ERF_Void;

	/* Process various event documents */
	else rc = genfulleventform(mode, authp, pptrs[1], pptrs[2], pptrs[3], buf_posp, buf_end, parm);

	/* Initialisation */
	p = 0;
	fp = 0;
	
	/* Outcome */
	switch (rc)
	{
		/* Success */
		case 0: return 0;

		/* Handle errors */
		default: erf(rc, 9);

		/* Access violations */
		case ERF_CantPubDoReps:	if (!p) p = FILE_CantPubDoReps;
		case ERF_NotRepOwner:	if (!p) p = FILE_NotRepOwner;
		case ERF_CantPubAddEv:	if (!p) p = FILE_CantPubAddEv;
		case ERF_CantAddEv:		if (!p) p = FILE_CantAddEv;
		case ERF_CantPubEditEv:	if (!p) p = FILE_CantPubEditEv;
		case ERF_CantEditEv:	if (!p) p = FILE_CantEditEv;
		case ERF_BadAuth:		if (!p) p = FILE_BadAuth;
		case ERF_UnknownRef:	if (!p) p = FILE_UnknownRef;
								authform(rc, p, authp, authname);
								return 0;

		/* Output source document */
		case ERF_Void:
		{
			/* Nothing wrong so far */
			char *q = 0; i = 0;

			/* HTML header and preamble */
		    puts("Content-type: text/html" LF
				 LF "<HTML><HEAD><TITLE>" GAME " Assassins' Automatic Umpire</TITLE></HEAD>"
				 LF "<BODY bgcolor=#000000 text=#FFFFDD link=#BBBBFF vlink=#FFCCFF alink=#FFFFFF>"
				 LF "<h1><hr/>Add new event<hr/></h1>");


			/* Open file */
			if (!(fp = fopen(FILE_AddEvGuide, "rb"))) q = "Could not open HTML file";

			/* Copy it over */
			if ((fp) && (i = ftofcpy(fp, stdout, -1L, fpbuf, MAX_fpbuf, -1L, -1L))) q = "Could not copy HTML file";

			/* Finished */
			if (fp) fclose(fp);

			/* HTTP header and HTML on further error */
			if (q) printf(LF "<h3><img src=\"" CROSS_URL "\" align=left " CROSS_SIZ " hspace=8>"
						  LF "<font color=#FFCCCC>A misconfiguration manifested whilst copying guide text.<br/>"
						  LF "You can ignore this. Code %i: %s.</font><br clear=left></h3>", i, q);
	
			/* Event entry form */
			(void) printblankform(0, authp, 0, 0, thetime, authp, 0, 4, 0, 0);

			/* For admin, comment entry form */
			if (authp->flags & PF_Admin)
			{
				printblankform(2, authp, authp, 0, thetime, authp, EF_ShowHeadline, 0, OPT_FragComRep, OPT_FragComHead);
			}

			puts("</body></html>");
			return 0;
		}
	}

	/* Old dispatch code
	rc = genfulleventform(1, authp, pla[1], pla[2], 0, &statbuf_pos, statbuf_end, 0);
	rc = genfulleventform(2, authp, pla[1], 0, 0, &statbuf_pos, statbuf_end, 0);
	rc = genfulleventform(4, authp, 0, 0, 0, &statbuf_pos, statbuf_end, 0);
	rc = genfulleventform(5, authp, 0, 0, pla[3], &statbuf_pos, statbuf_end, 0);
	rc = genfulleventform(7, authp, 0, 0, pla[3], &statbuf_pos, statbuf_end, 0);
	rc = genfulleventform(6, authp, 0, 0, pla[3], &statbuf_pos, statbuf_end, p);
	rc = genfulleventform(3, authp, pla[1], pla[2], 0, &statbuf_pos, statbuf_end, p);*/


_erf: /****** Error handling ******/

	/* Debug */
	if (rc) fprintf(stderr, "%li:   doev(%i):%i %s mode=%i id1=\"%s\" id2=\"%s\" id3=\"%s\"\n", uniqid,
					ERF_CONTEXT(rc), ERF_TYPE(rc), ((authp) ? "auth " : ""), mode, ipids[1], ipids[2], ipids[3]);

	/* Only retain low bits of normal rc's, and pass most of then on */
	return(ERF_LOWBITS(rc));
}

/***************************************************************************************************************************
** Main code. Doesn't make use of command line arguments.                                                                 **
***************************************************************************************************************************/
int main(int argc, char* argv[])
{
	char statbuf[MAX_statbuf], *statbuf_pos = statbuf, *statbuf_end = statbuf + MAX_statbuf - 1;
	char *p, *authname = "", *authpass = "", *thetimestr;
	int i = -3, rc = 0;


	/* Debug: no buffering */
	/* setbuf(stdout, 0); */

	/* Output HTTP headers */
    /* printf("Content-type: text/html" LF LF); */

	/* Initialse stuff: time, file creation umask, pid, make note in log and redirect stderr there */
	if ((rc = geninit(&thetimestr, 0117, 1, "Event")) < 0) exit(send_doc(-rc, 0));

	/* Logged error handler now established */
	else if (rc) erf(rc, 1);

	/* Parse CGI input */
	if ((i = parsecgi()) > 0) erf(i, 1);

	/* If no authentication name given, assume public user */
	if (!(authname = findval("authname", 0))) authname = PUB_ID;

	/* Authentication password */
	else authpass = findval("authpass", 1);

	/* Claim lockfile */
	if (!(lockfd = setflock("Event.c "))) erf(ERF_flockFailed, 2);

	/* Initial event details given -- output form to collect more data */
	if (findval("sub_prime", 0)) i = 0;

	/* Enough event data given -- add event */
	else if (findval("sub_ev", 0)) i = 1;

	/* Enough data given -- add comment */
	else if (findval("sub_com", 0)) i = 2;

	/* Process event edit form -- edit event */
	else if (findval("sub_eved", 0)) i = 4;

	/* Generate the report addition form */
	else if (findval("sub_addrep", 0)) i = 5;

	/* Process the reporting form -- add/edit the report */
	else if (findval("sub_rep", 0)) i = 7;

	/* Application to edit a report: generate the report edit form */
	else if (p = findsuf("sub_red_", 616)) i = 6;

	/* Generate the edit event form */
	else if (p = findval("ev_ref", 0)) i = 3;
	
	/* Otherwise, output source document */
	else i = -1;

	/* Do the actual processing */
	if (i >= -1) rc = doev(i, authname, authpass, &statbuf_pos, statbuf_end, p);

	/* Release the lock */
	setfunlock(lockfd);
	lockfd = 0;

	/* Any side effects necessary? */
	if (rc) rescore = rescore >> RESC_ERR__Shift;

	/* Rescore if necessary */
	if (rescore & RESC_Scoring) system(SYS_Scores" 2> "FILE_ScoreLog" 1>&2 &");

	/* Rebuild news if necessary */
	if (rescore & RESC_MkNews) system(SYS_MkNews" 2> "FILE_NewsLog" 1>&2 &");

	/* Launch stats page */
	if ((rescore & RESC_Flaunch) && flaunch) erf(launchwithstdin(CGI_ROOT "stats.cgi", flaunch), 30);


_erf: /***** Error handling *****/

	/* Debug */
	if (rc) fprintf(stderr, "%li:   main(%i):%i:%s authname=\"%s\" authpass=\"%s\" rescore=%u i=%i\n",
					uniqid, ERF_CONTEXT(rc), ERF_TYPE(rc), ERF_STR(rc), authname, authpass, rescore, i);

	/* Tidy up */
	if (lockfd) setfunlock(lockfd);

	/* Only retain low bits of normal rc's */
	if (rc) send_doc(ERF_LOWBITS(rc), 0);

	return 0;
}

/***************************************************************************************************************************
** Produce and output HTML document data for the various situations indicated by (which).                                 **
***************************************************************************************************************************/
void authform(int which, char* p, player_p authp, char* authname)
{
	unsigned long pfb = 0, pfm = 0;
	FILE *fp = 0;
	char *q = 0;
	int i = 0;


	/* HTTP header */
    puts("Content-type: text/html" LF LF);

	/* Open file */
	if ((p) && !(fp = fopen(p, "rb"))) q = "Could not open HTML file";

	/* Copy it over */
	if (fp)
	{
		if (i = ftofcpy(fp, stdout, -1, fpbuf, MAX_fpbuf, -1, -1)) q = "Could not copy HTML file";
		fclose(fp);
	}

	/* HTTP header and HTML on further error */
	if ((q) || (!p)) printf(LF "<HTML><HEAD><TITLE>" GAME " Assassins' Automatic Umpire: Error</TITLE></HEAD>"
							LF "<BODY bgcolor=#000000 text=#FFFFDD link=#BBBBFF vlink=#FFCCFF alink=#FFFFFF>"
							LF "<h1><hr/>%i: Could not carry out original request.<hr/></h1>"
							LF "<h3><img src=\"" CROSS_URL "\" align=left " CROSS_SIZ " hspace=8>", which);

	if (q) printf(LF "<font color=#FFCCCC>Additionally, an error or misconfiguration manifested"
				  LF "whilst copying error document.<br/>"
				  LF "%i: %s \"%s\"</font><br clear=left></h3>", i, q, p);

	if (!p) puts(LF "<font color=#FFCCCC>No specific error documentation available.</font><br clear=left></h3>");

	switch (which)
	{
		#if OPT_NoAddEvents
		case ERF_CantAddEv: pfb = pfm = PF_Admin; break;
		#endif

		#if !OPT_PubAddEvents
		case ERF_CantPubAddEv: pfm = PF_Public; break;
		#endif

		case ERF_NotRepOwner:
		case ERF_CantPubDoReps: pfm = PF_Public; break;

		case ERF_CantEditEv:
		case ERF_CantPubEditEv:
		{
			#if OPT_NoEditEvents
			pfb = pfm = PF_Admin; break;
			#else
			pfm = PF_Public; break;
			#endif
		}
	}

	puts(LF "<form method=\"POST\" action=\"" CGI_URL THIS_CGI "\">"
		 LF "<p>Please (re)select your name and enter your password.</p>"
		 LF "<p>Authenticate as: &nbsp; <select size=1 name=\"authname\">");

	if (i = makeidmenu((suggestid) ? suggestid : authname, pfm|PF_Login, pfb|PF_Login, OPT_IDMenuMode))
	{
		fprintf(stderr, "%li:   rc %d from makeidmenu(3) in authform(%i)\n", uniqid, i, which);
	}

	puts("</select> &nbsp; password: <input name=\"authpass\" type=\"password\" size=24 maxlength=64></p>");

	if (!(i = (which != ERF_UnknownRef)))
	{
		printf("<p>Event ref: <input name=\"ev_ref\" value=\"");
		encapsulate(findval("ev_ref", 1), stdout);
		printf("\" type=\"text\" size=10 maxlength=10></p>");
	}

	regurgitate(excl_auth + i);

	puts(LF "<p><input value=\"(Re)authenticate and resubmit\" type=\"submit\" name=\"reauth\"></p>"
		 LF "</form><hr/><p>" FOOTER "<hr/></p></body></html>");
}

int send_doc(int which, player_p authp)
{
	int i, j;

	/* Output HTTP headers */
    printf("Content-type: text/html" LF LF);

	if (which == ERF_BadAuth)
	{
		printf(LF "<HTML><HEAD><TITLE>" GAME " Assassins' Automatic Umpire</TITLE></HEAD>"
			   LF "<BODY bgcolor=#000000 text=#FFFFDD link=#BBBBFF vlink=#FFCCFF alink=#FFFFFF>"
			   LF "<h1><hr/>Automatic Umpire: Authentication<hr/></h1>"
			   LF "<h3><img src=\"" CROSS_URL "\" align=left " CROSS_SIZ " hspace=8>"
			   LF "<font color=#FFCCCC>Authentication failed</font></h3>"
			   LF "<form method=\"POST\" action=\"" CGI_URL THIS_CGI "\">"
			   LF "<p>Please select your name and enter your password.</p>"
			   LF "<p>Authenticate as: &nbsp; <select size=1 name=\"authname\">");

		if (i = makeidmenu((suggestid) ? (suggestid) : findval("authname", 1), PF_Login, PF_Login, OPT_IDMenuMode))
		{
			fprintf(stderr, "%li:   rc %d from makeidmenu(3) in send_doc(%d)\n", uniqid, i, which);
		}

		puts("</select> &nbsp; password: <input name=\"authpass\" type=\"password\" size=24 maxlength=64></p>");

		regurgitate(excl_auth);

		puts("<p><input value=\"Authenticate and resubmit\" type=\"submit\" name=\"reauth\"></p>"
			 LF "</form><hr/><p>" FOOTER "<hr/></p></body></html>");
	}

	else error_doc(which);

	return 0;
}

/*** The end ***/
