/*****************************************************
**  lib-uinfo-read.c                     (library)  **
**  ----------------------------------------------  **
**  ----------------------------------------------  **
**  Updated to use new time system.                 **
**                                                  **
************** BORING LEGAL NOTICE *******************
**                                                  **
**  Assassins' Game (Semi-)Automatic Umpire System  **
**  ----------------------------------------------  **
**  Prerelease: Copyright 2000-2002 Edward M Nokes  **
**  This code may not be distributed in source or   **
**  executable form without the express consent of  **
**  the author, Edward M Nokes (emn23@cam.ac.uk).   **
**  The author accepts no responsibility or         **
**  liability for anything occurring as a result    **
**  of using this software, etc. etc.               **
**                                                  **
*****************************************************/

#ifdef included__lib_uinfo_read
 #error "** lib-uinfo-read multiply #included **"
#else
 #define included__lib_uinfo_read fromfile

/***************************************************************************************************************************
** Declarations and definitions                                                                                           **
***************************************************************************************************************************/

/* Acquire targetting details? */
#define UREAD_GetTargs	(1<<0)

/* Function-pointer prototypes for hooks in uread() */
typedef int (*uread_qfn) (char* id, char* name, void* context);
typedef int (*uread_pfn) (player_p data, int state, void* context);
typedef int (*uread_xfn) (player_p data, char* text, int state, void* context);

/***************************************************************************************************************************
** Build a list of player information in various styles from 'uinfo' file. Returns nonzero result on failure. Takes a     **
** pointer (entry_f) to a "uread_entry_fn" (that takes name, address, flags, timer, kill/death count) and calls it        **
** twice per player, once with with (state) = 0 before outputting pseudonym information, and once with (state) = 1 after  **
** all the (pseud_f) calls (return code permitting). The (pseud_f) to "int fn(char* pseudo, int state, timmy* tim)" is    **
** called it once per pseudonym for each player with (state) bit 0 set if (pseudo) is player's current pseudonym and      **
** with bit 1 set if (pseudo) was the last pseudonym found for that player, and bit 2 set if (pseudo) is "first" use of   **
** the pseudonym; (tim) is the time-value at which the pseudonym was chosen. A bit-0-set return from (pseud_f) instructs  **
** uread() to skip to the next player, a bit-1-set return that the passed pseudonym data is not to be overwritten.        **
** A return from the first call to (entry_f) with bit 0 set instructs uread() not to call the (pseud_f) for this player   **
** (ie. skip to next player); one with bit 1 set instructs uread() not to process any more players, ie. to finish; one    **
** with bit 2 clear instructs uread() not to call the (entry_f) for the second time. One with bit 3 set indicates that    **
** the passed data is not to be overwritten. The return value from the second call to (entry_f) is ignored.               **
***************************************************************************************************************************/
int uread(uread_qfn qf, uread_pfn pf, uread_xfn xf, char* buf, char* bufend, char** bufused, void* ctx, int mode)
{
	char *bp, *p;
	FILE *fp = 0;
	int i, c, j, rc = 0;
	player_p pp = 0;
	target_p tarp = 0, *tarplist;

	/* Ensure at least bytes free in buffer */
	if ((bufend - buf) < chk[CHK_uinfo].len) erf(ERF_BufferExhaust, 60);

	/* Open player info file */
	if (!(fp = fopen(FILE_UserInfo, "rb"))) erf(ERF_fopenUinfo, 1);

	/* Check file header (ignoring initial comments) */
	if (i = chkheader(fp, CHK_uinfo, buf, 0, 0, 0, 0)) erf(i, 61);

	/* Scan through for valid lines */
	while ((c = fgetc(fp)) != EOF)
	{
		/* Ignore lines not beginning with "*" character */
		switch (c)
		{
			case '*'  : break;
			default   : toeol(fp);
			case '\n' : continue;
		}
		
		/* Position in buffer */
		bp = buf;

		/* Claim memory for a player struct (if none already available) */
		if ((!pp) && !(pp = malloc(sizeof(struct player)))) erf(ERF_mallocFailed, 2);

		/* Clear the struct's link and opaque extension word */
		pp->ext = 0; pp->link = 0;

		/* No targets, etc. read yet */
		pp->targs = pp->asses = pp->oldts = 0;

		/* Skip initial tabs, then read 'id' up to first tab or newline */
		if (i = tread(fp, pp->id = bp, bufend, &bp)) erf(ERF_BadTread+i, 3);

		/* Read password field */
		if (i = tread(fp, pp->pass = bp, bufend, &bp)) erf(ERF_BadTread+i, 4);

		/* Read name field */
		if (i = tread(fp, pp->name = bp, bufend, &bp)) erf(ERF_BadTread+i, 5);

		/* Call quick entry function if passed */
		if ((!qf) || (((i = qf(pp->id, pp->name, ctx)) >= 0) && (i & 1)))
		{
			/* Retain buffer if required */
			if (i & 8) buf = bp;

			/* Read college field */
			if (i = tread(fp, pp->coll = bp, bufend, &bp)) erf(ERF_BadTread+i, 6);

			/* Read mafia field */
			if (i = tread(fp, pp->mafia = bp, bufend, &bp)) erf(ERF_BadTread+i, 7);

			/* Read address field */
			if (i = tread(fp, pp->addr = bp, bufend, &bp)) erf(ERF_BadTread+i, 8);
	
			/* Read email address field */
			if (i = tread(fp, pp->email = bp, bufend, &bp)) erf(ERF_BadTread+i, 9);

			/* Read flags field */
			if (i = tread(fp, bp, bufend, 0)) erf(ERF_BadTread+i, 10);
			pp->flags = strtoul(bp, 0, 16);
	
			/* Read restime field */
			if (i = tread(fp, bp, bufend, 0)) erf(ERF_BadTread+i, 11);
			if (!(pp->timer = str2tim(bp))) erf(ERF_BadDateText, 12);
	
			/* Read kill count */
			if (i = tread(fp, bp, bufend, 0)) erf(ERF_BadTread+i, 13);
			pp->kills = atoi(bp);
	
			/* Read death count */
			if (i = tread(fp, bp, bufend, 0)) erf(ERF_BadTread+i, 14);
			pp->deaths = atoi(bp);
	
			/* Read score */
			if (i = tread(fp, bp, bufend, 0)) erf(ERF_BadTread+i, 15);
			pp->score = atol(bp);

			/* Read rank */
			if (i = tread(fp, bp, bufend, 0)) erf(ERF_BadTread+i, 16);
			pp->rank = atol(bp);

			/* Read dangerousness */
			if (i = tread(fp, bp, bufend, 0)) erf(ERF_BadTread+i, 17);
			pp->danger = atol(bp);

			/* Get targets, assassins, old targets */
			if (mode & UREAD_GetTargs) while (1)
			{
				/* Read */
				if (i = tread(fp, p = bp, bufend, &bp)) erf(ERF_BadTread+i, 40);
            
				/* End of list */
				if (p[0] == '-') break;

				/* Allocate storage */
				if (!(tarp = malloc(sizeof(struct target)))) erf(ERF_mallocFailed, 41);

				/* Assassin? */
				if (p[0] == 'A') tarplist = &pp->asses;

				/* Target? */
				else if (p[0] == 'T') tarplist = &pp->targs;

				/* Old target? */
				else if (p[0] == 'O') tarplist = &pp->oldts;

				/* Invalid data */
				else { printf("Error in targetting: %s",p); erf(ERF_InternalError, 42); }

				/* Valid id? */
				if (*(tarp->id = p+1) <= 32) erf(ERF_InternalError, 43);

				/* Clear extension word */
				tarp->ext = 0;

				/* Link into list */
				tarp->next = *tarplist;
				*tarplist = tarp;
			}

			/* Read "wantedness" field */
			if (i = tread(fp, p = bp, bufend, &bp)) erf(ERF_BadTread+i, 17);

			/* Decode (may now be null string) */
			pp->wanted = SESC(p);

			/* Read pseudonym field */
			if ((c = tread(fp, p = bp, bufend, &bp)) > 1) erf(ERF_BadTread+c, 18);

			/* Decode (may now be null string) */
			pp->pseud = SESC(p);

			/* Call supplied function */
			i = (pf) ? (pf(pp, 0, ctx)) : 0;
		
			/* Terminate and pass on error */
			if (i < 0) erf(-i, 19);

			/* Loop for additional data */
			else if ((c == 0) && (xf) && !(i & 1))
			{
				int n;
				char* bpt;

				/* Read an additional field */
				for (c = 1; ((j = tread(fp, bp, bufend, &bpt)) < 2); c = 0)
				{
					/* Call supplied function (and don't overwrite if bit 1 returned set) */
					if ((n = xf(pp, bp, (j)?(c+2):(c), ctx)) & 2) bp = bpt;

					/* Handle thrown errors */
					if (n < 0) erf(-n, 20);

					/* Allow early termination */
					else if ((n & 1) || (j)) break;
				}

				/* Handle any file access errors left hanging */
				if (j > 1) erf(ERF_BadTread+j, 21);
			}
	
			/* Call the supplied function again? If buffer and struct are not being overwritten,
			** their contents are guaranteed safe once this call returns a non-negative value. */
			if ((i & 4) && ((j = pf(pp, 1, ctx)) < 0)) erf(-j, 23);
		}

		/* Error from quick function */
		else if (i < 0) erf(-i, 22);

		/* Don't reuse buffer if required */
		if (i & 8) buf = bp;

		/* Don't preserve target allocations */
		if (0 && (pp) && !(i & 32))
		{
			target_p tarq = 0;

			/* Free memory blocks */
			for (tarp = pp->targs; (tarp); tarp = tarq) { tarq = tarp->next; free(tarp); }
			for (tarp = pp->asses; (tarp); tarp = tarq) { tarq = tarp->next; free(tarp); }
			for (tarp = pp->oldts; (tarp); tarp = tarq) { tarq = tarp->next; free(tarp); }

			/* Note deallocation */
			pp->targs = pp->asses = pp->oldts = 0;
		}

		/* Preserve this player struct? */
		if (i & 16) pp = 0;

		/* Early return if required */
		if (i & 2) break;

		/* Next please */
		toeol(fp);
	}


_erf: /***** Error handling *****/

	/* Update extent of buffer used */
	if (bufused) *bufused = buf;

	/* Get rid of temp player struct */
	if (pp) free(pp);

	/* Tidy up */
	if (fp) fclose(fp);
	if (rc <= 0) return (rc);

	/* Debug */
	fprintf(stderr, "%li !! uread(%i):%i\n", uniqid, ERF_CONTEXT(rc), ERF_TYPE(rc));

	/* Only retain low bits of normal rc's */
	return (ERF_LOWBITS(rc));
}

#endif
/*** The End ***/
