/*****************************************************
**  genstats.c                     >> bin/genstats  **
**  ----------------------------------------------  **
**  Generate TSV statistics tables from game data.  **
**  ----------------------------------------------  **
**  Converted to new time system???                 **
**                                                  **
*************** Command line syntax: *****************
**                                                  **
**  [+|^|-]<name>[=<value>] scanned left-to-right.  **
**  Flags: '+' to set, '^' to toggle, '-' to clear. **
**  Only first char of <name> used to do matching.  **
**  Flag: <name> = 'a' to not omit annulled events. **
**  Flag: <name> = 'd' to print debugging text.     **
**  Flag: <name> = 'n' to include np atts/kills     **
**  Use <name> = 'k' to print list of kills to file **
**  pathname <value> or stdout if <value> is empty. **
**  Similarly, <name> = 'f' for list of attempts,   **
**  'bx' for bjoernstats mode (all 'x' every hour), **
**  's' for scoring debugger (explains scoring),    **
**  'cx' for changelogs (all 'x' after changes).    **
**  Where <x> = 's'|'k'|'f'|'d'|'v'|'i'|'l'|'n' for **
**   scores, kills, attempts, deaths, survivals,    **
**   illegal np kills, legal np kills, np attempts. **
**                                                  **
*************** BORING LEGAL NOTICE ******************
**                                                  **
**  Assassins' Game (Semi-)Automatic Umpire System  **
**  ----------------------------------------------  **
**  Prerelease: Copyright 2000-2002 Edward M Nokes  **
**  This code may not be distributed in source or   **
**  executable form without the express consent of  **
**  the author, Edward M Nokes (emn23@cam.ac.uk).   **
**  The author accepts no responsibility or         **
**  liability for anything occurring as a result    **
**  of using this software, etc. etc.               **
**  ----------------------------------------------  **
*****************************************************/

#define NO_CGI
#define MAX_statbuf			(1024*1024)
#define EVENT_EXT_TYPE		int			/* type of  event_p->ext is irrelevent */
#define PLAYER_EXT_TYPE		int			/* type of player_p->ext is irrelevent */
#define PSTATE_EXT_TYPE		pstate_p	/* type of pstate_p->ext is pstate_p   */
#define PSTATE_PLP_TYPE		player_p	/* type of pstate_p->plp is player_p   */
#define PEXIST_EXT_TYPE		int			/* type of pexist_p->ext is irrelevent */

#include "!lib-general.c"
#include "!custom.c"
#ifdef TEST
 #include "!lib-evolveT.c"
#else
 #include "!lib-evolve.c"
#endif

/***************************************************************************************************************************
** Definitions and declarations                                                                                           **
***************************************************************************************************************************/

/* Time limits */
timeval tmptime, endtime = 0, inittime = 0, bj_endtime = 0;

static int num_evkills = 0, num_evfails = 0;
static pstate_p bj_pstp = 0;
static evolve_notes nbmask = 0;

/* Handy "what has changed" masks */
#define NBMASK_Score	(NB_ScoreKill|NB_ScoreBonus|NB_ScoreIllegal)
#define NBMASK_Fatal	(NB_ScTabFatal)
#define NBMASK_Attempt	(NB_ScTabAttempt)

/* Possible statistics (check MODE_bj/ch flags if adding any more) */
#define STAT_Score		0
#define STAT_Kills		1
#define STAT_Fails		2
#define STAT_Deths		3
#define STAT_Survs		4
#define STAT_IllNP		5
#define STAT_LegNP		6
#define STAT_AttNP		7
#define STAT_PKill      8
#define STAT_PFail		9
#define STAT__MAX		10

/* Mask of things to check for each statistic */
evolve_notes chmask[STAT__MAX] =
{
	NBMASK_Score, NBMASK_Fatal, NBMASK_Attempt, NBMASK_Fatal, NBMASK_Attempt,
	NBMASK_Fatal, NBMASK_Fatal, NBMASK_Attempt, NBMASK_Fatal, NBMASK_Attempt
};

/* Statistic titles */
char* statit[STAT__MAX][4] = 
{
	{ "Scores", "scores", " Scores: ",
						  "---------" },
	{ "TotKills", "total kill counts", " Total kill (including non-players) counts: ",
									   "--------------------------------------------" },
	{ "TotFails", "total failed attempt counts", " Total attempt (including non-players) counts: ",
												 "-----------------------------------------------" },
	{ "Deaths", "death counts", " Death counts: ",
								"---------------" },
	{ "Survs", "survival counts", " Survival counts: ",
								  "------------------" },
	{ "LegNPKills", "legal non-player kill counts", " Legal non-playerkill counts: ",
												    "------------------------------" },
	{ "IllNPKills", "illegal non-player kill counts", " Illegal non-player kill counts: ",
													  "---------------------------------" },
	{ "NPAttempts", "failed non-player attempt counts", " Failed non-player attempt counts: ",
														"-----------------------------------" },
	{ "PlKills", "player kill counts", " Player kill counts: ",
									 "---------------------" },
	{ "PlFails", "failed player attempt counts", " Player attempt counts: ",
											   "------------------------" }
};

/* Filenames and streams for each output method */
static FILE *fp_bjstats[STAT__MAX], *fp_chstats[STAT__MAX];
static char *do_bjstats[STAT__MAX], *do_chstats[STAT__MAX];
static FILE *fp_evfails = 0, *fp_evkills = 0, *fp_breakdown = 0;
static char *do_evfails = 0, *do_evkills = 0, *do_breakdown = 0;

/* Mode flags */
#define MODE_Debug			(1UL<<0)
#define MODE_Annulled		(1UL<<1)
#define MODE_Breakdown		(1UL<<2)	/* Breakdown of all scoring decisions *//* MODE_Scores */
#define MODE_EvKills		(1UL<<3)	/* List of all kill events *//* MODE_AsKill */
#define MODE_EvFails		(1UL<<4)	/* List of all failed attempt events *//* MODE_AsFail */
#define MODE_EvBoth			(1UL<<5)
#define MODE_BjUnfinished	(1UL<<9)
#define MODE_BjStats		(1UL<<10)	/* Hourly snapshots */
#define MODE_Bj				(1UL<<11)
#define MODE_ChStats		(1UL<<20)	/* To-the-minute changelogs */
#define MODE_Ch				(1UL<<21)

/* Default mode */
static unsigned long mode = 0;

/* Event categories */
#define S_Err	(0)
#define S_Att	(1)	/* Attempts */
#define S_Kil	(2)	/* Kills */
#define S_Dth	(3)	/* Deaths */
#define S_Sur	(4)	/* Survivals */
#define S_Leg	(5)	/* Legal NP kills */
#define S_Ill	(6)	/* Illegal NP kills */
#define S_Non	(7)	/* NP attempts */
#define S_MAX	(8)

/* Sanity check on flag bit allocation */
#if (((MODE_Bj) << (S_MAX)) > (MODE_ChStats))
#error "** Number of event categories exceeds number of free MODE flag bits **"
#endif

/* Scoring table categories */
int pscdef[SCDEF_MAX_Single][SCDEF_MAX] =
{
	/*                 Against:  ---------Assassins--------- --------- Police ---------- --NonPlayer-- ------ ------ */
	/*                           --Unwanted--- ---Wanted---- --Unwanted--- ---Wanted---- ------------- Umpire - ?? - */
	/* -(Killer)--(NonFatal)--   -Comp- -Inco- -Comp- -Inco- -Comp- -Inco- -Comp- -Inco- -Legl- -Illg- ------ ------ */
	/* Assass Unwanted Comp */ { S_Att, S_Err, S_Att, S_Att, S_Att, S_Err, S_Att, S_Att, S_Non, S_Non, S_Err, S_Err },
	/* Assass Unwanted Inco */ { S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err },
	/* Assass Wanted   Comp */ { S_Att, S_Err, S_Att, S_Att, S_Att, S_Err, S_Att, S_Att, S_Non, S_Non, S_Err, S_Err },
	/* Assass Wanted   Inco */ { S_Att, S_Err, S_Att, S_Att, S_Att, S_Err, S_Att, S_Att, S_Non, S_Non, S_Err, S_Err },
	/* Police Unwanted Comp */ { S_Att, S_Err, S_Att, S_Att, S_Att, S_Err, S_Att, S_Att, S_Non, S_Non, S_Err, S_Err },
	/* Police Unwanted Inco */ { S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err },
	/* Police Wanted   Comp */ { S_Att, S_Err, S_Att, S_Att, S_Att, S_Err, S_Att, S_Att, S_Non, S_Non, S_Err, S_Err },
	/* Police Wanted   Inco */ { S_Att, S_Err, S_Att, S_Att, S_Att, S_Err, S_Att, S_Att, S_Non, S_Non, S_Err, S_Err },
	/*   Legal Non-Player   */ { S_Att, S_Err, S_Att, S_Att, S_Att, S_Err, S_Att, S_Att, S_Non, S_Non, S_Err, S_Err },
	/* Illegal Non-Player   */ { S_Att, S_Err, S_Att, S_Att, S_Att, S_Err, S_Att, S_Att, S_Non, S_Non, S_Err, S_Err },
	/* Umpire               */ { S_Att, S_Err, S_Att, S_Att, S_Att, S_Err, S_Att, S_Att, S_Non, S_Non, S_Err, S_Err },
	/* Unknown              */ { S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err },
	/* None                 */ { S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err },
	/* --------------(Fatal)--  ------ ------ ------ ------ ------ ------ ------ ------ ------ ------ ------ ------ */
	/* Assass Unwanted Comp */ { S_Kil, S_Err, S_Kil, S_Kil, S_Kil, S_Err, S_Kil, S_Kil, S_Leg, S_Ill, S_Err, S_Err },
	/* Assass Unwanted Inco */ { S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err },
	/* Assass Wanted   Comp */ { S_Kil, S_Err, S_Kil, S_Kil, S_Kil, S_Err, S_Kil, S_Kil, S_Leg, S_Ill, S_Err, S_Err },
	/* Assass Wanted   Inco */ { S_Kil, S_Err, S_Kil, S_Kil, S_Kil, S_Err, S_Kil, S_Kil, S_Leg, S_Ill, S_Err, S_Err },
	/* Police Unwanted Comp */ { S_Kil, S_Err, S_Kil, S_Kil, S_Kil, S_Err, S_Kil, S_Kil, S_Leg, S_Ill, S_Err, S_Err },
	/* Police Unwanted Inco */ { S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err },
	/* Police Wanted   Comp */ { S_Kil, S_Err, S_Kil, S_Kil, S_Kil, S_Err, S_Kil, S_Kil, S_Leg, S_Ill, S_Err, S_Err },
	/* Police Wanted   Inco */ { S_Kil, S_Err, S_Kil, S_Kil, S_Kil, S_Err, S_Kil, S_Kil, S_Leg, S_Ill, S_Err, S_Err },
	/*   Legal Non-Player   */ { S_Kil, S_Err, S_Kil, S_Kil, S_Kil, S_Err, S_Kil, S_Kil, S_Leg, S_Ill, S_Err, S_Err },
	/* Illegal Non-Player   */ { S_Kil, S_Err, S_Kil, S_Kil, S_Kil, S_Err, S_Kil, S_Kil, S_Leg, S_Ill, S_Err, S_Err },
	/* Umpire               */ { S_Kil, S_Err, S_Kil, S_Kil, S_Kil, S_Err, S_Kil, S_Kil, S_Leg, S_Ill, S_Err, S_Err },
	/* Unknown              */ { S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err },
	/* None                 */ { S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err },
	/* -(Victim)--(NonFatal)--  ------ ------ ------ ------ ------ ------ ------ ------ ------ ------ ------ ------ */
	/* Assass Unwanted Comp */ { S_Sur, S_Err, S_Sur, S_Sur, S_Sur, S_Err, S_Sur, S_Sur, S_Sur, S_Sur, S_Sur, S_Err },
	/* Assass Unwanted Inco */ { S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Sur, S_Err },
	/* Assass Wanted   Comp */ { S_Sur, S_Err, S_Sur, S_Sur, S_Sur, S_Err, S_Sur, S_Sur, S_Sur, S_Sur, S_Sur, S_Err },
	/* Assass Wanted   Inco */ { S_Sur, S_Err, S_Sur, S_Sur, S_Sur, S_Err, S_Sur, S_Sur, S_Sur, S_Sur, S_Sur, S_Err },
	/* Police Unwanted Comp */ { S_Sur, S_Err, S_Sur, S_Sur, S_Sur, S_Err, S_Sur, S_Sur, S_Sur, S_Sur, S_Sur, S_Err },
	/* Police Unwanted Inco */ { S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Sur, S_Err },
	/* Police Wanted   Comp */ { S_Sur, S_Err, S_Sur, S_Sur, S_Sur, S_Err, S_Sur, S_Sur, S_Sur, S_Sur, S_Sur, S_Err },
	/* Police Wanted   Inco */ { S_Sur, S_Err, S_Sur, S_Sur, S_Sur, S_Err, S_Sur, S_Sur, S_Sur, S_Sur, S_Sur, S_Err },
	/*   Legal Non-Player   */ { S_Sur, S_Err, S_Sur, S_Sur, S_Sur, S_Err, S_Sur, S_Sur, S_Sur, S_Sur, S_Sur, S_Err },
	/* Illegal Non-Player   */ { S_Sur, S_Err, S_Sur, S_Sur, S_Sur, S_Err, S_Sur, S_Sur, S_Sur, S_Sur, S_Sur, S_Err },
	/* Umpire               */ { S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err },
	/* Unknown              */ { S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err },
	/* None                 */ { S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err },
	/* -(Victim)-----(Fatal)--  ------ ------ ------ ------ ------ ------ ------ ------ ------ ------ ------ ------ */
	/* Assass Unwanted Comp */ { S_Dth, S_Err, S_Dth, S_Dth, S_Dth, S_Err, S_Dth, S_Dth, S_Dth, S_Dth, S_Sur, S_Err },
	/* Assass Unwanted Inco */ { S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Sur, S_Err },
	/* Assass Wanted   Comp */ { S_Dth, S_Err, S_Dth, S_Dth, S_Dth, S_Err, S_Dth, S_Dth, S_Dth, S_Dth, S_Sur, S_Err },
	/* Assass Wanted   Inco */ { S_Dth, S_Err, S_Dth, S_Dth, S_Dth, S_Err, S_Dth, S_Dth, S_Dth, S_Dth, S_Sur, S_Err },
	/* Police Unwanted Comp */ { S_Dth, S_Err, S_Dth, S_Dth, S_Dth, S_Err, S_Dth, S_Dth, S_Dth, S_Dth, S_Sur, S_Err },
	/* Police Unwanted Inco */ { S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Sur, S_Err },
	/* Police Wanted   Comp */ { S_Dth, S_Err, S_Dth, S_Dth, S_Dth, S_Err, S_Dth, S_Dth, S_Dth, S_Dth, S_Sur, S_Err },
	/* Police Wanted   Inco */ { S_Dth, S_Err, S_Dth, S_Dth, S_Dth, S_Err, S_Dth, S_Dth, S_Dth, S_Dth, S_Sur, S_Err },
	/*   Legal Non-Player   */ { S_Dth, S_Err, S_Dth, S_Dth, S_Dth, S_Err, S_Dth, S_Dth, S_Dth, S_Dth, S_Sur, S_Err },
	/* Illegal Non-Player   */ { S_Dth, S_Err, S_Dth, S_Dth, S_Dth, S_Err, S_Dth, S_Dth, S_Dth, S_Dth, S_Sur, S_Err },
	/* Umpire               */ { S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err },
	/* Unknown              */ { S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err },
	/* None                 */ { S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err }
	/* -----------------------  ------ ------ ------ ------ ------ ------ ------ ------ ------ ------ ------ ------ */
};

/***************************************************************************************************************************
** Print snapshots.                                                                                                       **
***************************************************************************************************************************/
int printbjoernstats(evolve__ctx_t* ectx, timeval at, FILE* fparr[STAT__MAX], int m, evolve_notes chk)
{
	pstate_p pstp;
	int i;
	char todo[STAT__MAX];

	/* Check each statistic */
	for (i = 0; (i < STAT__MAX); i++)
	{
		/* Needs updating? */
		if (!(todo[i] = ((fparr[i]) && ((!m) || (chk & chmask[i]))) ? 1 : 0)) continue;

		/* Print hour of game */
		if (!m) fprintf(fparr[i], " %lu", T_h(at) + (((u32b) 24) * (T_d(at) - T_d(TIME_genstatsMin))));

		/* Print time and date */
		else fprintf(fparr[i], " %lu\t"TDISP"", T_d(at), Thm(at));
	}

	/* Maintain reversed linked list */
	for (bj_pstp = 0, pstp = ectx->plist; (pstp); pstp = (bj_pstp = pstp)->next) pstp->ext = bj_pstp;

	/* Scan list in reverse order */
	for (pstp = bj_pstp; (pstp); pstp = pstp->ext)
	{
		int kills = pstp->scoretab[S_Kil] + pstp->scoretab[S_Leg] + pstp->scoretab[S_Ill];
		int fails = pstp->scoretab[S_Att] + pstp->scoretab[S_Non];

		if (todo[STAT_Score]) fprintf(fparr[STAT_Score], "\t%li", ((pstp->score + SCORE_ROUND) >> SCORE_SHIFT));
		if (todo[STAT_Kills]) fprintf(fparr[STAT_Kills], "\t%i", kills);
		if (todo[STAT_Fails]) fprintf(fparr[STAT_Fails], "\t%i", fails);
		if (todo[STAT_Deths]) fprintf(fparr[STAT_Deths], "\t%i", pstp->scoretab[S_Dth]);
		if (todo[STAT_Survs]) fprintf(fparr[STAT_Survs], "\t%i", pstp->scoretab[S_Sur]);
		if (todo[STAT_IllNP]) fprintf(fparr[STAT_IllNP], "\t%i", pstp->scoretab[S_Ill]);
		if (todo[STAT_LegNP]) fprintf(fparr[STAT_LegNP], "\t%i", pstp->scoretab[S_Leg]);
		if (todo[STAT_AttNP]) fprintf(fparr[STAT_AttNP], "\t%i", pstp->scoretab[S_Non]);
		if (todo[STAT_PKill]) fprintf(fparr[STAT_PKill], "\t%i", pstp->scoretab[S_Kil]);
		if (todo[STAT_PFail]) fprintf(fparr[STAT_PFail], "\t%i", pstp->scoretab[S_Att]);
	}

	/* Terminate lines */
	for (i = 0; (i < STAT__MAX); i++)
	{
		if (todo[i]) fputs("\n", fparr[i]);
	}

	return 0;
}

void floof(FILE* fp, const char* p1, const char* fmt, ...)
{
	pstate_p pstp;
	va_list ap;

	fputs(p1, fp);

	for (pstp = bj_pstp; (pstp); pstp = pstp->ext) fprintf(fp, "\t%s", pstp->pid);

	va_start(ap, fmt);
	vfprintf(fp, fmt, ap);
	va_end(ap);
}

/***************************************************************************************************************************
** Hooks to generate some statistics.                                                                                     **
***************************************************************************************************************************/
int evhook(event_p eptr, pstate_p p1s, pstate_p p2s, int pg1, int pg2, void* context, evolve__ctx_t* evolvectx)
{
	char* p = 0;
	FILE* fp = 0;

	/* Ignore irrelevent events */
	if ((eptr->flags & EF_Ignore) || !(eptr->flags & EF_Assassination)) return 0;

	/* Optionally, include annulled events */
	if ((eptr->flags & EF_Annulled) && !(mode & MODE_Annulled)) return 0;

	/* List kill events */
	if ((eptr->flags & EF_Kill) && (fp = fp_evkills)) { p = "  _____kill_at:"; ++num_evkills; }

	/* List kill attempts */
	else if (!(eptr->flags & EF_Kill) && (fp = fp_evfails)) { p = "  __attempt_at:"; ++num_evfails; }

	/* Print out summary */
	if (p) fprintf(fp, "%s\t%lu\t"TDISP"\t%s\t%s\t%s%s%s\n", p,
				   T_d(eptr->when), Thm(eptr->when), eptr->id, eptr->vict_pid,
				   (eptr->flags & EF_Annulled) ? "Annulled " : "",
				   (eptr->flags & EF_NoDangerBonus) ? "NoDangerBonus " : "",
				   (eptr->flags & EF_ScoreBonus) ? "ScoreBonus " : "");

	return 0;
}

int thook(event_p ep, int match, void* context, evolve__ctx_t* ectx)
{
	int i;
	timeval tim;

	/* Debug: summary of entry conditions */
	if (mode & MODE_Debug)
	{
		tim = (ep) ? (ep->when) : 0;
		fprintf(stderr, "%li:   thook when="TDEBfmt" match=%i ct="TDEBfmt" bj="TDEBfmt" et="TDEBfmt"\n",
				uniqid, TDEBparm(tim), match, TDEBparm(ectx->endtime), TDEBparm(bj_endtime), TDEBparm(endtime));
	}

	/* Has time changed, and are we doing changelogs? */
	if ((match <= 0) && (mode & MODE_ChStats))
	{
		/* New timeval */
		tim = (ep) ? (ep->when) : (endtime);

		/* Something interesting happened? */
		if (((ectx->lasttime < inittime) && (tim >= inittime)) || 
			(((tim = ectx->lasttime) >= inittime) && (ectx->notes & nbmask)))
		{
			/* Print some stats */
			if (i = printbjoernstats(ectx, tim, fp_chstats, 1, ectx->notes & nbmask)) return -(i);
	
			/* Dealt with */
			ectx->notes = 0;
		}
	}

	/* Nothing more to do here? */
	if ((match < 0) || !(mode & MODE_BjUnfinished)) return 0;

    /* Paranoia */
    if ((match != 0) && (match != 3) && (match != 4)) return -(ERF_InternalError);

    /* Limit is time of next event, or end time for stat collection */
    if ((!ep) || ((tim = ep->when) > bj_endtime)) tim = bj_endtime;

	/* Deal with all snapshot points up to and including the limit */
	while (ectx->endtime <= tim)
	{
		/* Time of snapshot */
		tim = ectx->endtime;

		/* Print some stats */
		if (i = printbjoernstats(ectx, ectx->endtime, fp_bjstats, 0, 0)) return -(i);

		/* Finished? */
		if (ectx->endtime >= bj_endtime) break;

		/* Next snapshot time */
		if (!(ectx->endtime = addhours(ectx->endtime, 1))) return -(ERF_BadTimeVal);
    	
		/* Fiddle to do snapshot at evolution limit */
		if (ectx->endtime > bj_endtime) ectx->endtime = bj_endtime;
	}

	/* All snapshots dealt with */
	if (tim >= bj_endtime)
	{
		ectx->flags &= ~EVOLVE_FuncEndCond;
		ectx->endtime = endtime;
		mode &= ~MODE_BjUnfinished;
	}

	/* Continue (retesting end conditions) */
	return 0;
}

/***************************************************************************************************************************
** Main code. Makes use of command line.                                                                                  **
***************************************************************************************************************************/
int main(int argc, char* argv[])
{
	char statbuf[MAX_statbuf], *statbuf_end = statbuf + MAX_statbuf, *statbuf_pos = statbuf;
	int i, rc = 0, lockfd = 0;
	evolve_flags ffl = 0;
	char *p, *timestr;
	pstate_p pslist = 0;


	/* Clear arrays */
	for (i = 0; (i < STAT__MAX); i++)
	{
		do_chstats[i] = do_bjstats[i] = 0;
		fp_chstats[i] = fp_bjstats[i] = 0;
	}

	/* Initialse stuff: time, file creation umask, pid, output title, don't redirect stderr to log */
	if (i = geninit(&timestr, 0117, 0, "Genstats")) erf(i, 1);

	/* Check command line args */
	for (i = 1; (i < argc) && (p = argv[i]); i++)
	{
		char **ap = 0;
		long m, mf = 0;

		/* Qualifier? */
		switch (*(p++))
		{
			case '+': m = 1; break;
			case '-': m = 2; break;
			case '^': m = 3; break;
			default : m = 0; --p;
		}

		/* Only first one or two letters of command switches are matched */
		switch (tolower(p[0]))
		{
			/* Flags */
			case 'd':	mf = MODE_Debug; break;
			case 'a':	mf = MODE_Annulled; break;

			/* Snapshot statistics */
			case 'b':	ap = do_bjstats;
			case 'c':	if (!ap) ap = do_chstats;
						switch (tolower(p[1]))
						{
							case 's': ap += STAT_Score; break;
							case 'k': ap += STAT_Kills; break;
							case 'f': ap += STAT_Fails; break;
							case 'd': ap += STAT_Deths; break;
							case 'v': ap += STAT_Survs; break;
							case 'i': ap += STAT_IllNP; break;
							case 'l': ap += STAT_LegNP; break;
							case 'n': ap += STAT_AttNP; break;
							case 'p': ap += STAT_PKill; break;
							case 'q': ap += STAT_PFail; break;
							default: goto ukparm;
						}
						break;

			/* Other statistics */
			case 's':	ap = &do_breakdown; break;
			case 'k':	ap = &do_evkills; break;
			case 'f':	ap = &do_evfails; break;

			/* Misc */
			case 't':
			{
				if ((m == 2) || ((m == 3) && (endtime))) endtime = 0;
				else if ((p = strchr(p, '=')) && (p[1] == '\0')) endtime = thetime;
				else if ((p) && (tmptime = str2tim(p+1))) endtime = tmptime;
				else fprintf(stderr, "%li: ! Bad time parameter: \"%s\"\n", uniqid, argv[i]);
				break;
			}

			/* Unknown */
			default:
			{
	ukparm:		fprintf(stderr, "%li: ! Ignoring unknown parameter: \"%s\"\n", uniqid, argv[i]);
			}
		}

		/* Do flags */
		if (mf) switch (m)
		{
			case 1: mode |=  mf; break;	/* Set flag */
			case 2: mode &= ~mf; break;	/* Clear flag */
			case 3: mode ^=  mf; break;	/* Toggle flag */
			default: fprintf(stderr, "%li: ! Ignoring unqualified parameter: \"%s\"\n", uniqid, argv[i]);
		}

		/* Handle output files */
		if (ap)
		{
			/* If any (possibly empty) filename given, activite stats mode */
			if (p = strchr(p, '=')) *ap = p + 1;

			/* Unparameterised - activate if appropriate, set to empty filename */
			else if ((m <= 1) || ((m == 3) && (*ap == 0))) *ap = "";

			/* Unparameterised - deactivate if appropriate */
			else *ap = 0;
		}
	}

	/* End time of evolution */
	if (!endtime) fprintf(stderr, "%li:   To end of game; ", uniqid);
	else fprintf(stderr, "%li:   To "TDISP", %s; ", uniqid, Thm(endtime), fdayname(T_d(endtime), 0));
	
	/* Print final parameter status */
	fprintf(stderr, "Args:%s%s%s%s%s",
			(mode & MODE_Debug) ? " Debug" : "",
			(mode & MODE_Annulled) ? " IncAnnulled" : "",
			(do_evkills) ? " ListKills" : "",
			(do_evfails) ? " ListFails" : "",
			(do_breakdown) ? " Breakdown"   : "");

	/* Print bjoernstat/changelog status */
	for (i = 0; (i < STAT__MAX); i++)
	{
		if (do_bjstats[i]) fprintf(stderr, " Bj%s", statit[i][0]);
		if (do_chstats[i]) fprintf(stderr, " Ch%s", statit[i][0]);
	}
	fputs(".\n", stderr);

	/* Time of termination is to be end of game, if unspecified */
	if (!endtime) endtime = TMAX;

	/* Bjoernstats */
	for (i = 0; (i < STAT__MAX); i++) if (do_bjstats[i])
	{
		/* Empty file name -> use stdout */
		if (!(*do_bjstats[i])) fp_bjstats[i] = stdout;

		/* Use given filename */
		else if (!(fp_bjstats[i] = fopen(do_bjstats[i], "wb"))) erf(ERF_fopenFailed, 8);


		/* Header */
		fprintf(fp_bjstats[i], "Bjoernstatistic %s for "GAME"; Generated at %s\n"
							   " Hour: \t%s\n"
							   "-------\t%s\n", statit[i][1], timestr, statit[i][2], statit[i][3]);
		/* Set mode flag */
		mode |= (MODE_BjStats | (MODE_Bj << i));
	}

	/* Changelogs */
	for (i = 0; (i < STAT__MAX); i++) if (do_chstats[i])
	{
		/* Empty file name -> use stdout */
		if (!(*do_chstats[i])) fp_chstats[i] = stdout;
	
		/* Use given filename */
		else if (!(fp_chstats[i] = fopen(do_chstats[i], "wb"))) erf(ERF_fopenFailed, 9);
	
		/* Header */
		fprintf(fp_chstats[i], "Changelog of %s for "GAME"; Generated at %s\n"
							   "  Day  \t Time  \t%s\n"
							   "-------\t-------\t%s\n", statit[i][1], timestr, statit[i][2], statit[i][3]);
		/* Set mode flag */
		mode |= (MODE_ChStats | (MODE_Ch << i));
	
		/* Set "scan changes" mask */
		nbmask |= chmask[i];
	}

	/* Open scoring breakdown file */
	if (do_breakdown)
	{
		/* Empty file name -> use stdout */
		if (!(*do_breakdown)) fp_breakdown = stdout;

		/* Use given filename */
		else if (!(fp_breakdown = fopen(do_breakdown, "wb"))) erf(ERF_fopenFailed, 9);

		/* Header */
		fprintf(fp_breakdown, "Scoring breakdown for "GAME"; Generated at %s\n", timestr);

		/* Set mode flag */
		mode |= MODE_Breakdown;
	}

	/* Open "list of kills" file */
	if (do_evkills)
	{
		/* Empty file name -> use stdout */
		if (!(*do_evkills)) fp_evkills = stdout;

		/* Use given filename */
		else if (!(fp_evkills = fopen(do_evkills, "wb"))) erf(ERF_fopenFailed, 9);

		/* Set mode flag */
		mode |= MODE_EvKills;

		/* Different file from that of list of attempts */
		if ((!do_evfails) || (strcmp(do_evkills, do_evfails)))
		{
			fprintf(fp_evkills, "List of kills for "GAME"; Generated at %s\n"
								" Event summary \t  Day  \t Time  \t By    \t On    \t Comments \n"
								"---------------\t-------\t-------\t-------\t-------\t----------\n", timestr);
		}
	}

	/* Open "list of attempts" file */
	if (do_evfails)
	{
		char *p;

		/* Same file as list of kills */
		if ((do_evkills) && (streq(do_evkills, do_evfails)))
		{
			fp_evfails = fp_evkills;
			mode |= MODE_EvBoth;
		}

		/* Empty file name -> use stdout */
		else if (!(*do_evfails)) fp_evfails = stdout;

		/* Use given filename */
		else if (!(fp_evfails = fopen(do_evfails, "wb"))) erf(ERF_fopenFailed, 9);

		/* Set mode flag */
		mode |= MODE_EvFails;

		/* Be slightly clever */
		p = (mode & MODE_EvBoth) ? "kills and attempts" : "attempts"; 

		/* Header */
		fprintf(fp_evfails, "List of %s for "GAME"; Generated at %s\n"
				" Event summary \t  Day  \t Time  \t By    \t On    \t Comments \n"
				"---------------\t-------\t-------\t-------\t-------\t----------\n", p, timestr);
	}

	/* Ensure evhook used */
	if (mode & (MODE_EvFails|MODE_EvKills)) ffl |= EVOLVE_FuncAssass;

	/* General changelog stuff */
	if (mode & MODE_ChStats)	
	{
		/* Ensure evhook used */
		ffl |= EVOLVE_FuncOnTick|EVOLVE_FuncExhaust|EVOLVE_Exclusive;

		/* Time of initial report */
		inittime = TIME_chstatsMin;
	}

	/* General bjoernstats stuff */
	if (mode & MODE_BjStats)	
	{
		/* Set mode flag */
		mode |= MODE_BjUnfinished;

		/* Ensure thook used consistently */
		ffl |= EVOLVE_FuncEndCond|EVOLVE_FuncExhaust|EVOLVE_Exclusive;

		/* Time of last score readout */
		bj_endtime = TIME_genstatsMax;

		/* Can't go past end of evolution time */
		if (bj_endtime > endtime) bj_endtime = endtime;

		/* Time of first score readout - Use it as effective limit */
		tmptime = TIME_genstatsMin;
	}

	/* Effective time limit */
	else tmptime = endtime;

	/* No parameters? */
	if (!mode) erf(0,0);

	/* Attempt to lock files */
	if (!(lockfd = setflock("Genstats.c "))) erf(ERF_flockFailed, 2);

	/* Evolve the game */
	if (rc = evolve(ffl|EVOLVE_TimersToLimit|EVOLVE_MatchAll|EVLadd_Comments|EVOLVE_InheritPseud|EVLadd_ScorePlayer,
					0,0, &pslist, fp_breakdown, pscdef, 0,0,0,0, tmptime, 0,
					statbuf_pos, statbuf_end, &statbuf_pos, evhook, 0, 0, thook, 0, 0)) erf(rc, 4);

	/* Bjoernstat/Changelog Footers */
	for (i = 0; (i < STAT__MAX); i++)
	{
		 if (fp_bjstats[i]) floof(fp_bjstats[i], "\nPlayer:", "\n\n*** Bjoernstat %s completed.\n", statit[i][0]);
		 if (fp_chstats[i]) floof(fp_chstats[i], "\nPlayer:\t", "\n\n*** Changelog for %s completed.\n", statit[i][0]);
	}

	/* Scoring Breakdown Footer */
	if (fp_breakdown) fprintf(fp_breakdown, "\n*** Scoring breakdown completed.\n");

	/* Event list footers */
	if (!(mode & MODE_EvBoth))
	{
		if (fp_evkills) fprintf(fp_evkills, "---------------\t-------\t-------\t-------\t-------\t----------\n"
											"*** List of (%d) kills completed.\n", num_evkills);
		if (fp_evfails) fprintf(fp_evfails, "---------------\t-------\t-------\t-------\t-------\t----------\n"
											"*** List of (%d) attempts completed.\n", num_evfails);
	}
	else fprintf(fp_evkills, "---------------\t-------\t-------\t-------\t-------\t----------\n"
							 "*** List of (%d) kills and (%d) attempts completed.\n", num_evkills, num_evfails);


_erf: /***** Error handling *****/

	/* Close output files */
	for (i = 0; (i < STAT__MAX); i++)
	{
		if ((fp_bjstats[i]) && (fp_bjstats[i] != stdout)) fclose(fp_bjstats[i]);
		if ((fp_chstats[i]) && (fp_chstats[i] != stdout)) fclose(fp_chstats[i]);
	}

	/* Be careful to close any shared files only once */
	if ((fp_evkills) && (fp_evkills != stdout) && (fp_evfails != fp_evkills)) fclose(fp_evkills);
	if ((fp_evfails) && (fp_evfails != stdout)) fclose(fp_evfails);
	if ((fp_breakdown) && (fp_breakdown != stdout)) fclose(fp_breakdown);

	/* Debug */
	if (rc) fprintf(stderr, "%li !! Error %s (%i) in main(%i) mode=%lx\n", uniqid,
					ERF_STR(rc), ERF_TYPE(rc), ERF_CONTEXT(rc), mode);

	/* Note success */
	else fprintf(stderr, "%li:   -------> finished <-------\n", uniqid);

	/* Tidy up */
	if (lockfd) setfunlock(lockfd);

	/* Only retain low bits of normal rc's */
	return (ERF_LOWBITS(rc));
}

/*** The end ***/
