/*****************************************************
**  lib-uinfo-scan.c                     (library)  **
**  ----------------------------------------------  **
**  Library: code to automate the gathering of      **
**  player data from the "uinfo.txt" file.          **
**  ----------------------------------------------  **
**  Updated to use new time system.                 **
**                                                  **
************** BORING LEGAL NOTICE *******************
**                                                  **
**  Assassins' Game (Semi-)Automatic Umpire System  **
**  ----------------------------------------------  **
**  Prerelease: Copyright 2000-2002 Edward M Nokes  **
**  This code may not be distributed in source or   **
**  executable form without the express consent of  **
**  the author, Edward M Nokes (emn23@cam.ac.uk).   **
**  The author accepts no responsibility or         **
**  liability for anything occurring as a result    **
**  of using this software, etc. etc.               **
**                                                  **
*****************************************************/

#ifdef included__lib_uinfo_scan
 #error "** lib-uinfo-scan multiply #included **"
#else
 #define included__lib_uinfo_scan fromfile

#ifndef included__lib_uinfo_read
 #include "!lib-uinfo-read.c"
#endif

/***************************************************************************************************************************
** Definitions                                                                                                            **
***************************************************************************************************************************/

/* Flags */
#define SCANPL_AddShort		(((scanplinfo_flags)1)<<0)	/* Add: record short info for additional players */
#define SCANPL_AddFull		(((scanplinfo_flags)1)<<1)	/* Add: record full info for additional players */
#define SCANPL_MustMatch	(((scanplinfo_flags)1)<<2)	/* Complain if no data found for an interesting player */
#define SCANPL_GetTargs		(((scanplinfo_flags)1)<<3)	/* Acquire targetting data */

/* Type to hold flags */
typedef unsigned long scanplinfo_flags;

/* Structure to hold scanplinfo__xxx helper function context */
struct scanplinfo__ctx
{
	scanplinfo_flags flags;
	pshort_p plist;
	player_p pfull, thispl;
	int plcnt, earlyterm;
};

/***************************************************************************************************************************
** Declarations                                                                                                           **
***************************************************************************************************************************/
player_p emptyplidx[1] = { 0 };

/***************************************************************************************************************************
** Helper functions                                                                                                       **
***************************************************************************************************************************/

/* Add a new player to list with state flags (fl) */
pshort_p scanplinfo__add(char* id, scanplinfo_flags fl, player_p pptr, struct scanplinfo__ctx* ctx)
{
	pshort_p pshp;

	/*fprintf(stderr, " (%s)-> \n", id);*/

	/* Allocate storage space */
	if (!(pshp = malloc(sizeof(struct pshort)))) return 0;

	/* Initialise as unseen */
	pshp->name = 0;
	pshp->pptr = 0;

	/* Copy player id */
	strncpy(pshp->id, id, MAX_id-1);
	pshp->id[MAX_id-1] = '\0';

	/* Link in at head of short list */
	pshp->link = ctx->plist;
	ctx->plist = pshp;

	/* Not recording full player information */
	if (!(fl & SCANPL_AddFull)) pshp->pptr = 0;

	/* Deal with full 'player' struct */
	else
	{
		/* No pre-allocated struct, so make one ourself */
		if ((!pptr) && !(pptr = malloc(sizeof(struct player)))) return 0;

		/*fprintf(stderr, " (moo|%s|%p|%p) \n", pshp->pid, ctx->pfull, pptr);*/

		/* Save it, and link at head of full linked list */
		(pshp->pptr = pptr)->link = ctx->pfull;
		ctx->pfull = pptr;

		/* Player id */
		pptr->id = pshp->id;

		/* No extension word or name, yet */
		pptr->ext  = 0;
		pptr->name = 0;

		/* No targets, etc. read yet */
		pptr->targs = pptr->asses = pptr->oldts = 0;

		/* The rest of the struct is uninitialised */
	}

	/* Increment player count */
	ctx->plcnt++;

	/* Success */
	return (pshp);
}

/* Deal with basic player information */
int scanplinfo__qf(char* pid, char* pname, void* context)
{
	int rc = 0;
	pshort_p pshp;
	struct scanplinfo__ctx* ctx = context;
	

	/* Try to match player id in existing (short) list */
	for (pshp = ctx->plist; (pshp); pshp = pshp->link) if (!strcmp(pid, pshp->id)) break;

	/* Collect data on uninteresting players? */
	if ((!pshp) && (ctx->flags & (SCANPL_AddFull|SCANPL_AddShort)))
	{
		/* Allocate new short (and possibly full) player structures */
		if (!(pshp = scanplinfo__add(pid, ctx->flags, 0, ctx))) erf(ERF_mallocFailed, 2);
	}

	/* Bother about this player? */
	if (pshp)
	{
		/* Paranoia: duplicates in uinfo file */
		if (pshp->name) erf(ERF_DuplicateIds, 3);

		/* Record player name */
		pshp->name = pname;

		/* Collect full player info? Arrange conditions for other helper function if so */
		if (ctx->thispl = pshp->pptr) return 1+8+32;

		/* Or don't, but protect buffered text from overwriting */
		else return 8;
	}

	/* Next, please */
    return 0;

_erf: /***** Error handling *****/

	/* No error, or reserved error */
	if (rc <= 0) rc = ERF_InternalError;

	/* Debug */
	fprintf(stderr, "%li:   scanplinfo__qf(%i):%i\n", uniqid, ERF_CONTEXT(rc), ERF_TYPE(rc));

	/* Only retain low bits of normal rc's */
	return (-ERF_LOWBITS(rc));
}

/* Record all player information */
int scanplinfo__pf(player_p pp, int state, void* context)
{
	int rc = 0;
	player_p pptr, lnk;
	struct scanplinfo__ctx* ctx = context;
	PLAYER_EXT_TYPE ext;


	/* Missing player struct pointer */
	if (!(pptr = ctx->thispl)) erf(ERF_InternalError, 1);

	/* Paranoia: duplicates in uinfo file */
	if (pptr->name) erf(ERF_DuplicateIds, 2);

	/*fprintf(stderr, "%li: scanplinfo %s\n", uniqid, pptr->id);*/

	/* Temporarily record extension word */
	ext = pptr->ext;
	lnk = pptr->link;

	/* Copy structure */
	*pptr = *pp;

	/* Restore overwritten extension word */
	pptr->ext = ext;
	pptr->link = lnk;

	/* Quick exit if all interesting players found */
	if (!(ctx->flags & (SCANPL_AddShort|SCANPL_AddFull)) && (--ctx->earlyterm <= 0)) return 1+8+2+32;

	/* Next, please (preserve text, reuse pp) */
    return 1+8+32;

_erf: /***** Error handling *****/

	/* No error, or reserved error */
	if (rc <= 0) rc = ERF_InternalError;

	/* Debug */
	fprintf(stderr, "%li:   scanplinfo__pf(%i):%i\n", uniqid, ERF_CONTEXT(rc), ERF_TYPE(rc));

	/* Only retain low bits of normal rc's */
	return (-ERF_LOWBITS(rc));
}

/***************************************************************************************************************************
** The highest-level function.                                                                                            **
** If non-zero, (ipids) is an array of interesting player ids, null terminated, for which full player information will be **
** stored. If (plidx) is non-null, and (*plidx) is non-zero, (*plidx) should be an array of pointers to (full) 'player'   **
** structs, each entry of which should point to a pre-allocated 'player' struct if non-null - if any entries are null, a  **
** suitable 'player' struct will be allocated for ecah, and the entries set to point to them. If (plidx) is non-null, but **
** (*plidx) is null, the array of pointers will be automatically allocated and filled in with pointers to automatically   **
** allocated 'player' structs, and the address of the pointer array returned as (*plidx). If (plong) is non-zero, then    **
** the head of the linked list of all full 'player' structs will be returned as (*plong). If (plout) is non-zero, then    **
** the head of the linked list of all short player information structures will be returned as (*plout). All interesting   **
** players will be in this list, with an additional link to their full 'player' struct; additional players may be given   **
** entries in the list according to the options. The 'pid' member of both types of player struct will always be correct,  **
** but the 'name' member of structs for interesting players may be null, if no information was found for such players.    **
***************************************************************************************************************************/
int scanplinfo(scanplinfo_flags flags, char** ipids, player_p** plidx, pshort_p* plout, player_p* plong,
			   char* buf, char* bufend, char** bufused)
{
	int rc = 0, nin = 0, i;
	struct scanplinfo__ctx ctx;
	pshort_p pshp;
	player_p pptr, tpptr;


	/* Initial context */
	ctx.pfull  = 0;
	ctx.thispl = 0;
	ctx.plist  = 0;
	ctx.plcnt  = 0;
	ctx.flags  = flags;

	/* Sort out contradictory options */
	if ((flags & SCANPL_AddShort) && (flags & SCANPL_AddFull)) erf(ERF_BadParms, 1);

	/* Interested in returning linked lists? Initially empty */
	if (plout) *plout = ctx.plist;
	if (plong) *plong = ctx.pfull;

 	/* Count initial player ids */
	if (ipids) while (ipids[nin]) nin++;

	/* Were there any in list? (Set count of unmatched interesting players for possible early termination) */
	if (ctx.earlyterm = nin)
	{
		/* Index array desired, but no initial array of player pointers passed... */
		if ((plidx) && !(*plidx))
		{
			/* Allocate array */
			if (!(*plidx = malloc(sizeof(player_p) * nin))) erf(ERF_mallocFailed, 2);

			/* Initialise it */
			for (nin = 0; (ipids[nin]); nin++) (*plidx)[nin] = 0;
		}

		/* Deal with each initial player id */
		for (nin = 0, pptr = 0; (ipids[nin]); nin++)
		{
			/* Empty player ids */
			if (ipids[nin][0] == '\0')
			{
				if (plidx) (*plidx)[nin] = 0;
                continue;
			}

			/* Search list of already initialised players for matching id */
			for (pshp = ctx.plist; (pshp); pshp = pshp->link) if (!strcmp(pshp->id, ipids[nin])) break;

			/* Not already allocated; proceed to initialise new structures */
			if (!pshp)
			{
				/* Has a full struct been pre-allocated? */
				tpptr = (plidx) ? ((*plidx)[nin]) : (0);

				/* Find space for short (and maybe full) structs, initialise to blank state */
				if (!(pshp = scanplinfo__add(ipids[nin], SCANPL_AddFull, tpptr, &ctx))) erf(ERF_mallocFailed, 3);

				/* Note last of initial players (additional ones linked in at head of list) */
				pptr = pshp->pptr;
			}

			/* Complain if two pre-allocated player structs correspond to the same pid */
			else if ((plidx) && (*plidx)[nin]) erf(ERF_BadParms, 4);

			/* Place pointer in index, if required */
			if (plidx) (*plidx)[nin] = pshp->pptr;
		}
	}

	/* Return pointer to blank index if initial list empty */
	else if (plidx) *plidx = emptyplidx;

	/* Read targets? */
	i = (flags & SCANPL_GetTargs) ? UREAD_GetTargs : 0;

	/* Step through uinfo file, helper function collects data */
	if (rc = uread(scanplinfo__qf, scanplinfo__pf, 0, buf, bufend, bufused, &ctx, i)) erf(rc, 5);

	/* Returning linked lists? */
	if (plout) *plout = ctx.plist;
 	if (plong) *plong = ctx.pfull;

	/* Option: Complain if not all interesting players matched */
	if (flags & SCANPL_MustMatch)
	{
		/* Check each 'player' struct's 'name' member */
		for (; (pptr); pptr = pptr->link) if (!pptr->name) erf(ERF_MissingPid, 6);
	}


_erf: /***** Error handling *****/

	/* No error, or reserved error */
	if (rc <= 0) return (rc);

	/* Debug */
	fprintf(stderr, "%li:   scanplinfo(%i):%i\n", uniqid, ERF_CONTEXT(rc), ERF_TYPE(rc));

	/* Only retain low bits of normal rc's */
	return (ERF_LOWBITS(rc));
}

#endif
/*** The End ***/
